Package: rvinecopulib
Type: Package
Title: High Performance Algorithms for Vine Copula Modeling
Version: 0.7.3.1.0
Authors@R: c(
        person("Thomas", "Nagler",, "info@vinecopulib.org", role = c("aut", "cre")),
        person("Thibault", "Vatter",, "info@vinecopulib.org", role = c("aut"))
        )
Description: Provides an interface to 'vinecopulib', a C++ library for vine 
 copula modeling. The 'rvinecopulib' package implements the core features of the
 popular 'VineCopula' package, in particular inference algorithms for both vine 
 copula and bivariate copula models. Advantages over 'VineCopula' are a sleeker 
 and more modern API, improved performances, especially in high dimensions, 
 nonparametric and multi-parameter families, and the ability to model discrete 
 variables. The 'rvinecopulib' package includes 'vinecopulib' as header-only 
 C++ library (currently version 0.7.2). Thus users do not need to install 
 'vinecopulib' itself in order to use 'rvinecopulib'. Since their initial 
 releases, 'vinecopulib' is licensed under the MIT License, and 'rvinecopulib' 
 is licensed under the GNU GPL version 3.
License: GPL-3 | file LICENSE
Encoding: UTF-8
NeedsCompilation: yes
Depends: R (>= 3.0.2)
Imports: assertthat, graphics, grDevices, kde1d (>= 1.1.0), lattice,
        Rcpp (>= 0.12.12), stats, utils
Suggests: igraph, ggplot2, ggraph, testthat
LinkingTo: BH, Rcpp, RcppEigen, RcppThread (>= 2.1.2), wdm (>= 0.2.6),
        kde1d (>= 1.1.0)
BugReports: https://github.com/vinecopulib/rvinecopulib/issues
URL: https://vinecopulib.github.io/rvinecopulib/
RoxygenNote: 7.3.2
Packaged: 2025-06-13 11:05:32 UTC; n5
Author: Thomas Nagler [aut, cre],
  Thibault Vatter [aut]
Maintainer: Thomas Nagler <info@vinecopulib.org>
Repository: CRAN
Date/Publication: 2025-06-13 12:20:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 03:12:13 UTC; unix
Archs: rvinecopulib.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as_rvine_matrix,list)
S3method(as_rvine_matrix,matrix)
S3method(as_rvine_matrix,rvine_matrix)
S3method(as_rvine_matrix,rvine_structure)
S3method(as_rvine_structure,list)
S3method(as_rvine_structure,matrix)
S3method(as_rvine_structure,rvine_matrix)
S3method(as_rvine_structure,rvine_structure)
S3method(coef,bicop_dist)
S3method(contour,bicop)
S3method(contour,bicop_dist)
S3method(contour,vinecop)
S3method(contour,vinecop_dist)
S3method(dim,rvine_matrix)
S3method(dim,rvine_structure)
S3method(dim,vine_dist)
S3method(dim,vinecop_dist)
S3method(fitted,bicop)
S3method(fitted,vine)
S3method(fitted,vinecop)
S3method(logLik,bicop)
S3method(logLik,vine)
S3method(logLik,vinecop)
S3method(plot,bicop)
S3method(plot,bicop_dist)
S3method(plot,rvine_matrix)
S3method(plot,rvine_structure)
S3method(plot,vinecop)
S3method(plot,vinecop_dist)
S3method(predict,bicop_dist)
S3method(predict,vine)
S3method(predict,vinecop)
S3method(print,bicop)
S3method(print,bicop_dist)
S3method(print,rvine_list)
S3method(print,rvine_matrix)
S3method(print,rvine_structure)
S3method(print,summary_df)
S3method(print,vine)
S3method(print,vine_dist)
S3method(print,vinecop)
S3method(print,vinecop_dist)
S3method(summary,bicop)
S3method(summary,bicop_dist)
S3method(summary,vine)
S3method(summary,vine_dist)
S3method(summary,vinecop)
S3method(summary,vinecop_dist)
S3method(truncate_model,rvine_matrix)
S3method(truncate_model,rvine_structure)
S3method(truncate_model,vine_dist)
S3method(truncate_model,vinecop_dist)
export(as.bicop)
export(as_rvine_matrix)
export(as_rvine_structure)
export(bicop)
export(bicop_dist)
export(cvine_structure)
export(dbicop)
export(dvine)
export(dvine_structure)
export(dvinecop)
export(emp_cdf)
export(get_all_families)
export(get_all_ktaus)
export(get_all_pair_copulas)
export(get_all_parameters)
export(get_family)
export(get_ktau)
export(get_matrix)
export(get_pair_copula)
export(get_parameters)
export(get_structure)
export(hbicop)
export(inverse_rosenblatt)
export(is.rvine_matrix)
export(is.rvine_structure)
export(ktau_to_par)
export(mBICV)
export(pairs_copula_data)
export(par_to_ktau)
export(pbicop)
export(pseudo_obs)
export(pvine)
export(pvinecop)
export(rbicop)
export(rosenblatt)
export(rvine)
export(rvine_matrix)
export(rvine_matrix_sim)
export(rvine_structure)
export(rvine_structure_sim)
export(rvinecop)
export(truncate_model)
export(vine)
export(vine_dist)
export(vinecop)
export(vinecop_dist)
importFrom(Rcpp,evalCpp)
importFrom(assertthat,"on_failure<-")
importFrom(assertthat,assert_that)
importFrom(assertthat,is.count)
importFrom(assertthat,is.flag)
importFrom(assertthat,is.number)
importFrom(assertthat,is.scalar)
importFrom(assertthat,is.string)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,rgb)
importFrom(graphics,abline)
importFrom(graphics,box)
importFrom(graphics,contour)
importFrom(graphics,hist.default)
importFrom(graphics,pairs.default)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,points.default)
importFrom(graphics,polygon)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(graphics,text.default)
importFrom(kde1d,dkde1d)
importFrom(kde1d,pkde1d)
importFrom(kde1d,qkde1d)
importFrom(lattice,wireframe)
importFrom(stats,coef)
importFrom(stats,cor)
importFrom(stats,dcauchy)
importFrom(stats,dchisq)
importFrom(stats,dexp)
importFrom(stats,df)
importFrom(stats,dgamma)
importFrom(stats,dlnorm)
importFrom(stats,dnorm)
importFrom(stats,dt)
importFrom(stats,dunif)
importFrom(stats,dweibull)
importFrom(stats,ecdf)
importFrom(stats,logLik)
importFrom(stats,model.matrix)
importFrom(stats,pbeta)
importFrom(stats,pcauchy)
importFrom(stats,pchisq)
importFrom(stats,pexp)
importFrom(stats,pf)
importFrom(stats,pgamma)
importFrom(stats,plnorm)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,punif)
importFrom(stats,pweibull)
importFrom(stats,qbeta)
importFrom(stats,qcauchy)
importFrom(stats,qchisq)
importFrom(stats,qexp)
importFrom(stats,qf)
importFrom(stats,qgamma)
importFrom(stats,qlnorm)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,qunif)
importFrom(stats,qweibull)
importFrom(stats,runif)
importFrom(utils,capture.output)
importFrom(utils,modifyList)
importFrom(utils,write.table)
useDynLib(rvinecopulib, .registration = TRUE)

rvinecopulib
==========

[![R-CMD-check](https://github.com/vinecopulib/rvinecopulib/actions/workflows/R-CMD-check.yaml/badge.svg?branch=main)](https://github.com/vinecopulib/rvinecopulib/actions/workflows/R-CMD-check.yaml)
[![Coverage Status](https://img.shields.io/codecov/c/github/vinecopulib/rvinecopulib/main.svg)](https://app.codecov.io/github/vinecopulib/rvinecopulib?branch=main)
[![CRAN version](https://www.r-pkg.org/badges/version/rvinecopulib)](https://cran.r-project.org/package=rvinecopulib) 
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/rvinecopulib)](https://cran.r-project.org/package=rvinecopulib)

Vine copulas are a flexible class of dependence models consisting of bivariate 
building blocks (see e.g., Aas et al., 2009). You can find a comprehensive 
list of publications and other materials on [vine-copula.org](https://www.math.cit.tum.de/math/forschung/gruppen/statistics/vine-copula-models/).

This package is the [R](https://cran.r-project.org/) API to the C++ library 
[vinecopulib](https://github.com/vinecopulib/vinecopulib), a header-only 
C++ library for vine copula models based on [Boost](https://www.boost.org/) and 
[Eigen](http://eigen.tuxfamily.org/index.php?title=Main_Page).

It provides high-performance implementations of the core features of the popular
[VineCopula R library](https://github.com/tnagler/VineCopula), in particular
inference algorithms for both vine copula and bivariate copula models.
Advantages over VineCopula are  
* a sleaker and more modern API,
* shorter runtimes, especially in high dimensions,
* nonparametric and multi-parameter families,
* ability to model discrete variables.

As VineCopula, the package is primarily made for the statistical analysis of 
**vine copula models**. The package includes tools for parameter estimation, 
model selection, simulation, and visualization. Tools for estimation, selection 
and exploratory data analysis of **bivariate copula** models are also provided. 
Please see the [API documentation](https://vinecopulib.github.io/rvinecopulib/) 
for a detailed description of all functions.

Table of contents
-----------------

- [rvinecopulib](#rvinecopulib)
  - [Table of contents](#table-of-contents)
  - [How to install](#how-to-install)
  - [Package overview](#package-overview)
    - [Bivariate copula modeling: bicop_dist and bicop](#bivariate-copula-modeling-bicopdist-and-bicop)
    - [Vine copula modeling: vinecop_dist and vinecop](#vine-copula-modeling-vinecopdist-and-vinecop)
    - [Bivariate copula families](#bivariate-copula-families)
  - [References](#references)

------------------------------------------------------------------------


How to install
--------------


You can install:

-   the stable release on CRAN:

    ``` r
    install.packages("rvinecopulib")
    ```

-   the latest development version:

    ``` r
    remotes::install_github("vinecopulib/rvinecopulib")
    ```

------------------------------------------------------------------------

Package overview
----------------

Below, we list most functions and features you should know about. As usual in 
copula models, data are assumed to be serially independent and lie in the unit
hypercube. 

### Bivariate copula modeling: bicop_dist and bicop

  * `bicop_dist`: Creates a bivariate copula by specifying the family, rotation 
    and parameters. Returns an object of class `bicop_dist`. The class has the
    following methods:
     
     * `print`: a brief overview of the bivariate copula. 
            
     * `plot`, `contour`: surface/perspective and contour plots of the copula
        density. Possibly coupled with standard normal margins (default for
        `contour`). 
        
  * `dbicop`, `pbicop`, `rbicop`, `hbicop`: Density, distribution function, random 
    generation and H-functions (with their inverses) for bivariate copula 
    distributions. Additionally to the evaluation points, you can provide 
    either `family`, `rotation` and `parameter`, or an object of class 
    `bicop_dist`.

  * `bicop`: Estimates parameters of a bivariate copula. Estimation can be done 
    by maximum likelihood (`par_method = "mle"`) or inversion of the empirical 
    Kendall's tau (`par_method = "itau"`, only available for one-parameter 
    families) for parametric families, and using local-likelihood 
    approximations of order zero/one/two for nonparametric models 
    (`nonpar_method="constant"`/`nonpar_method="linear"`/`nonpar_method="quadratic"`). 
    If `family_set` is a vector of families, then the family is selected using
    `selcrit="loglik"`, `selcrit="aic"` or `selcrit="bic"`. The function 
    returns an object of classes `bicop` and `bicop_dist`.
    The class `bicop` has the following following methods:
    
     * `print`: a more comprehensive overview of the bivariate copula model 
       with fit statistics. 
            
     * `predict`, `fitted`: predictions and fitted values for a bivariate 
       copula model.
       
     * `nobs`, `logLik`, `AIC`, `BIC`: usual fit statistics.

### Vine copula modeling: vinecop_dist and vinecop

  * `vinecop_dist`: Creates a vine copula by specifying a nested list of 
    `bicop_dist` objects and a quadratic structure matrix. 
    Returns an object of class `vinecop_dist`. The class has the
    following methods:
     
     * `print`, `summary`: a brief and more comprehensive overview of the vine 
       copula. 
            
     * `plot`: plots of the vine structure. 
        
  * `dvinecop`, `pvinecop`, `rvinecop`: Density, distribution function, random 
    generation for vine copula distributions. 

  * `vinecop`: automated fitting for vine copula models. The function inherits 
    the parameters of `bicop`. Optionally, a quadratic `matrix` can be used as 
    input to pre-specify the vine structure. `tree_crit` describes the 
    criterion for tree selection, one of `"tau"`, `"rho"`, `"hoeffd"` for 
    Kendall's tau, Spearman's rho, and Hoeffding's D, respectively.
    Additionally, `threshold` allows to threshold the `tree_crit` and 
    `trunc_lvl` to truncate the vine copula, with `threshold_sel` and 
    `trunc_lvl_sel` to automatically select both parameters. The function 
    returns an object of classes `vinecop` and `vinecop_dist`.
    The class has the `vinecop` has the following following methods:
    
     * `print`, `summary`: a brief and more comprehensive overview of the vine 
       copula with additional fit statistics information.
            
     * `predict`, `fitted`: predictions and fitted values for a vine 
       copula model.
       
     * `nobs`, `logLik`, `AIC`, `BIC`: usual fit statistics.

### Bivariate copula families

In this package several bivariate copula families are included for bivariate 
and multivariate analysis using vine copulas. It provides 
functionality of elliptical (Gaussian and Student-t) as well as Archimedean 
(Clayton, Gumbel, Frank, Joe, BB1, BB6, BB7 and BB8) copulas to cover a large
range of dependence patterns. For Archimedean copula families,
rotated versions are included to cover negative dependence as well. 
Additionally, nonparametric families are also supported.

| type          | name                  | name in R  |
| ------------- | --------------------- | ---------- |
| -             | Independence          | "indep"    |
| Elliptical    | Gaussian              | "gaussian" |
| "             | Student t             | "t"        |
| Archimedean   | Clayton               | "clayton"  |
| "             | Gumbel                | "gumbel"   |
| "             | Frank                 | "frank"    |
| "             | Joe                   | "joe"      |
| "             | Clayton-Gumbel (BB1)  | "bb1"      |
| "             | Joe-Gumbel (BB6)      | "bb6"      |
| "             | Joe-Clayton (BB7)     | "bb7"      |
| "             | Joe-Frank (BB8)       | "bb8"      |
| Nonparametric | Transformation kernel | "tll"      |

Note that several convenience vectors of families are included:
* `"all"` contains all the families
* `"parametric"` contains the parametric families (all except `"tll"`)
* `"nonparametric"` contains the nonparametric families (`"indep"` and `"tll"`)
* `"one_par"` contains the parametric families with a single parameter
(`"gaussian"`, `"clayton"`, `"gumbel"`, `"frank"`, and `"joe"`)
* `"two_par"` contains the parametric families with two parameters
(`"t"`, `"bb1"`, `"bb6"`, `"bb7"`, and `"bb8"`)
* `"elliptical"` contains the elliptical families
* `"archimedean"` contains the archimedean families
* `"BB"` contains the BB families
* `"itau"` families for which estimation by Kendall's tau inversion is available
(`"indep"`,`"gaussian"`, `"t"`,`"clayton"`, `"gumbel"`, `"frank"`, `"joe"`)

The following table shows the parameter ranges of bivariate copula families with 
one or two parameters:

| Copula family        | `par[1]`   | `par[2]`   |
| :------------------- | :--------- | :--------- |
| Gaussian             | `(-1, 1)`  | -          |
| Student t            | `(-1, 1)`  | `(2,Inf)`  |
| Clayton              | `(0, Inf)` | -          |
| Gumbel               | `[1, Inf)` | -          |
| Frank                | `R \ {0}`  | -          |
| Joe                  | `(1, Inf)` | -          |
| Clayton-Gumbel (BB1) | `(0, Inf)` | `[1, Inf)` |
| Joe-Gumbel (BB6)     | `[1 ,Inf)` | `[1, Inf)` |
| Joe-Clayton (BB7)    | `[1, Inf)` | `(0, Inf)` |
| Joe-Frank (BB8)      | `[1, Inf)` | `(0, 1]`   |

------------------------------------------------------------------------

References
----------

Aas, K., C. Czado, A. Frigessi, and H. Bakken (2009). Pair-copula constructions 
of multiple dependence. Insurance: Mathematics and Economics 44 (2), 182-198.

#' Internal: Turn vector input into a matrix with two columns
#'
#' @param u input data
#' @param to_col if `u` is a vector, then `to_col = FALSE` (respectively
#' `to_col = TRUE`) transforms it into a matrix with a single row (respectively
#' single column)
#'
#'
#' @return either a matrix, or an error if u is neither a matrix, data.frame,
#' or a vector
#'
#' @noRd
if_vec_to_matrix <- function(u, to_col = FALSE) {
  if (is.null(u))
    return(NULL)
  assert_that(is.numeric(u) | is.data.frame(u))
  if (NCOL(u) == 1) {
    if (to_col) {
      u <- matrix(u, length(u), 1)
    } else {
      u <- matrix(u, 1, length(u))
    }
  }
  if (!is.matrix(u)) {
    u <- as.matrix(u)
  }

  u
}

#' Internal: Convert arguments to `bicop_dist` object.
#' @param family the family as passed in function call.
#' @param rotation the rotation as passed in function call.
#' @param parameters the parameters as passed in function call.
#' @return A `bicop_dist` object.
#' @noRd
args2bicop <- function(family, rotation, parameters, var_types = c("c", "c")) {
  if (all(inherits(family, "bicop_dist"))) {
    return(family)
  } else {
    if (missing(rotation)) {
      rotation <- 0
    }
    if (missing(parameters)) {
      parameters <- numeric(0)
    }
    assert_that(is.string(family), is.number(rotation), is.numeric(parameters))
    return(bicop_dist(family, rotation, parameters, var_types))
  }
}

process_family_set <- function(family_set, par_method) {
  family_set <- check_and_match_family_set(family_set)
  family_set <- expand_family_set(family_set)
  if (par_method == "itau") {
    if (any(!(family_set %in% family_set_itau))) {
      warning("Only families (",
        paste(family_set_itau, collapse = ", "),
        ") can be used with ", "'par_method = ", '"itau"', "'; ",
        "reducing family set.",
        call. = FALSE
      )
      family_set <- intersect(family_set, family_set_itau)
    }
  }

  family_set
}

#' Internal: Expand shortcuts in the familyset.
#' @noRd
expand_family_set <- function(family_set) {
  unique(unlist(lapply(family_set, expand_family)))
}

expand_family <- function(family) {
  switch(
    family,
    "archimedean" = family_set_archimedean,
    "elliptical" = family_set_elliptical,
    "bbs" = family_set_bb,
    "oneparametric" = family_set_onepar,
    "twoparametric" = family_set_twopar,
    "parametric" = family_set_parametric,
    "nonparametric" = family_set_nonparametric,
    "itau" = family_set_itau,
    "all" = family_set_all,
    family # default is no expansion
  )
}

#' Internal: Checks whether all families are known (including partial matching).
#' @noRd
check_and_match_family_set <- function(family_set) {
  matched_fams <- family_set_all_defs[pmatch(family_set, family_set_all_defs)]
  if (any(is.na(matched_fams))) {
    stop(
      "unknown families in family_set: ",
      paste0('"', family_set[is.na(matched_fams)], '"', collapse = ", ")
    )
  }
  matched_fams
}

# Multiple plot function
#
# ggplot objects can be passed in ..., or to plotlist (as a list of ggplot
# objects) - cols:   Number of columns in layout - layout: A matrix specifying
# the layout. If present, 'cols' is ignored.
#
# If the layout is something like matrix(c(1,2,3,3), nrow=2, byrow=TRUE), then
# plot 1 will go in the upper left, 2 will go in the upper right, and 3 will go
# all the way across the bottom.
#
multiplot <- function(..., plotlist = NULL, file, cols = 1, layout = NULL) {
  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)

  numPlots <- length(plots)

  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout)) {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots / cols)),
      ncol = cols, nrow = ceiling(numPlots / cols)
    )
  }

  if (numPlots == 1) {
    print(plots[[1]])
  } else {
    # Set up the page
    grid::grid.newpage()
    grid::pushViewport(grid::viewport(
      layout =
        grid::grid.layout(
          nrow(layout),
          ncol(layout)
        )
    ))

    # Make each plot, in the correct location
    for (i in 1:numPlots) {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))

      print(plots[[i]],
        vp = grid::viewport(
          layout.pos.row = matchidx$row,
          layout.pos.col = matchidx$col
        )
      )
    }
  }
}

# Get the depth of a list
depth <- function(this) ifelse(is.list(this), 1L + max(sapply(this, depth)), 0L)

supported_distributions <- c(
  "beta", "cauchy", "chisq", "exp", "f", "gamma",
  "lnorm", "norm", "t", "unif", "weibull"
)

#' @importFrom stats pbeta qbeta qbeta dcauchy pcauchy qcauchy dchisq pchisq
#' @importFrom stats qchisq dexp pexp qexp df pf qf dgamma pgamma qgamma
#' @importFrom stats dlnorm plnorm qlnorm dt pt qt dunif punif qunif
#' @importFrom stats dweibull pweibull qweibull
check_distr <- function(distr) {

  ## if provided with a kde1d object, then there is nothing to do
  if (inherits(distr, "kde1d")) {
    return(TRUE)
  }

  ## basic sanity checks
  if (!is.list(distr)) {
    return("a distribution should be a kde1d object or a list")
  }
  if (!any(is.element(names(distr), "distr"))) {
    return("a distribution should be a kde1d object or a list with a 'distr' element")
  }
  nn <- distr[["distr"]]
  if (!is.element(nn, supported_distributions)) {
    return("the provided name does not belong to supported distributions")
  }

  ## check that the provided parameters are consistent with the distribution
  qfun <- get(paste0("q", nn))
  par <- distr[names(distr) != "distr"]
  par$p <- 0.5
  e <- tryCatch(do.call(qfun, par), error = function(e) e)
  if (any(class(e) == "error")) {
    return(e$message)
  }

  return(TRUE)
}

get_npars_distr <- function(distr) {
  switch(distr$distr,
    beta = 2,
    cauchy = 2,
    chisq = ifelse("ncp" %in% names(distr), 2, 1),
    exp = 1,
    f = 3,
    gamma = 2,
    lnorm = 2,
    norm = 2,
    t = ifelse("ncp" %in% names(distr), 3, 2),
    unif = 2,
    weibull = ifelse("scale" %in% names(distr), 2, 1)
  )
}

#' @param data the data (after expand_factors() was called).
#' @param ctrl the margin controls object (after expand_margin_controls() was
#'   called).
#' @noRd
check_margin_controls <- function(data, ctrl) {
  nms <- colnames(data)
  if (is.null(nms)) {
    nms <- as.character(seq_len(ncol(data)))
  }
  lapply(seq_len(NCOL(data)), function(k) {
    msg_var <- paste0("Problem with margin_controls for variable ", nms[k], ": ")
    tryCatch(
      assert_that(
        is.numeric(ctrl$mult[k]), ctrl$mult[k] > 0,
        is.numeric(ctrl$xmin[k]), is.numeric(ctrl$xmax[k]),
        is.na(ctrl$bw[k]) | (is.numeric(ctrl$bw[k]) & (ctrl$bw[k] > 0)),
        is.numeric(ctrl$deg[k])
      ),
      error = function(e) stop(msg_var, e$message)
    )

    if (is.ordered(data[, k]) & (!is.nan(ctrl$xmin[k]) | !is.nan(ctrl$xmax[k]))) {
      stop(msg_var, "xmin and xmax are not meaningful for x of type ordered.")
    }

    if (!is.nan(ctrl$xmax[k]) & !is.nan(ctrl$xmin[k])) {
      if (ctrl$xmin[k] > ctrl$xmax[k]) {
        stop(msg_var, "xmin is larger than xmax.")
      }
    }
    if (!is.nan(ctrl$xmin[k])) {
      if (any(data[, k] < ctrl$xmin[k])) {
        stop(msg_var, "not all data are larger than xmin.")
      }
    }
    if (!is.nan(ctrl$xmax[k])) {
      if (any(data[, k] > ctrl$xmax[k])) {
        stop(msg_var, "not all data are samller than xmax.")
      }
    }
    if (!(ctrl$deg[k] %in% 0:2)) {
      stop(msg_var, "deg must be either 0, 1, or 2.")
    }
  })
}


#' @noRd
#' @importFrom assertthat assert_that on_failure<-
#' @importFrom assertthat is.number is.string is.flag is.scalar
in_set <- function(el, set) {
  all(el %in% set)
}


on_failure(in_set) <- function(call, env) {
  paste0(
    deparse(call$el),
    " must be one of {",
    paste0(eval(call$set, env), collapse = ", "),
    "}."
  )
}

correct_var_types <- function(var_types, data) {
  is.character(var_types) && in_set(var_types, c("c", "d"))
}

on_failure(correct_var_types) <- function(call, env) {
  paste0("var_types must be vector with elements 'c' or 'd'.")
}

#' Pseudo-Observations
#'
#' Compute the pseudo-observations for the given data matrix.
#'
#' @param x vector or matrix random variates to be converted (column wise) to
#' pseudo-observations.
#' @param ties_method similar to `ties.method` of [rank()] (only `"average"`,
#' `"first"` and `"random"` currently available).
#' @param lower_tail `logical` which, if `FALSE``, returns the
#'   pseudo-observations when applying the empirical marginal survival
#'   functions.
#' @details
#' Given `n` realizations \eqn{x_i=(x_{i1}, \ldots,x_{id})},
#' \eqn{i \in \left\lbrace 1, \ldots,n \right\rbrace }
#' of a random vector `X`, the pseudo-observations are defined via
#' \eqn{u_{ij}=r_{ij}/(n+1)} for
#' \eqn{i \in \left\lbrace 1, \ldots,n \right\rbrace}
#' and
#' \eqn{j \in \left\lbrace 1, \ldots,d \right\rbrace }, where
#' \eqn{r_{ij}} denotes the rank of \eqn{x_{ij}} among all \eqn{x_{kj}},
#' \eqn{k \in \left\lbrace 1, \ldots,n \right\rbrace }.
#'
#' The pseudo-observations can thus also be computed by component-wise applying
#' the empirical distribution functions to the data and scaling the result by
#' \eqn{n/(n+1)}. This asymptotically negligible scaling factor is used to force
#' the variates to fall inside the open unit hypercube, for example, to avoid
#' problems with density evaluation at the boundaries.
#'
#' When `lower_tail = FALSE`, then `pseudo_obs()` simply returns
#' `1 - pseudo_obs()`.
#'
#' @return
#' a vector of matrix of the same dimension as the input containing the
#' pseudo-observations.
#' @examples
#' # pseudo-observations for a vector
#' pseudo_obs(rnorm(10))
#'
#' # pseudo-observations for a matrix
#' pseudo_obs(cbind(rnorm(10), rnorm(10)))
#' @export
pseudo_obs <- function(x, ties_method = "average", lower_tail = TRUE) {
  assert_that(is.scalar(lower_tail) && is.logical(lower_tail))
  assert_that(is.character(ties_method) && is.scalar(ties_method))
  assert_that(in_set(ties_method, c("average", "first", "random")))
  assert_that(is.numeric(x) || is.matrix(x) || is.data.frame(x))

  x[] <- pseudo_obs_cpp(if_vec_to_matrix(x, TRUE), ties_method)
  if (!lower_tail) {
    x <- 1 - x
  }
  x
}

#' Corrected Empirical CDF
#'
#' The empirical CDF with tail correction, ensuring that its output is never
#' 0 or 1.
#'
#' @details The corrected empirical CDF is defined as
#' \deqn{
#' F_n(x) = \frac{1}{n + 1} \min\biggl\{1, \sum_{i = 1}^n 1(X_i \le x)\biggr\}
#' }
#'
#' @param x numeric vector of observations
#'
#' @return A function with signature `function(x)` that returns \eqn{F_n(x)}.
#'
#' @importFrom stats ecdf
#' @export
#' @examples
#' # fit ECDF on simulated data
#' x <- rnorm(100)
#' cdf <- emp_cdf(x)
#'
#' # output is bounded away from 0 and 1
#' cdf(-50)
#' cdf(50)
emp_cdf <- function(x) {
  assert_that(is.numeric(x))
  n <- length(x)
  Fn <- ecdf(x)
  function(xx) pmax(n * Fn(xx), 1) / (n + 1)
}

#' Truncates output of model data frames.
#'
#' @param x a `data.frame` whose print output should be truncated.
#' @noRd
#' @export
print.summary_df <- function(x, ..., rows = 1:10) {
  assert_that(is.numeric(rows), all(rows > 0))
  rows <- intersect(seq_len(nrow(x)), rows)
  x_print <- x[rows, ]
  cat("# A data.frame:", nrow(x), "x", ncol(x), "\n")
  print.data.frame(x_print, digits = 2, row.names = FALSE)
  if (nrow(x) > length(rows)) {
    cat("# ... with", nrow(x) - length(rows), "more rows\n")
  }
  invisible(x)
}

#' internal function
#' @noRd
print_truncation_info <- function(x) {
  if (dim(x)[2] < dim(x)[1] - 1) {
    cat(", ", dim(x)[2], "-truncated", sep = "")
  }
  cat("\n")
}

#' internal function
#' @noRd
print_fit_info <- function(x) {
  ll <- logLik(x)
  cat("nobs =", x$nobs, "  ")
  cat("logLik =", round(ll[1], 2), "  ")
  cat("npars =", round(attr(ll, "df"), 2), "  ")
  cat("AIC =", round(-2 * ll[1] + 2 * attr(ll, "df"), 2), "  ")
  cat("BIC =", round(-2 * ll[1] + log(x$nobs) * attr(ll, "df"), 2), "  ")
  cat("\n")
}

#' internal function : synchronize C++ random number generators with R
#' @importFrom stats runif
#' @noRd
get_seeds <- function() {
  as.numeric(sprintf("%20.0f", runif(20, 1e6, 1e7)))
}

test_that("pairs_copula_data works", {
  u <- replicate(3, runif(10))
  expect_silent(pairs_copula_data(u))
})

test_that("summary_df is printed correctly", {
  x <- data.frame(a = 1:20, B = "A")
  output <- capture_output_lines(x2 <- print.summary_df(x))
  expect_identical(x, x2)
  expect_length(output, 13)
})

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Exported statistical tools")

set.seed(0)

test_that("'pseudo_obs' works", {
  x <- rnorm(10)
  expect_eql(sort(pseudo_obs(x)), (1:10) / 11)
  expect_eql(sort(pseudo_obs(x, lower_tail = FALSE)), (1:10) / 11)
  expect_eql(pseudo_obs(x), 1 - pseudo_obs(x, lower_tail = FALSE))

  mat <- cbind(rnorm(10), rnorm(10))
  expect_eql(sort(pseudo_obs(mat)[, 1]), (1:10) / 11)
  expect_eql(sort(pseudo_obs(mat)[, 2]), (1:10) / 11)

  df <- data.frame(a = rnorm(10), b = rnorm(10))
  expect_eql(sort(pseudo_obs(df)[, 1]), (1:10) / 11)
  expect_eql(sort(pseudo_obs(df)[, 2]), (1:10) / 11)

  expect_is(pseudo_obs(x), "numeric")
  expect_is(pseudo_obs(mat), "matrix")
  expect_is(pseudo_obs(df), "data.frame")
  expect_eql(names(pseudo_obs(df)), c("a", "b"))
  expect_error(pseudo_obs("something"))
  expect_error(pseudo_obs(x, "something"))
})

test_that("'emp_cdf' works", {
  x <- 1:10
  cdf <- emp_cdf(x)

  expect_eql(pseudo_obs(x), cdf(x))
  expect_eql(cdf(c(0, 11)), c(1, 10) / 11)

  expect_is(cdf(x), "numeric")
  expect_is(cdf, "function")
  expect_error(cdf(data.frame(x = 1:10)))
})

#' Coerce various kind of objects to R-vine structures and matrices
#'
#' `as_rvine_structure` and `as_rvine_matrix` are new S3 generics allowing to
#' coerce objects into R-vine structures and matrices (see [rvine_structure()]
#' and [rvine_matrix()]).
#'
#' The coercion to `rvine_structure` and `rvine_matrix` can be applied to
#' different kind of objects Currently, `rvine_structure`, `rvine_matrix`,
#' `matrix` and `list` are supported.
#'
#' For `as_rvine_structure`:
#'
#'  * `rvine_structure` : the main use case is to re-check an object via
#'  `validate = TRUE`.
#'  * `rvine_matrix` and `matrix` : allow to coerce matrices into R-vine
#'  structures (see [rvine_structure()] for more details). The main difference
#'  between `rvine_matrix` and `matrix` is the nature of the validity
#'  checks.
#'  * `list` : must contain named elements `order` and `struct_array` to be
#'  coerced into an R-vine structure (see [rvine_structure()] for more details).
#'
#' For `as_rvine_matrix`:
#'
#'  * `rvine_structure` : allow to coerce an `rvine_structure` into an
#'  R-vine matrix (useful e.g. for printing).
#'  * `rvine_matrix`: similar to `as_rvine_structure` for `rvine_structure`,
#'  the main use case is to re-check an object via `validate = TRUE`.
#'  * `matrix` : allow to coerce matrices into R-vine
#'  matrices (mainly by checking that the matrix defines a valid
#'  R-vine, see [rvine_matrix()] for more details).
#'  * `list` : must contain named elements `order` and `struct_array` to be
#'  coerced into an R-vine matrix (see [rvine_structure()] for more details).
#'
#' @param x An object of class `rvine_structure`, `rvine_matrix`, `matrix` or
#' `list` that can be coerced into an R-vine structure or R-vine matrix
#' (see *Details*).
#' @param ... Other arguments passed on to individual methods.
#' @return Either an object of class `rvine_structure` or of class
#' `rvine_matrix` (see [rvine_structure()] or [rvine_matrix()]).
#' @examples
#' # R-vine structures can be constructed from the order vector and struct_array
#' rvine_structure(order = 1:4, struct_array = list(
#'   c(4, 4, 4),
#'   c(3, 3),
#'   2
#' ))
#'
#' # ... or a similar list can be coerced into an R-vine structure
#' as_rvine_structure(list(order = 1:4, struct_array = list(
#'   c(4, 4, 4),
#'   c(3, 3),
#'   2
#' )))
#'
#' # similarly, standard matrices can be coerced into R-vine structures
#' mat <- matrix(c(4, 3, 2, 1, 4, 3, 2, 0, 4, 3, 0, 0, 4, 0, 0, 0), 4, 4)
#' as_rvine_structure(mat)
#'
#' # or truncate and construct the structure
#' mat[3, 1] <- 0
#' as_rvine_structure(mat)
#'
#' # throws an error
#' mat[3, 1] <- 5
#' try(as_rvine_structure(mat))
#' @seealso rvine_structure rvine_matrix
#' @rdname as_rvine_structure
#' @aliases as_rvine_matrix
#' @export
as_rvine_structure <- function(x, ...) {
  UseMethod("as_rvine_structure")
}

#' @export
#' @rdname as_rvine_structure
as_rvine_matrix <- function(x, ...) {
  UseMethod("as_rvine_matrix")
}


#' @param validate When `TRUE``, verifies that the input is a valid
#' rvine-structure (see *Details*). You may want to suppress this when you
#' know that you already have a valid structure and you want to save some time,
#' or to explicitly enable it if you have a structure that you want to re-check.
#' @export
#' @rdname as_rvine_structure
as_rvine_structure.rvine_structure <- function(x, ..., validate = FALSE) {
  assert_that(is.flag(validate))
  if (validate) {
    validate_rvine_structure(x)
  } else {
    assert_that(is.rvine_structure(x))
  }

  x
}

#' @export
#' @rdname as_rvine_structure
as_rvine_matrix.rvine_structure <- function(x, ..., validate = FALSE) {
  assert_that(is.flag(validate))
  if (validate) {
    validate_rvine_structure(x)
  } else {
    assert_that(is.rvine_structure(x))
  }

  # extract order and dimension
  order <- x$order
  d <- dim(x)[1]
  trunc_lvl <- dim(x)[2]

  # set-up output
  matrix <- matrix(0, d, d)

  # fill output
  diag(matrix) <- order
  matrix <- matrix[d:1, ]
  for (i in seq_len(min(trunc_lvl, d - 1))) {
    newrow <- order[x[["struct_array"]][[i]]]
    matrix[i, 1:length(newrow)] <- newrow
  }

  class(matrix) <- c("rvine_matrix", class(matrix))
  attr(matrix, "d") <- d
  attr(matrix, "trunc_lvl") <- dim(x)[2]
  matrix
}

#' @param is_natural_order A flag indicating whether the `struct_array` element
#' of `x` is assumed to be provided in natural order already (a structure is in
#' natural order if the anti-diagonal is 1, .., d from bottom left to top
#' right).
#' @export
#' @rdname as_rvine_structure
as_rvine_structure.list <- function(x, ..., is_natural_order = FALSE) {
  assert_that(
    is.list(x),
    all(c("order", "struct_array") %in% names(x)),
    is.flag(is_natural_order)
  )

  rvine_structure(
    x[["order"]],
    x[["struct_array"]],
    is_natural_order
  )
}

#' @export
#' @rdname as_rvine_structure
as_rvine_matrix.list <- function(x, ..., is_natural_order = FALSE) {
  assert_that(
    is.list(x),
    all(c("order", "struct_array") %in% names(x)),
    is.flag(is_natural_order)
  )

  rvine_struct <- rvine_structure(
    x[["order"]],
    x[["struct_array"]],
    is_natural_order
  )

  as_rvine_matrix.rvine_structure(rvine_struct)
}


#' @export
#' @rdname as_rvine_structure
as_rvine_structure.rvine_matrix <- function(x, ..., validate = FALSE) {
  assert_that(is.flag(validate))
  if (validate) {
    validate_rvine_matrix(x)
  } else {
    assert_that(is.rvine_matrix(x))
  }

  # compute structure array in natural order
  d <- dim(x)[1]
  order <- order(diag(x[d:1, ]))
  struct_array <- lapply(1:(d - 1), function(i) order[x[i, 1:(d - i)]])

  # create and return x
  structure(list(
    order = diag(x[d:1, ]),
    struct_array = struct_array,
    d = d,
    trunc_lvl = dim(x)[2]
  ),
  class = c("rvine_structure", "list")
  )
}

#' @export
#' @rdname as_rvine_structure
as_rvine_matrix.rvine_matrix <- function(x, ..., validate = FALSE) {
  assert_that(is.flag(validate))
  if (validate) {
    validate_rvine_matrix(x)
  } else {
    assert_that(is.rvine_matrix(x))
  }

  x
}


#' @export
#' @rdname as_rvine_structure
as_rvine_structure.matrix <- function(x, ..., validate = TRUE) {
  x <- as_rvine_matrix(x, validate)
  as_rvine_structure.rvine_matrix(x)
}

#' @export
#' @rdname as_rvine_structure
as_rvine_matrix.matrix <- function(x, ..., validate = TRUE) {
  assert_that(is.flag(validate))
  if (validate) {
    x <- rvine_matrix(x)
  } else {
    assert_that(is.matrix(x) && is.numeric(x))
    x <- rvine_matrix_nocheck(x)
  }

  x
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Truncation generic")

mat <- matrix(c(3, 2, 1, 3, 2, 0, 3, 0, 0), 3, 3)

test_that("works with rvine_structure", {
  struct <- truncate_model(as_rvine_structure(mat), 1)

  expect_eql(unname(dim(struct)), c(3, 1))
  expect_eql(struct$order, 1:3)
  expect_length(struct$struct_array[[1]], 2)
  expect_warning(expect_identical(struct, truncate_model(struct, 2)))
  expect_error(truncate_model(struct, 7))
})

test_that("works with rvine_matrix", {
  mat <- truncate_model(as_rvine_matrix(mat), 1)

  expect_eql(unname(dim(mat)), c(3, 1))
  expect_eql(mat[1, ], c(3, 3, 3))
  expect_eql(mat[2, ], c(0, 2, 0))
  expect_eql(mat[3, ], c(1, 0, 0))
})

test_that("works with vinecop objects", {
  u <- replicate(4, runif(20))
  vc <- vinecop(u, family = "gauss")
  vc_trunc <- truncate_model(vc, 1)

  expect_eql(vc_trunc$npars, 3)
  expect_eql(
    vc_trunc$loglik,
    sum(sapply(vc$pair_copulas[[1]], function(x) x$loglik))
  )
  expect_eql(unname(dim(vc_trunc)), c(4, 1))
  expect_eql(vc_trunc$pair_copulas, vc$pair_copulas[1])
  expect_length(vc_trunc$structure$struct_array[[1]], 3)
  expect_warning(expect_identical(vc_trunc, truncate_model(vc_trunc, 2)))

  expect_warning(get_all_pair_copulas(vc_trunc, 2))
})

test_that("works with vine objects", {
  x <- replicate(3, runif(20))
  vd <- vine(x, copula_controls = list(family = "gauss"))
  vd_trunc <- truncate_model(vd, 1)

  expect_lt(vd_trunc$npars, vd$npars)
  expect_lt(vd_trunc$loglik, vd$loglik)
  expect_eql(unname(dim(vd_trunc)), c(3, 1))
  expect_eql(vd_trunc$copula$pair_copulas, vd$copula$pair_copulas[1])
  expect_length(vd_trunc$copula$structure$struct_array[[1]], 2)
  expect_warning(expect_identical(vd_trunc, truncate_model(vd_trunc, 2)))
})

#' Plotting \code{vinecop_dist} and `vinecop` objects.
#'
#' There are two plotting generics for \code{vinecop_dist} objects.
#' \code{plot.vinecop_dist} plots one or all trees of a given R-vine copula
#' model. Edges can be labeled with information about the corresponding
#' pair-copula. \code{contour.vinecop_dist} produces a matrix of contour plots
#' (using \code{\link[rvinecopulib:plot.bicop]{plot.bicop}}).
#'
#' If you want the contour boxes to be perfect squares, the plot height should
#' be \code{1.25/length(tree)*(d - min(tree))} times the plot width.
#'
#' The `plot()` method returns an object that (among other things) contains the
#' `igraph` representation of the graph; see *Examples*.
#'
#' @method plot vinecop_dist
#'
#' @aliases contour.vinecop_dist plot.vinecop contour.vinecop
#'
#' @param x \code{vinecop_dist} object.
#' @param tree \code{"ALL"} or integer vector; specifies which trees are
#' plotted.
#' @param var_names integer; specifies how to make use of variable names:
#' * `"ignore"`` = variable names are ignored,
#' * `"use"`` = variable names are used to annotate vertices,
#' * `"legend"`` = uses numbers in plot and adds a legend for variable names,
#' * `"hide"`` = no numbers or names, just the node.
#' @param edge_labels character; options are:
#' * `"family"` = pair-copula family (see `[bicop_dist()]`),
#' * `"tau"`` = pair-copula Kendall's tau
#' * `"family_tau"`` = pair-copula family and Kendall's tau,
#' * `"pair"`` = the name of the involved variables.
#' @param cex.nums numeric; expansion factor for font of the numbers.
#' @param \dots Unused for \code{plot} and passed to
#' \code{\link[rvinecopulib:contour.bicop]{contour.bicop}} for \code{contour}.
#'
#' @author Thomas Nagler, Thibault Vatter
#'
#' @seealso \code{\link[rvinecopulib:vinecop_dist]{vinecop_dist}},
#' \code{\link[rvinecopulib:plot.bicop]{plot.bicop}}
#'
#' @keywords plot
#'
#' @examples
#' # set up vine copula model
#' u <- matrix(runif(20 * 10), 20, 10)
#' vc <- vinecop(u, family = "indep")
#'
#' # plot
#' plot(vc, tree = c(1, 2))
#' plot(vc, edge_labels = "pair")
#'
#' # extract igraph representation
#' plt <- plot(vc, edge_labels = "family_tau")
#' igr_obj <- get("g", plt$plot_env)[[1]]
#' igr_obj  # print object
#' igraph::E(igr_obj)$name  # extract edge labels
#'
#' # set up another vine copula model
#' pcs <- lapply(1:3, function(j) # pair-copulas in tree j
#'   lapply(runif(4 - j), function(cor) bicop_dist("gaussian", 0, cor)))
#' mat <- rvine_matrix_sim(4)
#' vc <- vinecop_dist(pcs, mat)
#'
#' # contour plot
#' contour(vc)
#' @export
plot.vinecop_dist <- function(x, tree = 1, var_names = "ignore",
                              edge_labels = NULL, ...) {
  if (!requireNamespace("ggraph", quietly = TRUE)) {
    stop("The 'ggraph' package must be installed to plot.")
  }
  if (!requireNamespace("grid", quietly = TRUE)) {
    stop("The 'grid' package must be installed to plot.")
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("The 'ggplot2' package must be installed to plot.")
  }
  if (!requireNamespace("igraph", quietly = TRUE)) {
    stop("The 'igraph' package must be installed to plot.")
  }

  d <- dim(x)[1]
  trunc_lvl <- dim(x)[2]

  ## sanity checks
  if (!identical(tree, "ALL") && any(tree > trunc_lvl)) {
    stop("Selected tree does not exist.")
  }
  if (any(tree == "ALL")) {
    if (d > 5) {
      warning(paste(
        "tree = 'ALL' is not recommended for d > 5 and",
        " it is set as c(1,2), please use tree = 1:d"
      ))
      tree <- c(1, 2)
    } else {
      tree <- seq_len(trunc_lvl)
    }
  }
  assert_that(in_set(var_names, c("ignore", "use", "legend", "hide")))
  if (!is.null(edge_labels)) {
    assert_that(in_set(edge_labels, c("pair", "tau", "family", "family_tau")))
  }

  ## set names if empty
  if (is.null(x$names)) {
    x$names <- as.character(1:d)
  }
  if (var_names %in% c("ignore", "legend")) {
    names <- x$names
    x$names <- as.character(1:d)
  }

  #### loop through the trees and create graph objects
  g <- lapply(tree, get_graph, vc = x, edge_labels = edge_labels)

  plots <- vector("list", length(tree))
  name <- NULL # for the CRAN check
  for (i in seq_along(tree)) {
    p <- ggraph::ggraph(g[[i]], "igraph",
      algorithm = "tree", circular = TRUE
    )
    if (!is.null(edge_labels)) {
      p <- p +
        ggraph::geom_edge_link(ggplot2::aes(label = name),
          colour = "#000000",
          angle_calc = "along",
          label_dodge = grid::unit(7, "points")
        )
    } else {
      p <- p + ggraph::geom_edge_link(colour = "#000000")
    }
    p <- p +
      ggraph::geom_node_point(col = "#56B4E9", size = 3) +
      ggplot2::theme_void() +
      ggplot2::labs(title = paste0("Tree ", tree[i]))
    if (var_names != "hide") {
      p <- p + ggraph::geom_node_text(ggplot2::aes(label = name),
        fontface = "bold",
        repel = TRUE
      )
    }
    if (var_names == "legend") {
      p <- p + ggplot2::labs(caption = paste(x$names, names,
        sep = " = ",
        collapse = ", "
      ))
    }
    plots[[i]] <- p +
      ggplot2::theme(plot.margin = ggplot2::margin(5, 5, 5, 5, "pt"))
  }

  if (length(tree) > 3) {
    invisible(multiplot(plotlist = plots, cols = 2))
  } else {
    return(invisible(multiplot(plotlist = plots)))
  }
}

#' @rdname plot.vinecop_dist
#' @export
plot.vinecop <- plot.vinecop_dist

## creates a graph object for a tree in a given vinecop_dist
get_graph <- function(tree, vc, edge_labels) {
  M <- get_matrix(vc)
  d <- dim(vc)[1]

  I <- matrix(0, d - tree + 1, d - tree + 1)

  ## extract node and edge labels as numbers
  if (tree > 1) {
    V <- t(sapply(seq.int(d - tree + 1), function(j)
      M[c(d - j + 1, (tree - 1):1), j]))
  } else {
    V <- matrix(diag(M[d:1, ]), ncol = 1)
  }
  E <- t(sapply(seq.int(d - tree), function(j)
    M[c(d - j + 1, tree:1), j]))

  ## build adjacency matrix by matching V and E
  for (i in 1:nrow(E)) {
    ind.i <- which(apply(V, 1, function(x) all(x %in% E[i, ])))
    I[ind.i[1], ind.i[2]] <- I[ind.i[2], ind.i[1]] <- 1
  }

  ## convert to variable names
  if (tree > 1) {
    colnames(I) <- rownames(I) <- sapply(1:(d - tree + 1),
      get_name,
      tree = tree - 1,
      vc = vc
    )
  } else {
    colnames(I) <- rownames(I) <- vc$names[diag(M[d:1, ])]
  }

  ## build graph
  g <- igraph::graph_from_adjacency_matrix(I, mode = "undirected")

  ## add edge labels
  if (!is.null(edge_labels)) {
    igraph::E(g)$name <- sapply(tree, set_edge_labels,
      vc = vc,
      edge_labels = edge_labels
    )
  }

  g
}

## finds appropriate edge labels for the plot
set_edge_labels <- function(tree, vc, edge_labels) {
  d <- dim(vc)[1]
  get_edge_label <- switch(edge_labels,
    family = get_family,
    tau = function(vc, tree, edge)
      round(get_ktau(vc, tree, edge),
        digits = 2
      ),
    family_tau = get_family_tau,
    pair = get_name
  )
  sapply(1:(d - tree), function(edge) get_edge_label(vc, tree, edge))
}


## get info for a pair-copula
get_name <- function(vc, tree, edge) {
  M <- get_matrix(vc)
  d <- nrow(M)
  # conditioned set
  bef <- paste0(vc$names[M[c(d - edge + 1, tree), edge]],
    collapse = ","
  )
  # conditioning set
  aft <- ifelse(tree > 1,
    paste0(vc$names[M[(tree - 1):1, edge]], collapse = ","),
    ""
  )
  # paste together
  sep <- ifelse(tree > 1, ";", "")
  paste(bef, aft, sep = sep, collapse = "")
}

get_family_tau <- function(vc, tree, edge) {
  paste0(
    get_family(vc, tree, edge), "(",
    round(get_ktau(vc, tree, edge), digits = 2), ")"
  )
}

#' @method contour vinecop_dist
#' @rdname plot.vinecop_dist
#' @importFrom graphics par plot.new plot.window abline polygon
#' @importFrom graphics strheight strwidth text
#' @export
contour.vinecop_dist <- function(x, tree = "ALL", cex.nums = 1, ...) {
  ## check input
  d <- dim(x)[1]
  trunc_lvl <- dim(x)[2]
  assert_that(is.number(cex.nums))
  assert_that((length(tree) == 1 && tree == "ALL") || all(tree <= trunc_lvl),
    msg = "Selected tree does not exist."
  )
  if (trunc_lvl == 0) {
    warning("Model contains no pair copulas (trunc_lvl = 0).")
    return(invisible(NULL))
  }

  if (any(tree == "ALL")) {
    tree <- seq_len(trunc_lvl)
  }

  n_tree <- length(tree)

  if (!is.null(list(...)$var_names)) {
    stop("Only contour plots allowed. Don't use the var_names argument!")
  }
  if (!is.null(list(...)$margins)) {
    margins <- list(...)$margins
    assert_that(in_set(margins, c("unif", "norm", "exp", "flexp")))
  } else {
    margins <- "norm"
  }
  if (is.null(list(...)$xlim) & is.null(list(...)$ylim)) {
    xylim <- switch(margins,
      "unif" = c(1e-2, 1 - 1e-2),
      "norm" = c(-3, 3),
      "exp" = c(0, 6),
      "flexp" = c(-6, 0)
    )
    xlim <- ylim <- xylim
  } else {
    xlim <- list(...)$xlim
    ylim <- list(...)$ylim
    xylim <- range(c(list(...)$xlim, list(...)$ylim))
  }

  if (is.null(x$names)) {
    x$names <- as.character(1:d)
  }

  ## set up for plotting windows (restore settings on exit)
  usr <- par(mfrow = c(n_tree, d - min(tree)), mar = rep(0, 4))
  on.exit(par(usr))

  # contours: adjust limits for headings
  offs <- 0.25
  mult <- 1.35
  ylim[2] <- ylim[2] + offs * diff(ylim)

  ## run through trees -----------------------------------------------
  # initialize check variables
  cnt <- 0
  k <- d
  maxnums <- get_name(x, max(tree, length(x$pair_copulas)), 1)
  for (i in rev(tree)) {
    for (j in 1:(d - min(tree))) {
      if (j <= d - i) {
        if (length(x$pair_copulas) >= i) {
          pcfit <- x$pair_copulas[[i]][[j]]
        } else {
          pcfit <- bicop_dist()
        }

        # set up list of contour arguments
        args <- list(
          x = pcfit,
          drawlabels = FALSE,
          xlab = "",
          ylab = "",
          xlim = xlim,
          ylim = ylim,
          xaxt = "n",
          yaxt = "n",
          add = TRUE
        )

        # create empty plot
        plot.new()
        plot.window(
          xlim = xlim, ylim = ylim,
          xaxs = "i", yaxs = "i"
        )

        # call contour.bicop with ... arguments
        do.call(contour, modifyList(args, list(...)))

        # draw area for headings
        abline(h = ylim[2] - diff(ylim) / mult * offs)
        polygon(
          x = c(
            xlim[1] - diff(xlim),
            xlim[1] - diff(xlim),
            xlim[2] + diff(xlim),
            xlim[2] + diff(xlim)
          ),
          y = c(
            ylim[2] + diff(ylim) / mult * offs,
            ylim[2] - diff(ylim) / mult * offs,
            ylim[2] - diff(ylim) / mult * offs,
            ylim[2] + diff(ylim) / mult * offs
          ),
          col = "grey"
        )

        # add separating lines
        abline(v = xlim)
        abline(h = ylim)

        # add pair-copula ID
        cx1 <- 0.75 * diff(xlim) / strwidth(maxnums)
        ty <- ylim[2] - diff(ylim) / mult * offs
        cx2 <- 0.75 * (ylim[2] - ty) / strheight(maxnums)
        cx <- min(cx1, cx2)
        text(
          x = sum(xlim) / 2,
          y = ty + 0.225 / cex.nums * (ylim[2] - ty),
          cex = cex.nums * cx,
          labels = get_name(x, i, j),
          pos = 3,
          offset = 0
        )
      } else {
        plot.new()
      }
    }
  }
}

#' @rdname plot.vinecop_dist
#' @export
contour.vinecop <- contour.vinecop_dist

#' R-vine structure
#'
#' R-vine structures are compressed representations encoding the tree structure
#' of the vine, i.e. the conditioned/conditioning variables of each edge. The
#' functions `[cvine_structure()]` or `[dvine_structure()]` give a simpler way
#' to construct C-vines (every tree is a star) and D-vines (every tree is a
#' path), respectively (see *Examples*).
#'
#' The R-vine structure is essentially a lower-triangular matrix/triangular
#' array, with a notation that differs from the one in the VineCopula package.
#' An example array is
#' ```
#' 4 4 4 4
#' 3 3 3
#' 2 2
#' 1
#' ```
#' which encodes the following pair-copulas:
#'
#' \tabular{lll}{
#' tree \tab  edge \tab pair-copulas   \cr
#' 0    \tab  0    \tab `(1, 4)`       \cr
#'      \tab  1    \tab `(2, 4)`       \cr
#'      \tab  2    \tab `(3, 4)`       \cr
#' 1    \tab  0    \tab `(1, 3; 4)`    \cr
#'      \tab  1    \tab `(2, 3; 4)`    \cr
#' 2    \tab  0    \tab `(1, 2; 3, 4)`
#' }
#'
#' An R-vine structure can be converted to an R-vine matrix using
#' [as_rvine_matrix()], which encodes the same model with a square matrix filled
#' with zeros. For instance, the matrix corresponding to the structure above is:
#' ```
#' 4 4 4 4
#' 3 3 3 0
#' 2 2 0 0
#' 1 0 0 0
#' ```
#' Similarly, an R-vine matrix can be converted to an R-vine structure using
#' [as_rvine_structure()].
#'
#' Denoting by `M[i, j]` the array entry in row `i` and column `j` (the
#' pair-copula index for edge `e` in tree `t` of a `d` dimensional vine is
#' `(M[d + 1 - e, e], M[t, e]; M[t - 1, e], ..., M[1, e])`. Less formally,
#'
#' 1. Start with the counter-diagonal element of column `e` (first conditioned
#' variable).
#'
#' 2. Jump up to the element in row `t` (second conditioned variable).
#'
#' 3. Gather all entries further up in column `e` (conditioning set).
#'
#' Internally, the diagonal is stored separately from the off-diagonal elements,
#' which are stored as a triangular array. For instance, the off-diagonal
#' elements off the structure above are stored as
#' ```
#' 4 4 4
#' 3 3
#' 2
#' ```
#' for the structure above. The reason is that it allows for parsimonious
#' representations of truncated models. For instance, the 2-truncated model is
#' represented by the same diagonal and the following truncated triangular
#' array:
#' ```
#' 4 4 4
#' 3 3
#' ```
#'
#' A valid R-vine structure or matrix must satisfy several conditions which are
#' checked when [rvine_structure()], [rvine_matrix()], or some coercion methods
#' (see [as_rvine_structure()] and `as_rvine_matrix(`) are called:
#'
#' 1. It can only contain numbers between 1 and d (and additionally zeros for
#' R-vine matrices).
#'
#' 2. The anti-diagonal must contain the numbers 1, ..., d.
#'
#' 3. The anti-diagonal entry of a column must not be contained in any column
#' further to the right.
#'
#' 4. The entries of a column must be contained in all columns to the left.
#'
#' 5. The proximity condition must hold: For all t = 1, ..., d - 2 and e = 1,
#' ..., d - t there must exist an index j > d, such that
#' `(M[t, e], {M[1, e], ..., M[t - 1, e]})` equals either
#' `(M[d + 1 - j, j], {M[1, j], ..., M[t - 1, j]})` or
#' `(M[t - 1, j], {M[d + 1 - j, j], M[1, j], ..., M[t - 2, j]})`.
#'
#'
#' Condition 5 already implies conditions 2-4, but is more difficult to check by
#' hand.
#'
#' @param order a vector of positive integers.
#' @param struct_array a list of vectors of positive integers. The vectors
#'   represent rows of the r-rvine structure and the number of elements have to
#'   be compatible with the `order` vector. If empty, the model is 0-truncated.
#' @param is_natural_order whether `struct_array` is assumed to be provided in
#'   natural order already (a structure is in natural order if the anti-
#'   diagonal is 1, .., d from bottom left to top right).
#'
#' @return Either an `rvine_structure` or an `rvine_matrix`.
#' @export
#' @seealso [as_rvine_structure()], [as_rvine_matrix()],
#'   [plot.rvine_structure()], [plot.rvine_matrix()],
#'   [rvine_structure_sim()], [rvine_matrix_sim()]
#' @examples
#'
#' # R-vine structures can be constructed from the order vector and struct_array
#' rvine_structure(order = 1:4, struct_array = list(
#'   c(4, 4, 4),
#'   c(3, 3),
#'   2
#' ))
#'
#' # R-vine matrices can be constructed from standard matrices
#' mat <- matrix(c(4, 3, 2, 1, 4, 3, 2, 0, 4, 3, 0, 0, 4, 0, 0, 0), 4, 4)
#' rvine_matrix(mat)
#'
#' # coerce to R-vine structure
#' str(as_rvine_structure(mat))
#'
#' # truncate and construct the R-vine matrix
#' mat[3, 1] <- 0
#' rvine_matrix(mat)
#'
#' # or use directly the R-vine structure constructor
#' rvine_structure(order = 1:4, struct_array = list(
#'   c(4, 4, 4),
#'   c(3, 3)
#' ))
#'
#' # throws an error
#' mat[3, 1] <- 5
#' try(rvine_matrix(mat))
#'
#' # C-vine structure
#' cvine <- cvine_structure(1:5)
#' cvine
#' plot(cvine)
#'
#' # D-vine structure
#' dvine <- dvine_structure(c(1, 4, 2, 3, 5))
#' dvine
#' plot(dvine)
#'
#' @name rvine_structure
#' @aliases rvine_matrix is.rvine_structure is.rvine_matrix dvine_structure
#'   cvine_structure
rvine_structure <- function(order, struct_array = list(), is_natural_order = FALSE) {

  # sanity checks and extract dimension/trunc_lvl
  assert_that(is.vector(order) && all(sapply(order, is.count)),
    msg = "Order should be a vector of positive integers."
  )
  assert_that(is.vector(struct_array) || is.list(struct_array))
  assert_that(all(rapply(struct_array, function(i) sapply(i, is.count))),
    msg = "All elements of struct_array should be positive integers."
  )
  d <- length(order)
  assert_that(length(struct_array) < length(order),
    msg = paste("The length of struct array is incompatible with order,",
                "it should be in {1, ..., length(order) - 1}")
  )

  trunc_lvl <- length(struct_array)
  lens <- sapply(seq_len(trunc_lvl), function(j) length(struct_array[[j]]))
  assert_that(all(lens == seq(d - 1, d - trunc_lvl)),
    msg = paste(
      "The number of elements in struct_array is incompatible with order.",
      "The first entry of struct_array should be a vectors of size" ,
      "length(order) - 1, the second of size length(order) - 2), etc.")
  )

  # create and check output
  output <- rvine_structure_cpp(
    list(
      order = order,
      struct_array = struct_array,
      d = d,
      trunc_lvl = trunc_lvl
    ),
    TRUE, is_natural_order
  )
  output <- structure(output, class = c("rvine_structure", "list"))

  # return output
  output
}

#' Plotting R-vine structures
#'
#' Plot one or all trees of an R-vine structure.
#'
#' @param x an `rvine_structure` or `rvine_matrix` object.
#' @param ... passed to `plot.vinecop_dist()`.
#' @aliases plot.rvine_matrix
#' @export
#' @examples
#' plot(cvine_structure(1:5))
#' plot(rvine_structure_sim(5))
plot.rvine_structure <- function(x, ...) {
  d <- dim(x)[1]
  trunc_lvl <- dim(x)[2]
  pcs <- lapply(seq_len(min(d - 1, trunc_lvl)),
                function(i) lapply(seq_len(d - i), function(j) bicop_dist()))
  plot.vinecop_dist(vinecop_dist(pcs, x), ...)
}

#' @rdname plot.rvine_structure
#' @examples
#' mat <- rbind(c(1, 1, 1), c(2, 2, 0), c(3, 0, 0))
#' plot(rvine_matrix(mat))
#' plot(rvine_matrix_sim(5))
#' @export
plot.rvine_matrix <- function(x, ...) {
  plot(as_rvine_structure(x), ...)
}

#' @rdname rvine_structure
#' @param trunc_lvl the truncation level
#' @export
cvine_structure <- function(order, trunc_lvl = Inf) {
  if (is.count(order))
    order <- seq_len(order)
  assert_that(is.vector(order) && all(sapply(order, is.count)),
              msg = "Order should be a vector of positive integers.")
  assert_that(is.scalar(trunc_lvl) & is.number(trunc_lvl))

  d <- length(order)
  trunc_lvl <- min(trunc_lvl, d - 1)
  vars <- rev(seq_len(d))
  struct_array <- lapply(seq(1, trunc_lvl), function(i) rep(vars[i], d - i))

  output <- rvine_structure_cpp(
    list(
      order = order,
      struct_array = struct_array,
      d = d,
      trunc_lvl = trunc_lvl
    ),
    FALSE, TRUE
  )
  structure(output, class = c("rvine_structure", "list"))
}

#' @rdname rvine_structure
#' @export
dvine_structure <- function(order, trunc_lvl = Inf) {
  if (is.count(order))
    order <- seq_len(order)
  assert_that(is.vector(order) && all(sapply(order, is.count)),
              msg = "Order should be a vector of positive integers.")
  assert_that(is.scalar(trunc_lvl) & is.number(trunc_lvl))

  d <- length(order)
  trunc_lvl <- min(trunc_lvl, d - 1)
  vars <- rev(seq_len(d))
  struct_array <- lapply(seq(1, trunc_lvl), function(i) seq(i + 1, d))

  output <- rvine_structure_cpp(
    list(
      order = order,
      struct_array = struct_array,
      d = d,
      trunc_lvl = trunc_lvl
    ),
    FALSE, TRUE
  )
  structure(output, class = c("rvine_structure", "list"))
}

#' @param matrix an R-vine matrix, see *Details*.
#' @rdname rvine_structure
#' @export
rvine_matrix <- function(matrix) {
  validate_rvine_matrix(matrix)
  rvine_matrix_nocheck(matrix)
}

rvine_matrix_nocheck <- function(matrix) {
  d <- ncol(matrix)
  class(matrix) <- c("rvine_matrix", class(matrix))
  attr(matrix, "d") <- d
  attr(matrix, "trunc_lvl") <- ifelse(any(matrix[, 1] == 0),
    min(which(matrix[, 1] == 0)) - 1,
    d - 1
  )
  matrix
}

validate_rvine_structure <- function(structure, is_natural_order = TRUE) {
  assert_that(is.rvine_structure(structure))
  assert_that(is.flag(is_natural_order))
  rvine_structure_check_cpp(structure, is_natural_order)
}

validate_rvine_matrix <- function(matrix) {
  assert_that(is.rvine_matrix(matrix) ||
    (is.matrix(matrix) && is.numeric(matrix)))
  rvine_matrix_check_cpp(matrix)
}

#' @export
dim.rvine_matrix <- function(x) {
  output <- c(attr(x, "d"), attr(x, "trunc_lvl"))
  names(output) <- c("dim", "trunc_lvl")
  output
}

#' @export
#' @importFrom utils write.table
print.rvine_matrix <- function(x, ..., zero.print = " ", n = 10,
                               structure = FALSE) {
  if (structure) {
    cat(dim(x)[1], "-dimensional R-vine structure ('rvine_structure')", sep = "")
  } else {
    cat(dim(x)[1], "-dimensional R-vine matrix ('rvine_matrix')", sep = "")
  }
  print_truncation_info(x)
  write.table(format(x, zero.print = zero.print),
    quote = FALSE,
    row.names = FALSE, col.names = FALSE
  )
  invisible(x)
}

#' @export
print.rvine_structure <- function(x, ...) {
  print(as_rvine_matrix(x), structure = TRUE)
}

#' @export
dim.rvine_structure <- function(x) {
  output <- c(x$d, x$trunc_lvl)
  names(output) <- c("dim", "trunc_lvl")
  output
}

#' @export
is.rvine_structure <- function(structure) {
  inherits(structure, "rvine_structure")
}

#' @export
is.rvine_matrix <- function(matrix) {
  inherits(matrix, "rvine_matrix")
}


#' Simulate R-vine structures
#'
#' Simulates from a uniform distribution over all R-vine structures on d
#' variables. `rvine_structure_sim()` returns an [rvine_structure()] object,
#' `rvine_matrix_sim()` an [rvine_matrix()].
#'
#' @aliases rvine_matrix_sim
#'
#' @param d the number of variables
#' @param natural_order boolean; whether the structures should be in natural
#'   order (counter-diagonal is `1:d`).
#'
#' @seealso [rvine_structure()], [rvine_matrix()],
#'    [plot.rvine_structure()], [plot.rvine_matrix()]
#' @export
#' @examples
#' rvine_structure_sim(10)
#'
#' rvine_structure_sim(10, natural_order = TRUE)  # counter-diagonal is 1:d
#'
#' rvine_matrix_sim(10)
rvine_structure_sim <- function(d, natural_order = FALSE) {
  assert_that(is.count(d), is.flag(natural_order))
  rvine_structure_sim_cpp(d, natural_order, get_seeds())
}

#' @rdname rvine_structure_sim
#' @export
rvine_matrix_sim <- function(d, natural_order = FALSE) {
  as_rvine_matrix(rvine_structure_sim(d, natural_order))
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Class 'rvine_structure'")

mat <- matrix(c(4, 3, 2, 1, 4, 3, 2, 0, 4, 3, 0, 0, 4, 0, 0, 0), 4, 4)
mylist <- list(order = 1:4, struct_array = list(c(4, 4, 4), c(3, 3), 2))

test_that("constructor and as/is generics work", {
  expect_silent(rvs <- as_rvine_structure(mylist)) ## calls the constructor
  expect_silent(rvm <- as_rvine_matrix(mylist)) ## true coercion
  expect_silent(as_rvine_structure(mat)) ## calls the constructors
  expect_silent(as_rvine_matrix(mat)) ## true coercion
  expect_silent(as_rvine_matrix(as_rvine_matrix(mat)))
  expect_true(is.rvine_structure(rvs))
  expect_true(is.rvine_matrix(rvm))
  mat[1, 1] <- 0
  expect_silent(as_rvine_matrix(mat, validate = FALSE))
})

test_that("print/dim generics work", {
  rvs <- as_rvine_structure(mylist)
  rvm <- as_rvine_matrix(mat)
  expect_output(print(rvs))
  expect_output(print(rvm))
  expect_equiv(dim(rvs), c(4, 3))
  expect_equiv(dim(rvm), c(4, 3))
})

test_that("C- and D-vine structures work", {
  cv <- cvine_structure(1:6, trunc_lvl = 4)
  expect_equiv(dim(cv), c(6, 4))
  expect_equiv(cv$order, 1:6)
  expect_identical(cv, cvine_structure(6, 4))

  dv <- dvine_structure(1:6, trunc_lvl = 4)
  expect_equiv(dim(dv), c(6, 4))
  expect_equiv(dv$order, 1:6)
  expect_identical(dv, dvine_structure(6, 4))
})

test_that("plot functions work", {
  struct <- as_rvine_structure(mat)
  mat <- rvine_matrix(mat)
  vc <- vinecop_dist(list(replicate(3, bicop_dist(), simplify = FALSE)), struct)
  p <- plot(vc)
  ps <- plot(struct)
  pm <- plot(mat)
  p$plot_env <- NULL
  ps$plot_env <- NULL
  pm$plot_env <- NULL
  expect_equiv(p, ps)
  expect_equiv(p, pm)
})

test_that("d = 1 works", {
  struct <- rvine_structure(1)
  expect_length(struct$struct_array, 0)
  expect_length(struct$order, 1)

  mat <- as_rvine_matrix(struct)
  expect_eql(unname(dim(mat)), c(1, 0))

  expect_output(print(struct))
  expect_error(plot(struct))

  expect_silent(dvine_structure(1))
  expect_silent(cvine_structure(1))
})

#' Vine copula distributions
#'
#' Density, distribution function and random generation for the vine copula
#' distribution.
#'
#' @name vinecop_distributions
#' @aliases dvinecop pvinecop rvinecop dvinecop_dist pvinecop_dist rvinecop_dist
#' @param u matrix of evaluation points; must contain at least d columns, where
#'   d is the number of variables in the vine. More columns are required for
#'   discrete models, see *Details*.
#' @param vinecop an object of class `"vinecop_dist"`.
#' @param cores number of cores to use; if larger than one, computations are
#'   done in parallel on `cores` batches .
#' @details See [vinecop()] for the estimation and construction of vine copula
#' models.
#'
#' The copula density is defined as joint density divided by marginal
#' densities, irrespective of variable types.
#'
#' ## Discrete variables
#'
#' When at least one variable is discrete, two types of
#' "observations" are required in `u`: the first \eqn{n \; x \; d} block
#' contains realizations of \eqn{F_{X_j}(X_j)}. The second \eqn{n \; x \; d}
#' block contains realizations of \eqn{F_{X_j}(X_j^-)}. The minus indicates a
#' left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
#' \eqn{F_{X_j}(X_j^-) = F_{X_j}(X_j - 1)}. For continuous variables the left
#' limit and the cdf itself coincide. Respective columns can be omitted in the
#' second block.
#'
#' @return
#' `dvinecop()` gives the density, `pvinecop()` gives the distribution function,
#' and `rvinecop()` generates random deviates.
#'
#' The length of the result is determined by `n` for `rvinecop()`, and
#' the number of rows in `u` for the other functions.
#'
#' The `vinecop` object is recycled to the length of the
#' result.
#'
#' @seealso [vinecop_dist()], [vinecop()], [plot.vinecop()], [contour.vinecop()]
#'
#' @examples
#' ## simulate dummy data
#' x <- rnorm(30) * matrix(1, 30, 5) + 0.5 * matrix(rnorm(30 * 5), 30, 5)
#' u <- pseudo_obs(x)
#'
#' ## fit a model
#' vc <- vinecop(u, family = "clayton")
#'
#' # simulate from the model
#' u <- rvinecop(100, vc)
#' pairs(u)
#'
#' # evaluate the density and cdf
#' dvinecop(u[1, ], vc)
#' pvinecop(u[1, ], vc)
#'
#' ## Discrete models
#' vc$var_types <- rep("d", 5)  # convert model to discrete
#'
#' # with discrete data we need two types of observations (see Details)
#' x <- qpois(u, 1)  # transform to Poisson margins
#' u_disc <- cbind(ppois(x, 1), ppois(x - 1, 1))
#'
#' dvinecop(u_disc[1:5, ], vc)
#' pvinecop(u_disc[1:5, ], vc)
#'
#' # simulated data always has uniform margins
#' pairs(rvinecop(200, vc))
#' @rdname vinecop_methods
#' @export
dvinecop <- function(u, vinecop, cores = 1) {
  assert_that(inherits(vinecop, "vinecop_dist"))
  u <- if_vec_to_matrix(u, dim(vinecop)[1] == 1)
  vinecop_pdf_cpp(u, vinecop, cores)
}

#' @rdname vinecop_methods
#' @param n_mc number of samples used for quasi Monte Carlo integration.
#' @importFrom assertthat is.count
#' @export
pvinecop <- function(u, vinecop, n_mc = 10^4, cores = 1) {
  assert_that(
    inherits(vinecop, "vinecop_dist"),
    is.number(n_mc), is.count(cores)
  )
  u <- if_vec_to_matrix(u, dim(vinecop)[1] == 1)
  vinecop_cdf_cpp(as.matrix(u), vinecop, n_mc, cores, get_seeds())
}

#' @rdname vinecop_methods
#' @param n number of observations.
#' @param qrng if `TRUE`, generates quasi-random numbers using the multivariate
#' Generalized Halton sequence up to dimension 300 and the Generalized Sobol
#' sequence in higher dimensions (default `qrng = FALSE`).
#' @export
rvinecop <- function(n, vinecop, qrng = FALSE, cores = 1) {
  assert_that(
    is.number(n),
    inherits(vinecop, "vinecop_dist"),
    is.flag(qrng),
    is.number(cores)
  )

  U <- vinecop_sim_cpp(vinecop, n, qrng, cores, get_seeds())
  if (!is.null(vinecop$names)) {
    colnames(U) <- vinecop$names
  }

  U
}

#' @export
print.vinecop_dist <- function(x, ...) {
  cat(dim(x)[1], "-dimensional vine copula model ('vinecop_dist')", sep = "")
  print_truncation_info(x)
  invisible(x)
}

#' @importFrom utils capture.output
#' @export
summary.vinecop_dist <- function(object,
                                 trees = seq_len(dim(object)["trunc_lvl"]),
                                 ...) {
  mat <- as_rvine_matrix(get_structure(object))
  d <- dim(object)[1]
  trees <- intersect(trees, seq_len(dim(object)["trunc_lvl"]))
  n_pcs <- length(unlist(object$pair_copulas[trees], recursive = FALSE))
  mdf <- as.data.frame(matrix(NA, n_pcs, 10))
  names(mdf) <- c(
    "tree", "edge",
    "conditioned", "conditioning", "var_types",
    "family", "rotation", "parameters", "df", "tau"
  )
  k <- 1
  for (t in trees) {
    for (e in seq_len(d - t)) {
      mdf$tree[k] <- t
      mdf$edge[k] <- e
      mdf$conditioned[k] <- list(c(mat[d - e + 1, e], mat[t, e]))
      mdf$conditioning[k] <- list(mat[rev(seq_len(t - 1)), e])
      pc <- object$pair_copulas[[t]][[e]]
      mdf$var_types[k] <- paste(pc$var_types, collapse = ",")
      mdf$family[k] <- pc$family
      mdf$rotation[k] <- pc$rotation
      mdf$parameters[k] <- list(pc$parameters)
      if (pc$family %in% setdiff(family_set_nonparametric, "indep"))
        mdf$parameters[k] <- list("[30x30 grid]")
      mdf$df[k] <- pc$npars
      mdf$tau[k] <- par_to_ktau(pc)
      k <- k + 1
    }
  }
  class(mdf) <- c("summary_df", class(mdf))
  mdf
}

#' Predictions and fitted values for a vine copula model
#'
#' Predictions of the density and distribution function
#' for a vine copula model.
#'
#' @name vinecop_predict_and_fitted
#' @aliases fitted.vinecop predict.vinecop
#' @param object a `vinecop` object.
#' @param newdata points where the fit shall be evaluated.
#' @param what what to predict, either `"pdf"` or `"cdf"`.
#' @param n_mc number of samples used for quasi Monte Carlo integration when
#'    `what = "cdf"`.
#' @param cores number of cores to use; if larger than one, computations are
#'   done in parallel on `cores` batches.
#' @param ... unused.
#'
#' @details `fitted()` can only be called if the model was fit with the
#'    `keep_data = TRUE` option.
#'
#' ## Discrete variables
#'
#' When at least one variable is discrete, two types of
#' "observations" are required in `newdata`: the first \eqn{n \; x \; d} block
#' contains realizations of \eqn{F_{X_j}(X_j)}. The second \eqn{n \; x \; d}
#' block contains realizations of \eqn{F_{X_j}(X_j^-)}. The minus indicates a
#' left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
#' \eqn{F_{X_j}(X_j^-) = F_{X_j}(X_j - 1)}. For continuous variables the left
#' limit and the cdf itself coincide. Respective columns can be omitted in the
#' second block.
#'
#' @return
#' `fitted()` and `predict()` have return values similar to [dvinecop()]
#' and [pvinecop()].
#' @export
#' @rdname predict_vinecop
#' @examples
#' u <- sapply(1:5, function(i) runif(50))
#' fit <- vinecop(u, family = "par", keep_data = TRUE)
#' all.equal(predict(fit, u), fitted(fit), check.environment = FALSE)
predict.vinecop <- function(object, newdata, what = "pdf", n_mc = 10^4,
                            cores = 1, ...) {
  assert_that(
    in_set(what, c("pdf", "cdf")),
    is.number(n_mc),
    is.number(cores), cores > 0
  )
  newdata <- if_vec_to_matrix(newdata, dim(object)[1] == 1)
  switch(
    what,
    "pdf" = vinecop_pdf_cpp(newdata, object, cores),
    "cdf" = vinecop_cdf_cpp(newdata, object, n_mc, cores, get_seeds())
  )
}

#' @rdname predict_vinecop
#' @export
fitted.vinecop <- function(object, what = "pdf", n_mc = 10^4, cores = 1, ...) {
  if (is.null(object$data)) {
    stop("data have not been stored, use keep_data = TRUE when fitting.")
  }
  assert_that(
    in_set(what, c("pdf", "cdf")),
    is.number(n_mc),
    is.number(cores), cores > 0
  )
  switch(
    what,
    "pdf" = vinecop_pdf_cpp(object$data, object, cores),
    "cdf" = vinecop_cdf_cpp(object$data, object, n_mc, cores, get_seeds())
  )
}

#' @export
logLik.vinecop <- function(object, ...) {
  structure(object$loglik, "df" = object$npars)
}

#' Modified vine copula Bayesian information criterion (mBICv)
#'
#' Calculates the modified vine copula Bayesian information criterion.
#'
#' The modified vine copula Bayesian information criterion (mBICv) is defined as
#'
#' \deqn{BIC = -2 loglik +  \nu log(n) - 2 \sum_{t=1}^{d - 1} (q_t log(\psi_0^t)
#' - (d - t - q_t) log(1 - \psi_0^t)) }
#'
#' where \eqn{\mathrm{loglik}} is the log-likelihood and \eqn{\nu} is the
#' (effective) number of parameters of the model, \eqn{t} is the tree level
#' \eqn{\psi_0} is the prior probability of having a non-independence copula and
#' \eqn{q_t} is the number of non-independence copulas in tree \eqn{t}. The
#' mBICv is a consistent model selection criterion for parametric sparse vine
#' copula models.
#'
#' @param object a fitted `vinecop` object.
#' @param psi0 baseline prior probability of a non-independence copula.
#' @param newdata optional; a new data set.
#'
#' @references Nagler, T., Bumann, C., Czado, C. (2019). Model selection for
#'   sparse high-dimensional vine copulas with application to portfolio risk.
#'   *Journal of Multivariate Analysis, in press*
#'   (\url{https://arxiv.org/pdf/1801.09739.pdf})
#'
#' @export mBICV
#' @examples
#' u <- sapply(1:5, function(i) runif(50))
#' fit <- vinecop(u, family = "par", keep_data = TRUE)
#' mBICV(fit, 0.9) # with a 0.9 prior probability of a non-independence copula
#' mBICV(fit, 0.1) # with a 0.1 prior probability of a non-independence copula
mBICV <- function(object, psi0 = 0.9, newdata = NULL) {
  assert_that(inherits(object, "vinecop_dist"), is.number(psi0))
  ll <- ifelse(is.null(newdata),
    object$loglik,
    sum(log(dvinecop(newdata, object)))
  )
  - 2 * ll + compute_mBICV_penalty(object, psi0)
}

compute_mBICV_penalty <- function(object, psi0) {
  d <- dim(object)[1]
  smr <- summary(object)
  q_m <- tapply(smr$family, smr$tree, function(x) sum(x == "indep"))
  q_m <- c(q_m, rep(0, d - 1 - length(q_m)))
  m_seq <- seq_len(d - 1)
  pen <- object$npars * log(object$nobs)
  pen - 2 * sum(
    q_m * log(psi0^m_seq) + (d - seq_len(d - 1) - q_m) * log(1 - psi0^m_seq)
  )
}

#' @export
print.vinecop <- function(x, ...) {
  cat(dim(x)[1], "-dimensional vine copula fit ('vinecop')", sep = "")
  print_truncation_info(x)
  print_fit_info(x)
  invisible(x)
}


#' @export
summary.vinecop <- function(object, trees = seq_len(dim(object)["trunc_lvl"]), ...) {
  trees <- intersect(trees, seq_len(dim(object)["trunc_lvl"]))
  mdf <- summary.vinecop_dist(object, trees)
  d <- dim(object)[1]
  k <- 1
  for (t in trees) {
    for (e in seq_len(d - t)) {
      mdf$loglik[k] <- object$pair_copulas[[t]][[e]]$loglik
      k <- k + 1
    }
  }

  mdf
}

#' @export
dim.vinecop_dist <- function(x) {
  dim(x$structure)
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Fitting 'vinecop' models")

set.seed(5)
u <- sapply(1:5, function(i) runif(30))
fit <- vinecop(u, family = "nonpar")
fit_with_data <- vinecop(u, family = "nonpar", keep_data = TRUE)

test_that("returns proper 'vinecop' object", {
  expect_s3_class(fit, "vinecop")
  expect_s3_class(fit, "vinecop_dist")
  expect_identical(
    names(fit),
    c("pair_copulas", "structure", "var_types",
      "npars", "loglik", "threshold", "controls", "nobs")
  )
  expect_identical(
    names(fit_with_data),
    c("pair_copulas", "structure", "var_types",
      "npars", "loglik", "threshold", "data", "controls", "nobs")
  )

  colnames(u) <- paste(seq_len(ncol(u)))
  expect_identical(
    names(vinecop(u, family = "indep")),
    c("pair_copulas", "structure", "var_types",
      "npars", "loglik", "threshold", "names", "controls", "nobs")
  )
})

test_that("works with structure", {
  u <- sapply(1:2, function(i) runif(30))
  expect_silent(fit <- vinecop(u, structure = matrix(c(1:2, 1:0), 2, 2)))
})

if (Sys.info()["sysname"] != "SunOS") {
  test_that("runs in parallel", {
    expect_silent(fit <- vinecop(u, cores = 2))
  })
}

test_that("S3 generics work", {
  expect_eql(predict(fit, u), fitted(fit_with_data))
  expect_eql(
    predict(fit, u, what = "cdf"),
    fitted(fit_with_data, what = "cdf"),
    tolerance = 0.01
  )
  expect_error(predict(fit, u, what = "hfunc1"))
  fit$data <- NULL
  expect_error(fitted(fit))
  expect_length(attr(logLik(fit), "df"), 1)
})

test_that("print/summary generics work", {
  expect_output(print(fit))
  expect_s3_class(s <- summary(fit), "summary_df")
  expect_is(s, "data.frame")
})

test_that("truncation works", {
  fit_truncated <- truncate_model(fit, trunc_lvl = 1)
  expect_silent(dvinecop(u, fit_truncated))
  expect_silent(rvinecop(50, fit_truncated))

  fit_truncated <- vinecop(u, family = "par", trunc_lvl = 1)
  expect_silent(dvinecop(u, fit_truncated))
  expect_silent(rvinecop(50, fit_truncated))
})

test_that("partial selection works", {
  fit_partial <- vinecop(u[, sample(1:5)],
                         structure = truncate_model(fit$structure, 1),
                         trunc_lvl = 3)
  expect_eql(unname(dim(fit_partial)[2]), 3)

  m_old <- as_rvine_matrix(fit$structure)
  m_new <- as_rvine_matrix(fit_partial$structure)
  tree1_old_edges <- c(paste(diag(m_old[5:2, ]), m_old[1, -5]),
                       paste(m_old[1, -5], diag(m_old[5:2, ])))
  expect_true(all(paste(diag(m_new[5:2, ]), m_new[1, -5]) %in% tree1_old_edges))
})

test_that("d = 1 works", {
  vc <- vinecop(runif(20), structure = rvine_structure(1))
  vc2 <- vinecop(runif(20))
  expect_identical(vc, vc2)

  expect_eql(AIC(vc), 0)
  expect_eql(mBICV(vc), 0)
  expect_eql(dim(summary(vc))[1], 0)
})

#' Plotting tools for `bicop_dist` and `bicop` objects
#'
#' There are several options for plotting bicop_dist objects. The density of a
#' bivariate copula density can be visualized as surface/perspective or contour
#' plot. Optionally, the density can be coupled with standard normal margins
#' (default for contour plots).
#'
#' @method plot bicop_dist
#'
#' @aliases plot.bicop_dist contour.bicop_dist plot.bicop contour.bicop
#' @param x \code{bicop_dist object.}
#' @param type plot type; either \code{"surface"} or \code{"contour"}.
#' @param margins options are: \code{"unif"} for the original copula density,
#'   \code{"norm"} for the transformed density with standard normal margins,
#'   \code{"exp"} with standard exponential margins, and  \code{"flexp"} with
#'   flipped exponential margins. Default is \code{"norm"} for \code{type =
#'   "contour"}, and \code{"unif"} for \code{type = "surface"}.
#' @param size integer; the plot is based on values on a `size x size` grid,
#'   default is 100.
#' @param \dots optional arguments passed to \code{\link{contour}} or
#'   \code{\link{wireframe}}.
#' @seealso [bicop_dist()], [graphics::contour()], [wireframe()]
#' @keywords plot
#' @examples
#'
#' ## construct bicop_dist object for a student t copula
#' obj <- bicop_dist(family = "t", rotation = 0, parameters = c(0.7, 4))
#'
#' ## plots
#' plot(obj) # surface plot of copula density
#' contour(obj) # contour plot with standard normal margins
#' contour(obj, margins = "unif") # contour plot of copula density
#' @importFrom graphics contour plot
#' @importFrom grDevices rgb col2rgb colorRampPalette
#' @importFrom lattice wireframe
#' @importFrom stats pnorm qnorm dnorm pexp qexp dexp
#' @importFrom utils modifyList
#' @export
plot.bicop_dist <- function(x, type = "surface", margins, size, ...) {
  ## partial matching and sanity check for type
  allowed_types <- c("contour", "surface")
  type <- allowed_types[pmatch(type, allowed_types)]
  assert_that(in_set(type, allowed_types))

  ## choose margins if missing, else partial matching and sanity check
  if (missing(margins)) {
    margins <- switch(type, "contour" = "norm", "surface" = "unif")
  } else {
    assert_that(is.string(margins))
    allowed_margins <- c("norm", "unif", "exp", "flexp")
    margins <- allowed_margins[pmatch(margins, allowed_margins)]
    assert_that(in_set(margins, allowed_margins))
  }

  ## choose size if missing and sanity check
  if (missing(size)) {
    size <- switch(type, "contour" = 100L, "surface" = 40L)
  }
  assert_that(is.number(size))
  size <- round(size)
  if (size < 5) {
    warning("size too small, set to 5")
    size <- 5
  }

  if (is.null(list(...)$xlim) & is.null(list(...)$ylim)) {
    xylim <- switch(margins,
      "unif" = c(1e-2, 1 - 1e-2),
      "norm" = c(-3, 3),
      "exp" = c(0, 6),
      "flexp" = c(-6, 0)
    )
  } else {
    xylim <- range(c(list(...)$xlim, list(...)$ylim))
  }

  ## prepare for plotting with selected margins
  if (margins == "unif") {
    points <- switch(type,
      "contour" = seq(1e-5, 1 - 1e-5, length.out = size),
      "surface" = 1:size / (size + 1)
    )
    g <- as.matrix(expand.grid(points, points))
    points <- g[1L:size, 1L]
    adj <- 1
    gu <- g[, 1L]
    gv <- g[, 2L]
    levels <- c(0.2, 0.6, 1, 1.5, 2, 3, 5, 10, 20)
    xlim <- ylim <- xylim
    at <- c(seq(0, 3, length.out = 50), seq(5, 100, length.out = 50))
  } else if (margins == "norm") {
    points <- pnorm(seq(xylim[1L], xylim[2L], length.out = size))
    g <- as.matrix(expand.grid(points, points))
    points <- qnorm(g[1L:size, 1L])
    adj <- tcrossprod(dnorm(points))
    levels <- c(0.01, 0.025, 0.05, 0.1, 0.15, 0.2, 0.3, 0.4, 0.5)
    gu <- qnorm(g[, 1L])
    gv <- qnorm(g[, 2L])
    xlim <- ylim <- xylim
    # at <- c(seq(0, 0.3, length.out = 50), seq(0.3, 100, length.out = 50))
  } else if (margins == "exp") {
    ll <- ifelse(type == "contour", 1e-2, 1e-1)
    points <- pexp(seq(ll, xylim[2L], length.out = size))
    g <- as.matrix(expand.grid(points, points))
    points <- qexp(g[1L:size, 1L])
    adj <- tcrossprod(dexp(points))
    levels <- c(0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.3, 0.5)
    gu <- qexp(g[, 1L])
    gv <- qexp(g[, 2L])
    xlim <- ylim <- xylim
    # at <- c(0, exp(seq(-10, 0, length.out = 79)), seq(1, 10, length.out = 20))
  } else if (margins == "flexp") {
    ll <- ifelse(type == "contour", 1e-2, 1e-1)
    points <- pexp(-seq(xylim[1L], -ll, length.out = size))
    g <- as.matrix(expand.grid(points, points))
    points <- -qexp(g[1L:size, 1L])
    adj <- tcrossprod(dexp(qexp(g[1L:size, 1L])))
    levels <- c(0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.3, 0.5)
    gu <- -qexp(g[, 1L])
    gv <- -qexp(g[, 2L])
    g <- 1 - g
    xlim <- ylim <- xylim
    # at <- c(0, exp(seq(-10, 0, length.out = 79)), seq(1, 10, length.out = 20))
  }

  ## evaluate on grid
  vt <- x$var_types
  x$var_types <- c("c", "c")
  vals <- dbicop(g, x)
  x$var_types <- vt
  cop <- matrix(vals, size, size)

  ## adjust for margins
  dens <- cop * adj
  if (length(unique(c(dens))) == 1) {
    dens[1] <- 1.000001 * dens[1]
  }

  if (margins %in% c("exp", "flexp")) {
    dens <- pmin(dens, 6)
  }

  jet.colors <- colorRampPalette(c(
    "#00007F", "blue", "#007FFF", "cyan",
    "#7FFF7F", "yellow", "#FF7F00", "red",
    "#7F0000"
  ),
  bias = 2
  )
  ## actual plotting
  if (type == "contour") {
    # set default parameters
    pars <- list(
      x = points,
      y = points,
      z = dens,
      levels = levels,
      xlim = xlim,
      ylim = ylim,
      xlab = switch(margins,
        "unif" = expression(u[1]),
        "norm" = expression(z[1]),
        "exp" = expression(e[1]),
        "flexp" = expression(e[1])
      ),
      ylab = switch(margins,
        "unif" = expression(u[2]),
        "norm" = expression(z[2]),
        "exp" = expression(e[2]),
        "flexp" = expression(e[2])
      )
    )
  } else if (type == "surface") {
    # list with coordinates
    lst <- list(u = gu, v = gv, c = as.vector(dens))

    # set default parameters
    pars <- list(
      x = c ~ u * v,
      data = lst,
      scales = list(arrows = FALSE),
      drape = TRUE, colorkey = FALSE,
      screen = list(z = 25, x = -55),
      shade = FALSE,
      aspect = c(1, 1),
      light.source = c(10, 0, 10),
      zoom = 0.85,
      par.settings = list(axis.line = list(col = "transparent")),
      col = NA,
      col.regions = jet.colors(100),
      xlab = switch(margins,
        "unif" = expression(u[1]),
        "norm" = expression(z[1]),
        "exp" = expression(e[1]),
        "flexp" = expression(e[1])
      ),
      ylab = switch(margins,
        "unif" = expression(u[2]),
        "norm" = expression(z[2]),
        "exp" = expression(e[2]),
        "flexp" = expression(e[2])
      ),
      zlab = "",
      zlim = switch(margins,
        "unif" = c(0, max(3, 1.1 * max(lst$c))),
        "norm" = c(0, max(0.4, 1.1 * max(lst$c))),
        "exp" = c(0, max(1, 1.1 * max(lst$c))),
        "flexp" = c(0, max(1, 1.1 * max(lst$c)))
      )
    )
  }

  pars <- modifyList(pars, list(...))
  plot_obj <- switch(type,
    contour = do.call(contour, pars),
    surface = print(do.call(wireframe, pars))
  )
  invisible(plot_obj)
}

#' @rdname plot.bicop_dist
#' @export
plot.bicop <- plot.bicop_dist

#' @method contour bicop_dist
#' @rdname plot.bicop_dist
#' @export
contour.bicop_dist <- function(x, margins = "norm", size = 100L, ...) {
  plot(x, type = "contour", margins = margins, size = size, ...)
}

#' @rdname plot.bicop_dist
#' @export
contour.bicop <- contour.bicop_dist

#' (Inverse) Rosenblatt transform
#'
#' The Rosenblatt transform takes data generated from a model and turns it into
#' independent uniform variates, The inverse Rosenblatt transform computes
#' conditional quantiles and can be used simulate from a stochastic model,
#' see *Details*.
#'
#' @name rosenblatt
#' @aliases rosenblatt inverse_rosenblatt
#'
#' @param x matrix of evaluation points; must be in \eqn{(0, 1)^d} for copula
#'   models.
#' @param u matrix of evaluation points; must be in \eqn{(0, 1)^d}.
#' @param model a model object; classes currently supported are
#'    `bicop_dist()`, `vinecop_dist()`, and `vine_dist()`.
#' @param cores if `>1`, computation is parallelized over `cores` batches (rows
#'    of `u`).
#'
#' @details
#' The Rosenblatt transform (Rosenblatt, 1952) \eqn{U = T(V)} of a random vector
#' \eqn{V = (V_1,\ldots,V_d) ~ F} is defined as
#' \deqn{
#'   U_1= F(V_1), U_{2} = F(V_{2}|V_1), \ldots, U_d =F(V_d|V_1,\ldots,V_{d-1}),
#' }
#' where \eqn{F(v_k|v_1,\ldots,v_{k-1})} is the conditional distribution of
#' \eqn{V_k} given \eqn{V_1 \ldots, V_{k-1}, k = 2,\ldots,d}. The vector
#' \eqn{U  = (U_1, \dots, U_d)} then contains independent standard uniform
#' variables. The inverse operation
#' \deqn{
#'   V_1 = F^{-1}(U_1), V_{2} = F^{-1}(U_2|U_1), \ldots,
#'   V_d =F^{-1}(U_d|U_1,\ldots,U_{d-1}),
#' }
#' can be used to simulate from a distribution. For any copula \eqn{F}, if
#' \eqn{U} is a vector of independent random variables, \eqn{V = T^{-1}(U)} has
#' distribution \eqn{F}.
#'
#' The formulas above assume a vine copula model with order \eqn{d, \dots, 1}.
#' More generally, `rosenblatt()` returns the variables
#' \deqn{
#'   U_{M[d + 1- j, j]}= F(V_{M[d + 1- j, j]} | V_{M[d - j, j - 1]}, \dots, V_{M[1, 1]}),
#' }
#' where \eqn{M} is the structure matrix. Similarly, `inverse_rosenblatt()`
#' returns
#' \deqn{
#'   V_{M[d + 1- j, j]}= F^{-1}(U_{M[d + 1- j, j]} | U_{M[d - j, j - 1]}, \dots, U_{M[1, 1]}).
#' }
#'
#'
#' @examples
#' # simulate data with some dependence
#' x <- replicate(3, rnorm(200))
#' x[, 2:3] <- x[, 2:3] + x[, 1]
#' pairs(x)
#'
#' # estimate a vine distribution model
#' fit <- vine(x, copula_controls = list(family_set = "par"))
#'
#' # transform into independent uniforms
#' u <- rosenblatt(x, fit)
#' pairs(u)
#'
#' # inversion
#' pairs(inverse_rosenblatt(u, fit))
#'
#' # works similarly for vinecop models
#' vc <- fit$copula
#' rosenblatt(pseudo_obs(x), vc)
#' @export
rosenblatt <- function(x, model, cores = 1) {
  assert_that(
    inherits(model, c("bicop_dist", "vinecop_dist", "vine_dist")),
    is.number(cores)
  )

  to_col <- if (inherits(model, "bicop_dist")) FALSE else (dim(model)[1] == 1)
  x <- if_vec_to_matrix(x, to_col)
  col_names <- colnames(x)

  if (inherits(model, "bicop_dist")) {
    assert_that(ncol(x) == 2, all((x > 0) & (x < 1)))
    x <- cbind(x[, 1], bicop_hfunc1_cpp(x, model))
  } else if (inherits(model, "vinecop_dist")) {
    assert_that(all((x > 0) & (x < 1)))
    x <- vinecop_rosenblatt_cpp(x, model, cores)
  } else {
    # prepare marginals if only one is specified
    if (!inherits(vine, "vine") & depth(model$margins) == 1) {
      model$margins <- replicate(dim(model)[1], model$margins, simplify = FALSE)
    }
    x <- dpq_marg(x, model, "p")
    x <- vinecop_rosenblatt_cpp(x, model$copula, cores)
  }
  colnames(x) <- col_names

  x
}

#' @rdname rosenblatt
#' @export
inverse_rosenblatt <- function(u, model, cores = 1) {
  assert_that(
    all((u > 0) & (u < 1)),
    inherits(model, c("bicop_dist", "vinecop_dist", "vine_dist")),
    is.number(cores)
  )

  to_col <- if (inherits(model, "bicop_dist")) FALSE else (dim(model)[1] == 1)
  u <- if_vec_to_matrix(u, to_col)
  col_names <- colnames(u)

  if (inherits(model, "bicop_dist")) {
    assert_that(ncol(u) == 2)
    u <- cbind(u[, 1], bicop_hinv1_cpp(u, model))
  } else if (inherits(model, "vinecop_dist")) {
    u <- vinecop_inverse_rosenblatt_cpp(u, model, cores)
  } else {
    u <- vinecop_inverse_rosenblatt_cpp(u, model$copula, cores)
    # prepare marginals if only one is specified
    if (!inherits(vine, "vine") & depth(model$margins) == 1) {
      model$margins <- replicate(dim(model)[1], model$margins, simplify = FALSE)
    }
    u <- dpq_marg(u, model, "q")
  }
  colnames(u) <- col_names

  u
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Inverse Rosenblatt transform")

pc <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(list(pc, pc), list(pc))
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
vc <- vinecop_dist(pcs, mat)
vd <- vine_dist(list(distr = "norm"), pcs, mat)

test_that("rosenblatt works with bivariate copulas", {
  u <- rbicop(20, pc)
  expect_eql(inverse_rosenblatt(rosenblatt(u, pc), pc), u)
  pc <- bicop(u, family = "clay")
  expect_eql(inverse_rosenblatt(rosenblatt(u, pc), pc), u)
})

test_that("rosenblatt works with vine copulas", {
  u <- rvinecop(20, vc)
  expect_eql(inverse_rosenblatt(rosenblatt(u, vc), vc), u)
  vc <- vinecop(u, structure = mat, family = "clay")
  expect_eql(inverse_rosenblatt(rosenblatt(u, vc), vc), u)
})

test_that("rosenblatt works with vine distribution", {
  u <- rvine(20, vd)
  expect_eql(inverse_rosenblatt(rosenblatt(u, vd), vd), u)
  vd <- vine(u, copula_controls = list(structure = mat, family = "clay"))
  expect_eql(inverse_rosenblatt(rosenblatt(u, vd), vd), u)
})

#' Exploratory pairs plot for copula data
#'
#' This function provides pair plots for copula data. It shows
#' bivariate contour plots on the lower panel, scatter plots and
#' correlations on the upper panel and histograms on the diagonal panel.
#'
#' @param data the data (must lie in the unit hypercube).
#' @param \dots other parameters passed to `pairs.default()`,
#' `contour.bicop()`, `points.default()`, `hist.default()`, or `bicop()`.
#'
#' @importFrom stats cor
#' @importFrom graphics box hist.default pairs.default points.default text.default
#' @export
#'
#' @examples
#' u <- replicate(3, runif(100))
#' pairs_copula_data(u)
pairs_copula_data <- function(data, ...) {
  assert_that(is.matrix(data) || is.data.frame(data))
  assert_that(all(data < 1) && all(data > 0))

  old_par <- par(xaxt = "n", yaxt = "n")
  on.exit(par(old_par))

  labels <- colnames(data)
  if (is.null(labels))
    labels <- paste0("var", seq_len(ncol(data)))
  args <- list(x = data,
               labels = labels,
               lower.panel = function(x, y) lp_pairs_copula_data(x, y, ...),
               diag.panel = function(x) dp_pairs_copula_data(x, ...),
               upper.panel = function(x, y)  up_pairs_copula_data(x, y, ...),
               label.pos = 0.85,
               cex.labels = 1,
               gap = 0)
  call_with_dots(pairs.default, args, ...)
}

call_with_dots <- function(fun, args, ..., except = "") {
  dots <- list(...)
  dots <- dots[names(dots) %in% setdiff(names(formals(fun)), except)]
  args <- modifyList(args, dots)
  do.call(fun, args)
}

## lower panel: empirical contour plot
lp_pairs_copula_data <- function(x, y, ...) {
  old_par <- par(usr = c(-3, 3, -3, 3), new = TRUE)
  on.exit(par(old_par))

  args <- list(data = cbind(x, y), family_set = "tll")
  cop <- call_with_dots(bicop, args, ...)

  args <- list(x = cop,
               size = 100,
               axes = FALSE,
               drawlabels = FALSE)
  call_with_dots(contour.bicop, args, ...)
}

## upper panel: scatter plot (copula data) and correlation
up_pairs_copula_data <- function(x, y, ...) {
  old_par <- par(usr = c(0, 1, 0, 1), new = TRUE)
  on.exit(par(old_par))

  args <- list(x = x, y = y, pch = ".", cex = 2, col = "gray40")
  call_with_dots(points.default, args, ...)

  cor <- call_with_dots(cor, list(x = x, y = y, method = "kendall"), ...)
  txt <- format(x = cor, digits = 2, nsmall = 2)[1]
  args <- list(x = 0.5, y = 0.5,
               labels = txt,
               cex = 1 + abs(cor) * 2)
  call_with_dots(text.default, args, ...)
}

## diagonal panel: histograms (copula data)
dp_pairs_copula_data <- function(x, ...) {
  old_par <- par(usr = c(0, 1, 0, 1.6), new = TRUE)
  on.exit(par(old_par))

  args <- list(x = x,
               freq = FALSE,
               add = TRUE,
               border = "white",
               main = "",
               col = "grey")
  call_with_dots(hist.default, args, ..., except = c("col", "freq"))
  box()
  abline(h = 1, col = "black", lty = 3)
}

#' Fitting vine copula models
#'
#' Automated fitting and model selection for vine copula models with continuous
#' or discrete data. Selection of the structure is performed using the algorithm
#' of Dissmann et al. (2013).
#'
#' @inheritParams bicop
#' @param family_set a character vector of families; see [bicop()] for
#'   additional options.
#' @param structure an `rvine_structure` object, namely a compressed
#'   representation of the vine structure, or an object that can be coerced into
#'   one (see [rvine_structure()] and [as_rvine_structure()]). The dimension
#'   must be `length(pair_copulas[[1]]) + 1`; `structure = NA` performs
#'   automatic selection based on Dissman's algorithm. See *Details* for partial
#'   selection of the structure.
#' @param psi0 prior probability of a non-independence copula (only used for
#'   `selcrit = "mbic"` and `selcrit = "mbicv"`).
#' @param trunc_lvl the truncation level of the vine copula; `Inf` means no
#'   truncation, `NA` indicates that the truncation level should be selected
#'   automatically by [mBICV()].
#' @param tree_crit the criterion for tree selection, one of `"tau"`, `"rho"`,
#'   `"hoeffd"`, `"mcor"`, or `"joe"` for Kendall's \eqn{\tau}, Spearman's
#'   \eqn{\rho}, Hoeffding's \eqn{D}, maximum correlation, or logarithm of
#'   the partial correlation, respectively.
#' @param threshold for thresholded vine copulas; `NA` indicates that the
#'   threshold should be selected automatically by [mBICV()].
#' @param show_trace logical; whether a trace of the fitting progress should be
#'   printed.
#' @param cores number of cores to use; if more than 1, estimation of pair
#'   copulas within a tree is done in parallel.
#' @param var_types variable types, a length d vector; e.g., `c("c", "c")` for
#'   two continuous variables, or `c("c", "d")` for first variable continuous
#'   and second discrete.
#'
#' @details
#'
#' ## Missing data
#' If there are missing data (i.e., `NA` entries), incomplete observations are
#' discarded before fitting a pair-copula. This is done on a pair-by-pair basis
#' so that the maximal available information is used.
#'
#' ## Discrete variables
#' The dependence measures used to select trees (default: Kendall's tau) are
#' corrected for ties (see [wdm::wdm]).
#'
#' Let `n` be the number of observations and `d` the number of variables.
#' When at least one variable is discrete, two types of
#' "observations" are required in `data`: the first `n  x  d` block
#' contains realizations of \eqn{F_{X_j}(X_j)}. The second `n  x  d`
#' block contains realizations of \eqn{F_{X_j}(X_j^-)}. The minus indicates a
#' left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
#' \eqn{F_{X_j}(X_j^-) = F_{X_j}(X_j - 1)}. For continuous variables the left
#' limit and the cdf itself coincide. Respective columns can be omitted in the
#' second block.
#'
#' ## Partial structure selection
#'
#' It is possible to fix the vine structure only in the first trees and select
#' the remaining ones automatically. To specify only the first `k` trees, supply
#' a `k`-truncated `rvine_structure()` or `rvine_matrix()`. All trees up to
#' `trunc_lvl` will then be selected automatically.
#'
#' @return
#'
#' Objects inheriting from `vinecop` and `vinecop_dist` for [vinecop()]. In
#' addition to the entries provided by [vinecop_dist()], there are:
#'
#' * `threshold`, the (set or estimated) threshold used for thresholding the
#' vine.
#'
#' * `data` (optionally, if `keep_data = TRUE` was used), the dataset that was
#' passed to [vinecop()].
#'
#' * `controls`, a `list` with fit controls that was passed to [vinecop()].
#'
#' * `nobs`, the number of observations that were used to fit the model.
#'
#' @seealso [vinecop()], [dvinecop()], [pvinecop()], [rvinecop()],
#'   [plot.vinecop()], [contour.vinecop()]
#' @references
#' Dissmann, J. F., E. C. Brechmann, C. Czado, and D. Kurowicka (2013).
#' *Selecting and estimating regular vine copulae and application to
#' financial returns.* Computational Statistics & Data Analysis, 59 (1),
#' 52-69.
#' @export
#'
#' @examples
#' ## simulate dummy data
#' x <- rnorm(30) * matrix(1, 30, 5) + 0.5 * matrix(rnorm(30 * 5), 30, 5)
#' u <- pseudo_obs(x)
#'
#' ## fit and select the model structure, family and parameters
#' fit <- vinecop(u)
#' summary(fit)
#' plot(fit)
#' contour(fit)
#'
#' ## select by log-likelihood criterion from one-paramter families
#' fit <- vinecop(u, family_set = "onepar", selcrit = "bic")
#' summary(fit)
#'
#' ## Gaussian D-vine
#' fit <- vinecop(u, structure = dvine_structure(1:5), family = "gauss")
#' plot(fit)
#' contour(fit)
#'
#' ## Partial structure selection with only first tree specified
#' structure <- rvine_structure(order = 1:5, list(rep(5, 4)))
#' structure
#' fit <- vinecop(u, structure = structure, family = "gauss")
#' plot(fit)
#'
#' ## 1-truncated model with random structure
#' fit <- vinecop(u, structure = rvine_structure_sim(5), trunc_lvl = 1)
#' contour(fit)
#'
#' ## Model for discrete data
#' x <- qpois(u, 1)  # transform to Poisson margins
#' # we require two types of observations (see Details)
#' u_disc <- cbind(ppois(x, 1), ppois(x - 1, 1))
#' fit <- vinecop(u_disc, var_types = rep("d", 5))
#'
#' ## Model for mixed data
#' x <- qpois(u[, 1], 1)  # transform first variable to Poisson margin
#' # we require two types of observations (see Details)
#' u_disc <- cbind(ppois(x, 1), u[, 2:5], ppois(x - 1, 1))
#' fit <- vinecop(u_disc, var_types = c("d", rep("c", 4)))
vinecop <- function(data, var_types = rep("c", NCOL(data)), family_set = "all",
                    structure = NA, par_method = "mle",
                    nonpar_method = "constant", mult = 1, selcrit = "aic",
                    weights = numeric(), psi0 = 0.9, presel = TRUE,
                    trunc_lvl = Inf, tree_crit = "tau", threshold = 0,
                    keep_data = FALSE, show_trace = FALSE, cores = 1) {
  assert_that(
    is.character(family_set),
    inherits(structure, "matrix") ||
      inherits(structure, "rvine_structure") ||
      (is.scalar(structure) && is.na(structure)),
    is.string(par_method),
    is.string(nonpar_method),
    is.number(mult), mult > 0,
    is.string(selcrit),
    is.numeric(weights),
    is.number(psi0), psi0 > 0, psi0 < 1,
    is.flag(presel),
    is.scalar(trunc_lvl),
    is.string(tree_crit),
    is.scalar(threshold),
    is.flag(keep_data),
    is.number(cores), cores > 0,
    correct_var_types(var_types)
  )

  # check if families known (w/ partial matching) and expand convenience defs
  family_set <- process_family_set(family_set, par_method)

  ## pre-process input
  if (is.scalar(structure) && is.na(structure)) {
    structure <- rvine_structure(seq_along(var_types))
  } else {
    structure <- as_rvine_structure(structure)
  }

  ## fit and select copula model
  vinecop <- vinecop_select_cpp(
    data = as.matrix(data),
    structure = structure,
    family_set = family_set,
    par_method = par_method,
    nonpar_method = nonpar_method,
    mult = mult,
    selection_criterion = selcrit,
    weights = weights,
    psi0 = psi0,
    preselect_families = presel,
    truncation_level = ifelse( # Inf cannot be passed to C++
      is.finite(trunc_lvl),
      trunc_lvl,
      .Machine$integer.max
    ),
    tree_criterion = tree_crit,
    threshold = threshold,
    select_truncation_level = is.na(trunc_lvl),
    select_threshold = is.na(threshold),
    show_trace = show_trace,
    num_threads = cores,
    var_types = var_types
  )

  ## make all pair-copulas bicop objects
  vinecop$pair_copulas <- lapply(
    vinecop$pair_copulas,
    function(tree) lapply(tree, as.bicop)
  )

  ## add information about the fit
  vinecop$names <- colnames(data)
  if (keep_data) {
    vinecop$data <- data
  }
  vinecop$controls <- list(
    family_set = family_set,
    par_method = par_method,
    nonpar_method = nonpar_method,
    mult = mult,
    selcrit = selcrit,
    weights = weights,
    presel = presel,
    trunc_lvl = trunc_lvl,
    tree_crit = tree_crit,
    threshold = threshold
  )
  vinecop$nobs <- NROW(data)
  vinecop
}

#' Vine copula models
#'
#' Create custom vine copula models by specifying the pair-copulas, structure,
#' and variable types.
#'
#' @return
#'
#' Object of class `vinecop_dist`, i.e., a list containing:
#'
#' * `pair_copulas`, a list of lists. Each element of `pair_copulas` corresponds
#' to a tree, which is itself a list of [bicop_dist()] objects.
#'
#' * `structure`, a compressed representation of the vine structure, or an
#' object that can be coerced into one (see [rvine_structure()] and
#' [as_rvine_structure()]).
#'
#' * `npars`, a `numeric` with the number of (effective) parameters.
#'
#' * `var_types` the variable types.
#'
#' @inheritParams vinecop
#' @param pair_copulas A nested list of '[bicop_dist()]' objects, where
#'   \code{pair_copulas[[t]][[e]]} corresponds to the pair-copula at edge `e` in
#'   tree `t`.
#' @seealso [rvine_structure()], [rvine_matrix()], [vinecop()],
#'   [plot.vinecop_dist()], [contour.vinecop_dist()], [dvinecop()],
#'   [pvinecop()], [rvinecop()]
#' @export
#' @examples
#' # specify pair-copulas
#' bicop <- bicop_dist("bb1", 90, c(3, 2))
#' pcs <- list(
#'   list(bicop, bicop), # pair-copulas in first tree
#'   list(bicop) # pair-copulas in second tree
#' )
#'
#' # specify R-vine matrix
#' mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
#'
#' # set up vine copula model
#' vc <- vinecop_dist(pcs, mat)
#'
#' # visualization
#' plot(vc)
#' contour(vc)
#'
#' # simulate from the model
#' pairs(rvinecop(200, vc))
vinecop_dist <- function(pair_copulas, structure,
                         var_types = rep("c", dim(structure)[1])) {
  # create object
  vinecop <- structure(
    list(
      pair_copulas = pair_copulas,
      structure = as_rvine_structure(structure)
    ),
    class = "vinecop_dist"
  )

  # sanity checks
  assert_that(is.list(pair_copulas), correct_var_types(var_types))
  if (length(pair_copulas) > 0) {
    if (length(pair_copulas) > length(pair_copulas[[1]])) {
      stop("'pair_copulas' has more trees than variables.")
    }
  }

  pc_lst <- unlist(pair_copulas, recursive = FALSE)
  if (!all(sapply(pc_lst, function(x) inherits(x, "bicop_dist")))) {
    stop("some objects in pair_copulas aren't of class 'bicop_dist'")
  }

  vinecop$structure <- truncate_model(
    vinecop$structure,
    length(vinecop$pair_copulas)
  )
  vinecop$var_types <- var_types
  vinecop_check_cpp(vinecop)
  vinecop$npars <- do.call(sum, lapply(pc_lst, function(x) x[["npars"]]))
  vinecop$loglik <- NA

  vinecop
}

family_set_archimedean <- c(
  "clayton", "gumbel", "frank", "joe"
)

family_set_elliptical <- c(
  "gaussian", "t"
)

family_set_bb <- c(
  "bb1", "bb6", "bb7", "bb8"
)

family_set_onepar <- c(
  "gaussian", family_set_archimedean
)

family_set_twopar <- c(
  "t", family_set_bb
)

family_set_parametric <- c(
  "indep", family_set_onepar, family_set_twopar
)

family_set_nonparametric <- c(
  "indep", "tll"
)

family_set_itau <- c(
  "indep", family_set_onepar, "t"
)

family_set_rotationless <- c(
  "frank", family_set_elliptical, family_set_nonparametric
)

family_set_all <- unique(
  c(family_set_parametric, family_set_nonparametric)
)

family_set_defs <- c(
  "archimedean", "elliptical", "bbs", "oneparametric", "twoparametric",
  "parametric", "nonparametric", "itau", "all"
)

family_set_all_defs <- c(
  family_set_all, family_set_defs
)

#' Bivariate copula distributions
#'
#' Density, distribution function, random generation and h-functions (with their
#' inverses) for the bivariate copula distribution.
#'
#' @name bicop_distributions
#' @aliases dbicop pbicop rbicop hbicop dbicop_dist pbicop_dist rbicop_dist
#'   hbicop_dist
#'
#' @param u evaluation points, a matrix with at least two columns, see
#'   *Details*.
#' @param family the copula family, a string containing the family name (see
#'   \code{\link{bicop}} for all possible families).
#' @param rotation the rotation of the copula, one of `0`, `90`, `180`, `270`.
#' @param parameters a vector or matrix of copula parameters.
#' @param var_types variable types, a length 2 vector; e.g., `c("c", "c")` for
#'   both continuous (default), or `c("c", "d")` for first variable continuous
#'   and second discrete.
#'
#' @note The functions can optionally be used with a [bicop_dist] object, e.g.,
#'   `dbicop(c(0.1, 0.5), bicop_dist("indep"))`.
#'
#' @details
#' See [bicop] for the various implemented copula families.
#'
#' The copula density is defined as joint density divided by marginal
#' densities, irrespective of variable types.
#'
#' H-functions (`hbicop()`) are conditional distributions derived
#' from a copula. If \eqn{C(u, v) = P(U \le u, V \le v)} is a copula, then
#' \deqn{h_1(u, v) = P(V \le v | U = u) = \partial C(u, v) / \partial u,}
#' \deqn{h_2(u, v) = P(U \le u | V = v) = \partial C(u, v) / \partial v.}
#' In other words, the H-function number refers to the conditioning variable.
#' When inverting H-functions, the inverse is then taken with respect to the
#' other variable, that is `v` when `cond_var = 1` and `u` when `cond_var = 2`.
#'
#' ## Discrete variables
#' When at least one variable is discrete, more than two columns are required
#' for `u`: the first \eqn{n \times 2} block contains realizations of
#' \eqn{F_{X_1}(x_1), F_{X_2}(x_2)}. The second \eqn{n \times 2} block contains
#' realizations of \eqn{F_{X_1}(x_1^-), F_{X_1}(x_1^-)}. The minus indicates a
#' left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
#' \eqn{F_{X_1}(x_1^-) = F_{X_1}(x_1 - 1)}. For continuous variables the left
#' limit and the cdf itself coincide. Respective columns can be omitted in the
#' second block.
#'
#' @return
#' `dbicop()` gives the density, `pbicop()` gives the distribution function,
#' `rbicop()` generates random deviates, and `hbicop()` gives the h-functions
#' (and their inverses).
#'
#' The length of the result is determined by `n` for `rbicop()`, and
#' the number of rows in `u` for the other functions.
#'
#' The numerical arguments other than `n` are recycled to the length of the
#' result.
#'
#' @seealso [bicop_dist()], [bicop()]
#
#' @examples
#' ## evaluate the copula density
#' dbicop(c(0.1, 0.2), "clay", 90, 3)
#' dbicop(c(0.1, 0.2), bicop_dist("clay", 90, 3))
#'
#' ## evaluate the copula cdf
#' pbicop(c(0.1, 0.2), "clay", 90, 3)
#'
#' ## simulate data
#' plot(rbicop(500, "clay", 90, 3))
#'
#' ## h-functions
#' joe_cop <- bicop_dist("joe", 0, 3)
#' # h_1(0.1, 0.2)
#' hbicop(c(0.1, 0.2), 1, "bb8", 0, c(2, 0.5))
#' # h_2^{-1}(0.1, 0.2)
#' hbicop(c(0.1, 0.2), 2, joe_cop, inverse = TRUE)
#'
#' ## mixed discrete and continuous data
#' x <- cbind(rpois(10, 1), rnorm(10, 1))
#' u <- cbind(ppois(x[, 1], 1), pnorm(x[, 2]), ppois(x[, 1] - 1, 1))
#' pbicop(u, "clay", 90, 3, var_types = c("d", "c"))
#'
#' @rdname bicop_methods
#' @export
dbicop <- function(u, family, rotation, parameters, var_types = c("c", "c")) {
  bicop <- args2bicop(family, rotation, parameters, var_types)
  bicop_pdf_cpp(if_vec_to_matrix(u), bicop)
}
#' @rdname bicop_methods
#' @export
pbicop <- function(u, family, rotation, parameters, var_types = c("c", "c")) {
  bicop <- args2bicop(family, rotation, parameters, var_types)
  bicop_cdf_cpp(if_vec_to_matrix(u), bicop)
}

#' @param n number of observations. If `length(n) > 1``, the length is taken to
#'   be the number required.
#' @param qrng if `TRUE`, generates quasi-random numbers using the bivariate
#' Generalized Halton sequence (default `qrng = FALSE`).
#' @rdname bicop_methods
#' @export
rbicop <- function(n, family, rotation, parameters, qrng = FALSE) {
  if (length(n) > 1) {
    n <- length(n)
  }
  if (inherits(family, "bicop_dist") & !missing(rotation)) {
    qrng <- rotation
  }
  assert_that(is.flag(qrng))

  bicop <- args2bicop(family, rotation, parameters)
  U <- bicop_sim_cpp(bicop, n, qrng, get_seeds())
  if (!is.null(bicop$names)) {
    colnames(U) <- bicop$names
  }

  U
}


#' @rdname bicop_methods
#' @param cond_var either `1` or `2`; `cond_var = 1` conditions on the first
#'    variable, `cond_var = 2` on the second.
#' @param inverse whether to compute the h-function or its inverse.
#' @export
hbicop <- function(u, cond_var, family, rotation, parameters, inverse = FALSE,
                   var_types = c("c", "c")) {
  assert_that(in_set(cond_var, 1:2), is.flag(inverse))
  bicop <- args2bicop(family, rotation, parameters, var_types)
  u <- if_vec_to_matrix(u)

  if (!inverse) {
    if (cond_var == 1) {
      return(bicop_hfunc1_cpp(u, bicop))
    } else {
      return(bicop_hfunc2_cpp(u, bicop))
    }
  } else {
    if (cond_var == 1) {
      return(bicop_hinv1_cpp(u, bicop))
    } else {
      return(bicop_hinv2_cpp(u, bicop))
    }
  }
}

#' Conversion between Kendall's tau and parameters
#'
#' @param family a copula family (see [bicop_dist()]) or a [bicop_dist] object.
#' @param rotation the rotation of the copula, one of `0`, `90`, `180`, `270`.
#' @param parameters vector or matrix of copula parameters, not used when
#'   `family` is a `bicop_dist` object.
#'
#' @export
#'
#' @examples
#' # the following are equivalent
#' par_to_ktau(bicop_dist("clayton", 0, 3))
#' par_to_ktau("clayton", 0, 3)
#'
#' ktau_to_par("clayton", 0.5)
#' ktau_to_par(bicop_dist("clayton", 0, 3), 0.5)
#' @name par_to_ktau
#' @rdname par_to_ktau
par_to_ktau <- function(family, rotation, parameters) {
  bicop <- args2bicop(family, rotation, parameters)
  bicop_par_to_tau_cpp(bicop)
}

#' @rdname par_to_ktau
#' @param tau Kendall's \eqn{\tau}.
#' @export
ktau_to_par <- function(family, tau) {
  assert_that(is.number(tau), !is.na(tau))
  bicop <- args2bicop(family)
  if (!(bicop$family %in% family_set_rotationless)) {
    bicop$rotation <- ifelse(tau > 0, 0, 90)
  }
  bicop_tau_to_par_cpp(bicop, tau)
}

#' Predictions and fitted values for a bivariate copula model
#'
#' Predictions of the density, distribution function, h-functions (with their
#' inverses) for a bivariate copula model.
#'
#' @name bicop_predict_and_fitted
#' @aliases predict.bicop fitted.bicop
#' @param object a `bicop` object.
#' @param newdata points where the fit shall be evaluated.
#' @param what what to predict, one of `"pdf"`, `"cdf"`, `"hfunc1"`, `"hfunc2"`,
#'   `"hinv1"`, `"hinv2"`.
#' @param ... unused.
#' @return `fitted()` and `logLik()` have return values similar to [dbicop()],
#' [pbicop()], and [hbicop()].
#'
#' @details `fitted()` can only be called if the model was fit with the
#'   `keep_data = TRUE` option.
#'
#'
#' ## Discrete variables
#' When at least one variable is discrete, more than two columns are required
#' for `newdata`: the first \eqn{n \times 2} block contains realizations of
#' \eqn{F_{X_1}(x_1), F_{X_2}(x_2)}. The second \eqn{n \times 2} block contains
#' realizations of \eqn{F_{X_1}(x_1^-), F_{X_1}(x_1^-)}. The minus indicates a
#' left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
#' \eqn{F_{X_1}(x_1^-) = F_{X_1}(x_1 - 1)}. For continuous variables the left
#' limit and the cdf itself coincide. Respective columns can be omitted in the
#' second block.
#'
#'
#' @examples
#' # Simulate and fit a bivariate copula model
#' u <- rbicop(500, "gauss", 0, 0.5)
#' fit <- bicop(u, family = "par", keep_data = TRUE)
#'
#' # Predictions
#' all.equal(predict(fit, u, "hfunc1"), fitted(fit, "hfunc1"),
#'           check.environment = FALSE)
#' @rdname predict_bicop
#' @export
predict.bicop_dist <- function(object, newdata, what = "pdf", ...) {
  assert_that(in_set(what, what_allowed))
  newdata <- if_vec_to_matrix(newdata)
  switch(
    what,
    "pdf" = bicop_pdf_cpp(newdata, object),
    "cdf" = bicop_cdf_cpp(newdata, object),
    "hfunc1" = bicop_hfunc1_cpp(newdata, object),
    "hfunc2" = bicop_hfunc2_cpp(newdata, object),
    "hinv1" = bicop_hinv1_cpp(newdata, object),
    "hinv2" = bicop_hinv2_cpp(newdata, object)
  )
}

what_allowed <- c("pdf", "cdf", "hfunc1", "hfunc2", "hinv1", "hinv2")

#' @rdname predict_bicop
#' @export
fitted.bicop <- function(object, what = "pdf", ...) {
  if (is.null(object$data)) {
    stop("data have not been stored, use keep_data = TRUE when fitting.")
  }
  assert_that(in_set(what, what_allowed))
  switch(
    what,
    "pdf" = bicop_pdf_cpp(object$data, object),
    "cdf" = bicop_cdf_cpp(object$data, object),
    "hfunc1" = bicop_hfunc1_cpp(object$data, object),
    "hfunc2" = bicop_hfunc2_cpp(object$data, object),
    "hinv1" = bicop_hinv1_cpp(object$data, object),
    "hinv2" = bicop_hinv2_cpp(object$data, object)
  )
}

#' @importFrom stats logLik
#' @export
logLik.bicop <- function(object, ...) {
  structure(object$loglik, "df" = object$npars)
}

#' @export
print.bicop_dist <- function(x, ...) {
  x0 <- x
  if (x$family %in% setdiff(family_set_nonparametric, "indep")) {
    x$parameters <- paste0(round(x$npars, 2), sep = " d.f.")
  }
  cat("Bivariate copula ('bicop_dist'): ",
    "family = ", x$family,
    ", rotation = ", x$rotation,
    ", parameters = ", ifelse(length(x$parameters) > 1,
      paste(round(x$parameters, 2),
        collapse = ", "
      ),
      x$parameters
    ),
    ", var_types = ", paste(x$var_types, collapse = ","),
    sep = ""
  )
  cat("\n")
  invisible(x0)
}

#' @export
summary.bicop_dist <- function(object, ...) {
  print.bicop_dist(object, ...)
}

#' @export
print.bicop <- function(x, ...) {
  x0 <- x
  if (x$family %in% setdiff(family_set_nonparametric, "indep")) {
    pars_formatted <- paste0(round(x$npars, 2), sep = " d.f.")
  } else {
    pars_formatted <- paste(round(x$parameters, 2), collapse = ", ")
  }
  cat("Bivariate copula fit ('bicop'): ",
    "family = ", x$family,
    ", rotation = ", x$rotation,
    ", parameters = ", pars_formatted,
    ", var_types = ", paste(x$var_types, collapse = ","),
    "\n",
    sep = ""
  )

  invisible(x0)
}

#' @export
summary.bicop <- function(object, ...) {
  print.bicop(object, ...)
  cat("nobs =", object$nobs, "  ")

  info <- bicop_fit_info(object)
  cat("logLik =", round(info$logLik, 2), "  ")
  cat("npars =", round(info$npars, 2), "  ")
  cat("AIC =", round(info$AIC, 2), "  ")
  cat("BIC =", round(info$BIC, 2), "  ")
  attr(object, "info") <- info
  cat("\n")

  invisible(object)
}

#' @importFrom stats coef
#' @export
coef.bicop_dist <- function(object, ...) {
  object$parameters
}

bicop_fit_info <- function(bc) {
  ll <- logLik(bc)
  list(
    nobs = bc$nobs,
    logLik = ll[1],
    npars = attr(ll, "df"),
    AIC = -2 * ll[1] + 2 * attr(ll, "df"),
    BIC = -2 * ll[1] + log(bc$nobs) * attr(ll, "df")
  )
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Discrete variables")

set.seed(5)

test_that("bicop_dist works", {
  cop <- bicop_dist("gum", 90, 4, c("d", "c"))
  expect_identical(cop$var_types, c("d", "c"))

  u <- rbicop(10, "gum", 90, 4)
  u <- cbind(u, u)
  u[, 1] <- ceiling(u[, 1] * 3) / 3
  u[, 3] <- floor(u[, 3] * 3) / 3

  # only check for errors
  dbicop(u, cop)
  pbicop(u, cop)
  hbicop(u, 1, cop)
  hbicop(u, 2, cop, TRUE)
  rbicop(10, cop)

  dbicop(u, "gum", 90, 4, c("d", "c"))
  pbicop(u, "gum", 90, 4, c("d", "c"))
  hbicop(u, 2, "gum", 90, 4, FALSE, c("d", "c"))
  hbicop(u, 1, "gum", 90, 4, TRUE, c("d", "c"))
})


test_that("bicop works", {
  u <- rbicop(1000, "gum", 90, 4)
  u <- cbind(u, u)
  u[, 1] <- ceiling(u[, 1] * 3) / 3
  u[, 3] <- floor(u[, 3] * 3) / 3
  cop <- bicop(u,
               family = "gumbel",
               var_types = c("d", "c"),
               presel = FALSE,
               keep_data = TRUE)
  expect_eql(cop$family, "gumbel")
  expect_eql(cop$rotation, 90)
  expect_eql(cop$parameters[1], 4, tol = 0.5)
  expect_identical(cop$var_types, c("d", "c"))
})

# -----------------------------------------------------------------

set.seed(0)
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(list(bicop, bicop), list(bicop))
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
var_types <- c("d", "d", "c")

test_that("vinecop_dist works", {
  cop <- vinecop_dist(pcs, mat, var_types)
  expect_eql(cop$var_types, var_types)

  u <- rvinecop(20, cop)
  u <- cbind(u, u)
  u[, 1] <- ceiling(u[, 1] * 3) / 3
  u[, 3] <- floor(u[, 3] * 3) / 3
  u[, 2] <- ceiling(u[, 2] * 10) / 10
  u[, 3] <- floor(u[, 4] * 10) / 10

  # only check for errors
  dvinecop(u, cop)
  pvinecop(u, cop)
})

test_that("vinecop works", {
  u <- replicate(3, runif(20))
  u <- cbind(u, u)
  u[, 1] <- ceiling(u[, 1] * 3) / 3
  u[, 3] <- floor(u[, 3] * 3) / 3
  u[, 2] <- ceiling(u[, 2] * 10) / 10
  u[, 3] <- floor(u[, 4] * 10) / 10
  cop <- vinecop(u, var_types = var_types, family = "tll")
  summary(cop)
  expect_identical(cop$var_types, var_types)
})


# -----------------------------------------------------------------

test_that("vine works", {
  n <- 20
  x1 <- rnorm(n)
  x2 <- ordered(sample(5, n, TRUE), 1:5)
  x3 <- x1 + as.numeric(x2) + rnorm(n, sd = 0.5)
  x <- data.frame(x1, x2, x3)

  fit <- vine(x)
  sim <- rvine(n * 10, fit)
  expect_true(is.data.frame(sim))
  expect_eql(sort(unique(sim[, 2])), ordered(1:5))

  summary(fit)
  expect_identical(fit$copula$var_types, c("c", "d", "c"))
  # only check for errors
  dvine(x, fit)
  pvine(x, fit)
  rvine(10, fit)
})

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Class 'bicop_dist'")

test_that("constructor creates proper bicop_dist object", {
  dist <- bicop_dist("gumbel", 90, 3)
  expect_s3_class(dist, "bicop_dist")
  expect_identical(names(dist),
                   c("family", "rotation", "parameters", "var_types", "npars"))
})


test_that("checks for family/rotation/parameters consistency", {
  expect_error(bicop_dist("asdf", 90, 3))
  expect_error(bicop_dist("frank", 3, 3))
  expect_error(bicop_dist("frank", 90, -3))
  expect_error(bicop_dist("frank", 90, 1:2))
})

test_that("partial matching for family names", {
  expect_eql(bicop_dist("ind")$family, "indep")
  expect_eql(bicop_dist("gauss")$family, "gaussian")
  expect_error(bicop_dist("g"))
})

test_that("d/p/r/h functions work", {
  set.seed(0)
  dist <- bicop_dist("bb1", 270, c(1, 2))
  u <- rbicop(50, "bb1", 270, c(1, 2))
  expect_false(any(rbicop(50, dist, qrng = FALSE) ==
    rbicop(50, dist, qrng = FALSE)))

  set.seed(1)
  u1 <- rbicop(50, dist, qrng = TRUE)
  set.seed(1)
  expect_true(all(u1 == rbicop(50, dist, qrng = TRUE)))
  expect_gte(min(dbicop(c(0.1, 0.2), dist)), 0)
  expect_gte(min(dbicop(u, dist)), 0)
  expect_gte(min(pbicop(u, dist)), 0)
  expect_lte(max(pbicop(u, dist)), 1)
  expect_lte(max(hbicop(c(0.1, 0.2), 1, dist)), 1)
  expect_lte(max(hbicop(u, 2, dist)), 1)
  expect_lte(max(hbicop(c(0.1, 0.2), 1, dist, inverse = TRUE)), 1)
  expect_lte(max(hbicop(u, 2, dist, inverse = TRUE)), 1)
})

test_that("plot functions work", {
  dist <- bicop_dist("gumbel", 90, 3)
  # we could check some values in the plot objects
  expect_silent(p <- plot(dist))
  expect_silent(p <- contour(dist))
  expect_silent(p <- plot(dist, margins = "norm"))
  expect_silent(p <- contour(dist, margins = "unif"))
  expect_silent(p <- plot(dist, margins = "exp"))
  expect_silent(p <- contour(dist, margins = "flexp"))
})

test_that("parameter <-> tau conversion works", {
  dist <- bicop_dist("joe", 90, 3)

  expect_eql(coef(dist), dist$parameters)

  # one-parameter family
  tau <- par_to_ktau(dist)
  expect_identical(tau, par_to_ktau("joe", 90, 3))

  par <- ktau_to_par(dist, tau)
  expect_identical(par, ktau_to_par("joe", tau))

  expect_eql(3, par[1])

  # two-parameter
  tau <- par_to_ktau("bb1", 0, c(1, 2))
  expect_error(ktau_to_par("bb1", 0.5))

  # rotationless
  ktau_to_par("frank", 0.5)
  ktau_to_par("frank", -0.5)
})

test_that("print method produces output", {
  dist <- bicop_dist("indep")
  expect_output(print(dist))
  expect_output(summary(dist))
})

test_that("getters work", {
  dist <- bicop_dist("gumbel", 90, 3)

  # test get_pair_copula
  expect_identical(dist, get_pair_copula(dist))
  expect_warning(get_pair_copula(dist, 1))
  expect_warning(get_pair_copula(dist, NA, 1))

  # test other getters
  expect_equiv(get_parameters(dist), coef(dist))
  expect_equiv(get_ktau(dist), par_to_ktau(dist))
  expect_equiv(get_family(dist), "gumbel")
})

#' Vine copula models
#'
#' Automated fitting or creation of custom vine copula models
#'
#' @aliases vine_dist
#' @param data a matrix or data.frame. Discrete variables have to be declared as
#' `ordered()`.
#' @param margins_controls a list with arguments to be passed to
#' [kde1d::kde1d()]. Currently, there can be
#'   * `mult` numeric vector of length one or d; all bandwidths for marginal
#'   kernel density estimation are multiplied with `mult`. Defaults to
#'   `log(1 + d)` where `d` is the number of variables after applying
#'   `rvinecopulib:::expand_factors()`.
#'   * `xmin` numeric vector of length d; see [kde1d::kde1d()].
#'   * `xmax` numeric vector of length d; see [kde1d::kde1d()].
#'   * `bw` numeric vector of length d; see [kde1d::kde1d()].
#'   * `deg` numeric vector of length one or d; [kde1d::kde1d()].
#' @param copula_controls a list with arguments to be passed to [vinecop()].
#' @param weights optional vector of weights for each observation.
#' @param keep_data whether the original data should be stored; if you want to
#'   store the pseudo-observations used for fitting the copula, use the
#'   `copula_controls` argument.
#' @param cores the number of cores to use for parallel computations.
#' @details
#' `vine_dist()` creates a vine copula by specifying the margins, a nested list
#' of `bicop_dist` objects and a quadratic structure matrix.
#'
#' `vine()` provides automated fitting for vine copula models.
#' `margins_controls` is a list with the same parameters as
#' [kde1d::kde1d()] (except for `x`). `copula_controls` is a list
#' with the same parameters as [vinecop()] (except for `data`).
#'
#' @return Objects inheriting from `vine_dist` for [vine_dist()], and
#' `vine` and `vine_dist` for [vine()].
#'
#' Objects from the `vine_dist` class are lists containing:
#'
#' * `margins`, a list of marginals (see below).
#' * `copula`, an object of the class `vinecop_dist`, see [vinecop_dist()].
#'
#' For objects from the `vine` class, `copula` is also an object of the class
#' `vine`, see [vinecop()]. Additionally, objects from the `vine` class contain:
#'
#' * `margins_controls`, a `list` with the set of fit controls that was passed
#' to [kde1d::kde1d()] when estimating the margins.
#' * `copula_controls`, a `list` with the set of fit controls that was passed
#' to [vinecop()] when estimating the copula.
#' * `data` (optionally, if `keep_data = TRUE` was used), the dataset that was
#' passed to [vine()].
#' * `nobs`, an `integer` containing the number of observations that was used
#' to fit the model.
#'
#' Concerning `margins`:
#'
#' * For objects created with [vine_dist()], it simply corresponds to the `margins`
#' argument.
#' * For objects created with [vine()], it is a list of objects of class `kde1d`,
#' see [kde1d::kde1d()].
#'
#' @examples
#' # specify pair-copulas
#' bicop <- bicop_dist("bb1", 90, c(3, 2))
#' pcs <- list(
#'   list(bicop, bicop), # pair-copulas in first tree
#'   list(bicop) # pair-copulas in second tree
#' )
#'
#' # specify R-vine matrix
#' mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
#'
#' # set up vine copula model with Gaussian margins
#' vc <- vine_dist(list(distr = "norm"), pcs, mat)
#'
#' # show model
#' summary(vc)
#'
#' # simulate some data
#' x <- rvine(50, vc)
#'
#' # estimate a vine copula model
#' fit <- vine(x, copula_controls = list(family_set = "par"))
#' summary(fit)
#'
#' ## model for discrete data
#' x <- as.data.frame(x)
#' x[, 1] <- ordered(round(x[, 1]), levels = seq.int(-5, 5))
#' fit_disc <- vine(x, copula_controls = list(family_set = "par"))
#' summary(fit_disc)
#'
#' @importFrom kde1d dkde1d pkde1d qkde1d
#' @export
vine <- function(data,
                 margins_controls = list(
                   mult = NULL,
                   xmin = NaN,
                   xmax = NaN,
                   bw = NA,
                   deg = 2
                 ),
                 copula_controls = list(
                   family_set = "all",
                   structure = NA,
                   par_method = "mle",
                   nonpar_method = "constant",
                   mult = 1,
                   selcrit = "aic",
                   psi0 = 0.9,
                   presel = TRUE,
                   trunc_lvl = Inf,
                   tree_crit = "tau",
                   threshold = 0,
                   keep_data = FALSE,
                   show_trace = FALSE,
                   cores = 1
                 ),
                 weights = numeric(),
                 keep_data = FALSE,
                 cores = 1) {
  ## basic sanity checks (copula_controls are checked by vinecop)
  data <- expand_factors(data)

  d <- ncol(data)
  var_types <- rep("c", d)
  var_types[sapply(data, is.ordered)] <- "d"

  assert_that(is.list(margins_controls))
  allowed_margins_controls <- c("mult", "xmin", "xmax", "bw", "deg")
  assert_that(in_set(names(margins_controls), allowed_margins_controls))
  assert_that(is.list(copula_controls))
  if (is.null(copula_controls$keep_data)) {
    copula_controls$keep_data <- TRUE
  }
  copula_controls$cores <- cores

  ## expand the required arguments and compute default mult if needed
  margins_controls <- expand_margin_controls(margins_controls, d, data)
  check_margin_controls(data, margins_controls)

  ## estimation of the marginals
  vine <- list()
  vine$margins <- fit_margins_cpp(prep_for_margins(data),
                                  sapply(data, nlevels),
                                  mult = margins_controls$mult,
                                  xmin = margins_controls$xmin,
                                  xmax = margins_controls$xmax,
                                  bw = margins_controls$bw,
                                  deg = margins_controls$deg,
                                  weights = weights,
                                  cores)
  vine$margins_controls <- margins_controls
  vine$margins <- finalize_margins(data, vine$margins)

  ## estimation of the R-vine copula --------------
  copula_controls$data <- compute_pseudo_obs(data, vine)
  copula_controls$var_types <- var_types
  copula_controls$weights <- weights
  vine$copula <- do.call(vinecop, copula_controls)
  vine$copula_controls <- copula_controls[-which(names(copula_controls) == "data")]

  finalize_vine(vine, data, weights, keep_data)
}

prep_for_margins <- function(data) {
  data <- lapply(data, function(x) if (is.ordered(x)) as.numeric(x) - 1 else x)
  do.call(cbind, data)
}

compute_pseudo_obs <- function(data, vine) {
  d <- ncol(data)
  u <- dpq_marg(data, vine)
  if (any(sapply(data, is.factor))) {
    u_sub <- u
    for (k in seq_len(d)) {
      if (is.factor(data[, k])) {
        lv <- as.numeric(data[, k]) - 1
        lv0 <- which(lv == 0)
        lv[lv0] <- 1
        xlv <- ordered(levels(data[, k])[lv], levels = levels(data[, k]))
        u_sub[, k] <- eval_one_dpq(xlv, vine$margins[[k]])
        u_sub[lv0, k] <- 0
      }
    }
  } else {
    u_sub <- NULL
  }
  cbind(u, u_sub)
}

#' @importFrom stats model.matrix
#' @noRd
expand_factors <- function(data) {
  if (is.data.frame(data)) {
    data <- lapply(data, function(x) {
      if (is.numeric(x) | is.ordered(x))
        return(x)
      x <- model.matrix(~ x)[, -1, drop = FALSE]
      x <- as.data.frame(x)
      x <- lapply(x, function(y) ordered(y, levels = 0:1))
    })
  }
  as.data.frame(data)
}

#' @param margins A list with with each element containing the specification of a
#' marginal [stats::Distributions]. Each marginal specification
#' should be a list with containing at least the distribution family (`"distr"`)
#' and optionally the parameters, e.g.
#' `list(list(distr = "norm"), list(distr = "norm", mu = 1), list(distr = "beta", shape1 = 1, shape2 = 1))`.
#' Note that parameters that have no default values have to be provided.
#' Furthermore, if `margins` has length one, it will be recycled for every component.
#' @param pair_copulas A nested list of 'bicop_dist' objects, where
#'    \code{pair_copulas[[t]][[e]]} corresponds to the pair-copula at edge `e` in
#'    tree `t`.
#' @param structure an `rvine_structure` object, namely a compressed
#' representation of the vine structure, or an object that can be coerced
#' into one (see [rvine_structure()] and [as_rvine_structure()]).
#' The dimension must be `length(pair_copulas[[1]]) + 1`.
#' @rdname vine
#' @export
vine_dist <- function(margins, pair_copulas, structure) {
  structure <- as_rvine_structure(structure)

  # sanity checks for the marg
  if (!(length(margins) %in% c(1, dim(structure)[1]))) {
    stop("marg should have length 1 or dim(structure)[1]")
  }
  stopifnot(is.list(margins))
  if (depth(margins) == 1) {
    check_marg <- check_distr(margins)
    npars_marg <- ncol(matrix) * get_npars_distr(margins)
  } else {
    check_marg <- lapply(margins, check_distr)
    npars_marg <- sum(sapply(margins, get_npars_distr))
  }
  is_ok <- sapply(check_marg, isTRUE)
  if (!all(is_ok)) {
    msg <- "Some objects in marg aren't properly defined.\n"
    msg <- c(msg, paste0("margin ", seq_along(check_marg)[!is_ok], " : ",
                         unlist(check_marg[!is_ok]), ".",
                         sep = "\n"
    ))
    stop(msg)
  }

  # create the vinecop object
  copula <- vinecop_dist(pair_copulas, structure)

  # create object
  structure(list(
    margins = margins,
    copula = copula,
    npars = copula$npars + npars_marg,
    loglik = NA
  ), class = "vine_dist")
}

expand_margin_controls <- function(controls, d, data) {
  default_controls <- list(mult = NULL, xmin = NaN, xmax = NaN, bw = NA, deg = 2)
  controls <- modifyList(default_controls, controls)
  if (is.null(controls[["mult"]])) {
    controls[["mult"]] <- log(1 + d)
  }
  for (par in names(controls)) {
    if (length(controls[[par]]) != ncol(data))
      controls[[par]] <- rep(controls[[par]], ncol(data))
  }
  controls
}

finalize_margins <- function(data, margins) {
  for (k in seq_along(margins)) {
    margins[[k]]$x <- data[[k]]
    margins[[k]]$nobs <- nrow(data)
  }
  margins
}

finalize_vine <- function(vine, data, weights, keep_data) {
  ## compute npars/loglik
  npars <- loglik <- 0
  for (k in seq_len(ncol(data))) {
    npars <- npars + vine$margins[[k]]$edf
    loglik <- loglik + vine$margins[[k]]$loglik
  }

  ## add the npars/loglik of the copulas
  vine$npars <- npars + vine$copula$npars
  vine$loglik <- loglik + vine$copula$loglik

  ## add data
  if (keep_data) {
    vine$data <- data
    vine$weights <- weights
  } else {
    vine$data <- matrix(NA, ncol = ncol(data))
    colnames(vine$data) <- colnames(data)
  }

  ## add number of observations
  vine$nobs <- nrow(data)
  vine$names <- vine$copula$names <- colnames(data)

  ## store levels for discrete variables
  vine$var_levels <- lapply(data, levels)

  ## create and return object
  structure(vine, class = c("vine", "vine_dist"))
}

#' Extracts components of `bicop_dist` and `vinecop_dist` objects
#'
#' Extracts either the structure matrix  (for `vinecop_dist` only), or
#' pair-copulas, their parameters, Kendall's taus, or families (for `bicop_dist`
#' and `vinecop_dist`).
#'
#' @name getters
#' @aliases get_pair_copula get_all_pair_copulas get_parameters
#'   get_all_parameters get_ktau get_all_ktaus get_matrix get_structure
#' @param object a `bicop_dist`, `vinecop_dist` or `vine_dist` object.
#' @details #' The [get_structure] method (for `vinecop_dist` or `vine_dist`
#' objects only) extracts the structure (see [rvine_structure] for more
#' details).
#'
#' The [get_matrix] method (for `vinecop_dist` or `vine_dist` objects only)
#' extracts the structure matrix (see [rvine_structure] for more details).
#'
#' The other `get_xyz` methods for `vinecop_dist` or `vine_dist` objects return
#' the entries corresponding to the pair-copula indexed by its `tree` and
#' `edge`. When `object` is of class `bicop_dist`, `tree` and `edge` are not
#' required. \cr
#'
#' * [get_pair_copula()] = the pair-copula itself (see [bicop]).
#'
#' * [get_parameters()] = the parameters of the pair-copula (i.e., a `numeric`
#' scalar, vector, or matrix).
#'
#' * [get_family()] = a character for the family (see [bicop] for
#' implemented families).
#'
#' * [get_ktau()] = a `numeric` scalar with the pair-copula Kendall's tau.
#'
#' The `get_all_xyz` methods (for `vinecop_dist` or `vine_dist` objects only)
#' return lists of lists, with each element corresponding to a tree in `trees`,
#' and then elements of the sublists correspond to edges. The returned lists
#' have two additional attributes:
#'
#' * `"d"` = the dimension of the model.
#'
#' * `"trees"` = the extracted trees.
#'
#' @examples
#' # specify pair-copulas
#' bicop <- bicop_dist("bb1", 90, c(3, 2))
#' pcs <- list(
#'   list(bicop, bicop), # pair-copulas in first tree
#'   list(bicop) # pair-copulas in second tree
#' )
#'
#' # specify R-vine matrix
#' mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
#'
#' # set up vine copula model
#' vc <- vinecop_dist(pcs, mat)
#'
#' # get the structure
#' get_structure(vc)
#' all(get_matrix(vc) == mat)
#'
#' # get pair-copulas
#' get_pair_copula(vc, 1, 1)
#' get_all_pair_copulas(vc)
#' all.equal(get_all_pair_copulas(vc), pcs,
#'           check.attributes = FALSE, check.environment = FALSE)
#' @return
#' The structure matrix, or pair-copulas, their parameters, Kendall's taus,
#' or families.
#' @rdname getters
#' @export
get_structure <- function(object) {
  assert_that(inherits(object, "vinecop_dist") ||
    inherits(object, "vine_dist"))
  if (inherits(object, "vinecop_dist")) {
    return(object$structure)
  } else {
    return(object$copula$structure)
  }
}

#' @export
get_matrix <- function(object) {
  as_rvine_matrix(get_structure(object))
}


#' @param tree tree index (not required if `object` is of class `bicop_dist`).
#' @param edge edge index (not required if `object` is of class `bicop_dist`).
#'
#' @rdname getters
#' @export
get_pair_copula <- function(object, tree = NA, edge = NA) {

  ## sanity checks
  assert_that(inherits(object, "bicop_dist") ||
    inherits(object, "vinecop_dist") ||
    inherits(object, "vine_dist"))

  if (inherits(object, "bicop_dist")) {
    if (!is.scalar(tree) || !is.na(tree)) {
      warning("tree argument not used for bicop_dist objects")
    }
    if (!is.scalar(edge) || !is.na(edge)) {
      warning("edge argument not used for bicop_dist objects")
    }
    return(object)
  } else {
    d <- dim(object)
    assert_that(is.number(tree),
      tree >= 1,
      tree <= d[2],
      msg = "tree should be a number between 1 and
                    the truncation level."
    )
    assert_that(is.number(edge),
      edge >= 1,
      edge <= d[1] - tree,
      msg = "tree should be a number between 1 and
                    dimension minus tree."
    )

    ## return pair-copula
    if (inherits(object, "vinecop_dist")) {
      return(object$pair_copulas[[tree]][[edge]])
    } else {
      return(object$copula$pair_copulas[[tree]][[edge]])
    }
  }
}

#' @rdname getters
#' @export
get_parameters <- function(object, tree = NA, edge = NA) {
  coef(get_pair_copula(object, tree, edge))
}

#' @rdname getters
#' @export
get_ktau <- function(object, tree = NA, edge = NA) {
  par_to_ktau(get_pair_copula(object, tree, edge))
}

#' @rdname getters
#' @export
get_family <- function(object, tree = NA, edge = NA) {
  get_pair_copula(object, tree, edge)$family
}

#' @param trees the trees to extract from `object` (`trees = NA` extracts all
#' trees).
#' @rdname getters
#' @export
get_all_pair_copulas <- function(object, trees = NA) {
  assert_that(inherits(object, "vinecop_dist") ||
    inherits(object, "vine_dist"))
  d <- dim(object)

  if (inherits(object, "vinecop_dist")) {
    pcs <- object$pair_copulas
  } else {
    pcs <- object$copula$pair_copulas
  }

  if (!any(is.na(trees))) {
    assert_that(is.numeric(trees),
      all(trees >= 1),
      all(trees <= d[1] - 1),
      msg = "the elements of trees should be numbers between 1 and d - 1."
    )
  }

  t <- length(pcs)
  if (any(is.na(trees))) {
    trees <- seq_len(t)
  } else {
    if (any(trees > t)) {
      warning(
        "vine copula is ", t, "-truncated; ",
        "only returning available trees."
      )
      trees <- trees[trees <= t]
    }
  }
  res <- structure(pcs[trees], class = c("rvine_list", "pair_copulas"))
  attr(res, "d") <- d[1]
  attr(res, "trunc_lvl") <- d[2]
  attr(res, "trees") <- trees
  return(res)
}

get_all_xyz <- function(object, trees, func, list_name) {
  pcs <- get_all_pair_copulas(object, trees)
  res <- structure(lapply(pcs, function(tree) lapply(tree, func)),
    class = c("rvine_list", list_name)
  )
  attr(res, "d") <- attr(pcs, "d")
  attr(res, "trees") <- attr(pcs, "trees")
  return(res)
}

#' @rdname getters
#' @export
get_all_parameters <- function(object, trees = NA) {
  get_all_xyz(object, trees, coef, "parameters")
}

#' @rdname getters
#' @export
get_all_ktaus <- function(object, trees = NA) {
  get_all_xyz(object, trees, par_to_ktau, "ktaus")
}

#' @rdname getters
#' @export
get_all_families <- function(object, trees = NA) {
  get_all_xyz(object, trees, function(pc) pc$family, "families")
}

#' @export
print.rvine_list <- function(x, ...) {
  what <- switch(class(x)[2],
    pair_copulas = "pair-copulas",
    parameters = "copula parameters",
    ktaus = "Kendall's taus",
    families = "copula families"
  )

  d <- attr(x, "d")
  trees <- attr(x, "trees")
  msg <- paste(
    "Nested list of lists for the", what, "of a",
    d, "dimensional vine with"
  )

  ntrees <- length(trees)
  if (ntrees == d - 1) {
    cat(paste(msg, "all trees: \n"))
  } else if (ntrees == 1) {
    cat(paste(msg, " tree ", trees, ". \n", sep = ""))
  } else {
    trees_print <- paste(
      paste(trees[1:(ntrees - 1)], collapse = ", "),
      "and", trees[ntrees]
    )
    cat(paste(msg, " trees ", trees_print, ". \n", sep = ""))
  }
  cat(rep("-", 50), "\n")
  arg <- deparse(substitute(x))
  for (t in seq_along(trees)) {
    if (t > 10) {
      if (length(trees) - 10 > 1) {
        trees_print <- "more trees.\n"
      } else {
        trees_print <- "more tree.\n"
      }
      cat("# ... with", length(trees) - 10, trees_print)
      break
    }
    cat(paste(arg, "[[", t, "]] -> a list with the ", d - trees[t],
      " ", what, " of tree ", trees[t], ". \n",
      sep = ""
    ))
  }
  invisible(x)
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Class 'vinecop_dist'")

set.seed(0)
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(list(bicop, bicop), list(bicop))
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
vc <- vinecop_dist(pcs, mat)


test_that("constructor creates proper vinecop_dist object", {
  expect_s3_class(vc, "vinecop_dist")
  expect_identical(
    names(vc),
    c("pair_copulas", "structure", "var_types", "npars", "loglik")
  )
})


test_that("d/p/r- functions work", {
  u <- rvinecop(50, vc)
  expect_false(any(rvinecop(50, vc, qrng = FALSE) ==
    rvinecop(50, vc, qrng = FALSE)))
  set.seed(1)
  u <- rvinecop(50, vc, qrng = TRUE)
  set.seed(1)
  expect_true(all(u == rvinecop(50, vc, qrng = TRUE)))
  expect_gte(min(dvinecop(u, vc)), 0)
  expect_gte(min(pvinecop(u, vc, 100)), 0)
  expect_lte(max(pvinecop(u, vc, 100)), 1)
})


test_that("constructor catches wrong input", {
  # wrong number of pcs
  pcs2 <- pcs
  pcs2[[1]][[2]] <- NULL
  expect_error(vinecop_dist(pcs[-1], mat))

  # not all pcs are of class 'bicop_dist'
  pcs2[[1]][[2]] <- list(this = "stupid")
  expect_error(vinecop_dist(pcs2, mat))

  # wrong R-vine matrix
  mat[3, 3] <- 5
  expect_error(vinecop_dist(pcs, mat))
})

test_that("works with truncated vines", {
  # takes and returns truncated pair_copulas list
  trunc_vine <- vinecop_dist(pcs[-2], mat)
  expect_length(trunc_vine$pair_copulas, 1)

  # summary table is truncated too
  expect_s3_class(summary(vinecop_dist(pcs[-2], mat)), "summary_df")
  expect_silent(smr <- summary(vinecop_dist(pcs[-2], mat)))
  expect_eql(nrow(smr), 2)
})

test_that("print/summary/dim generics work", {
  expect_output(print(vc))
  expect_s3_class(summary(vc), "summary_df")
  expect_silent(s <- summary(vc))
  expect_is(s, "data.frame")
  expect_eql(nrow(s), 3)
  expect_eql(ncol(s), 10)
  expect_equiv(dim(vc), c(3, 2))
})

test_that("plot functions work", {
  pcs <- lapply(1:4, function(j) # pair-copulas in tree j
    lapply(runif(5 - j), function(cor) bicop_dist("gaussian", 0, cor)))
  mat <- matrix(
    c(
      1, 2, 3, 4, 5,
      1, 2, 3, 4, 0,
      1, 2, 3, 0, 0,
      1, 2, 0, 0, 0,
      1, 0, 0, 0, 0
    ),
    5, 5
  )
  vc <- vinecop_dist(pcs, mat)

  # we could check some values in the plot objects
  expect_silent(p <- plot(vc, edge_labels = "family", var_names = "legend"))
  expect_silent(p <- plot(vc, edge_labels = "tau", var_names = "use"))
  expect_silent(p <- plot(vc, edge_labels = "pair"))
  expect_silent(p <- plot(vc, edge_labels = "family_tau"))
  expect_silent(p <- plot(vc, var_names = "hide"))
  expect_error(p <- plot(vc, edge_labels = "no"))
  expect_error(p <- plot(vc, var_names = "isaidno"))
  expect_error(p <- plot(vc, tree = 10))
  expect_silent(p <- plot(vc, "ALL"))
  expect_silent(p <- contour(vc, xlim = c(0.2, 0.8), ylim = c(0.2, 0.8)))
  expect_silent(p <- contour(vc, margins = "unif"))
  expect_error(p <- contour(vc, margins = "nonono"))
  expect_error(p <- contour(vc, var_names = "comeon"))

  # contour for truncated vines
  vc$pair_copulas[[4]] <- NULL
  expect_silent(p <- contour(vc, margins = "unif"))
})

test_that("getters work", {

  # test get_structure
  expect_silent(pcc <- get_structure(vc))
  expect_error(get_structure(12))

  # test get_matrix
  expect_equiv(as_rvine_matrix(mat), get_matrix(vc))
  expect_error(get_matrix(12))

  # test get_pair_copulas
  expect_silent(pcc <- get_pair_copula(vc, 1, 1))
  expect_eql(bicop, bicop_dist(pcc$family, pcc$rotation, pcc$parameters))
  expect_error(get_pair_copula(12, 1, 1))
  expect_error(get_pair_copula(vc, 1:2, 1))
  expect_error(get_pair_copula(vc, 1, 1:2))
  expect_error(get_pair_copula(vc, 0, 1))
  expect_error(get_pair_copula(vc, 1, 0))
  expect_error(get_pair_copula(vc, 12, 1))
  expect_error(get_pair_copula(vc, 1, 12))

  # test get_all_pair_copulas
  expect_equiv(pcs, get_all_pair_copulas(vc))
  expect_equiv(pcs[1:2], get_all_pair_copulas(vc, 1:2))
  expect_error(get_all_pair_copulas(12))
  expect_error(get_all_pair_copulas(vc, 0))
  expect_error(get_all_pair_copulas(vc, 12))

  # test other getters
  expect_equiv(get_parameters(vc, 1, 1), coef(pcs[[1]][[1]]))
  expect_equiv(
    get_all_parameters(vc),
    lapply(pcs, function(tree) lapply(tree, coef))
  )
  expect_equiv(get_ktau(vc, 1, 1), par_to_ktau(bicop))
  expect_equiv(
    get_all_ktaus(vc),
    lapply(pcs, function(tree)
      lapply(tree, function(pc) par_to_ktau(pc)))
  )
  expect_equiv(get_family(vc, 1, 1), "bb1")
  expect_equiv(
    get_all_families(vc),
    lapply(pcs, function(tree)
      lapply(tree, function(pc) pc$family))
  )

  # test printed output of getters
  expect_output(print(get_all_pair_copulas(vc)))
  expect_output(print(get_all_pair_copulas(vc, 1)))
})


test_that("d = 1 works", {
  vc <- vinecop_dist(list(), rvine_structure(1))
  u <- runif(5)
  expect_eql(dim(summary(vc))[1], 0)

  expect_eql(dvinecop(u, vc), rep(1, 5))
  expect_eql(pvinecop(u, vc), u, tol = 1e-2)
  expect_eql(c(rosenblatt(u, vc)), u)
  expect_eql(c(inverse_rosenblatt(u, vc)), u)
  expect_silent(rvinecop(10, vc))
  expect_error(plot(vc))
  expect_warning(contour(vc))
})

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Class 'vine_dist'")

set.seed(0)
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(list(bicop, bicop), list(bicop))
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
vc <- vine_dist(list(distr = "norm"), pcs, mat)

test_that("constructor creates proper `vine_dist` object", {
  expect_s3_class(vc, "vine_dist")
  expect_identical(names(vc), c("margins", "copula", "npars", "loglik"))
})


test_that("d/p/r- functions work", {
  u <- rvine(50, vc)
  expect_false(any(rvine(50, vc, qrng = FALSE) == rvine(50, vc, qrng = FALSE)))
  set.seed(1)
  u <- rvine(50, vc, qrng = TRUE)
  set.seed(1)
  expect_true(all(u == rvine(50, vc, qrng = TRUE)))
  expect_gte(min(dvine(u, vc)), 0)
  expect_gte(min(pvine(u, vc, 100)), 0)
  expect_lte(max(pvine(u, vc, 100)), 1)
})


test_that("constructor catches wrong input", {
  # missing margin name
  expect_error(vine_dist(list(stupid = "norm"), cop))

  # unused margin argument
  expect_error(vine_dist(list(distr = "norm", stupid = 42), cop))

  # missing margin argument
  expect_error(vine_dist(list(distr = "beta", scale1 = 1), cop))

  # length of margins vector do not correspond to cop
  expect_error(vine_dist(list(
    list(distr = "norm"),
    list(distr = "gamma", shape = 1)
  ), cop))
})

test_that("print/summary/dim generics work", {
  expect_output(print(vc))

  s <- summary(vc)
  expect_is(s$margins, "data.frame")
  expect_is(s$copula, "data.frame")
  expect_eql(nrow(s$margins), 3)
  expect_eql(ncol(s$margins), 2)
  expect_eql(nrow(s$copula), 3)
  expect_eql(ncol(s$copula), 10)

  expect_equiv(dim(vc)[1], 3)
  expect_equiv(dim(vc)[2], 2)
})

test_that("getters work", {

  # test get_matrix
  expect_equiv(as_rvine_matrix(mat), get_matrix(vc))
  expect_error(get_matrix(12))

  # test get_pair_copulas
  expect_silent(pcc <- get_pair_copula(vc, 1, 1))
  expect_eql(bicop, bicop_dist(pcc$family, pcc$rotation, pcc$parameters))
  expect_error(get_pair_copula(12, 1, 1))
  expect_error(get_pair_copula(vc, 1:2, 1))
  expect_error(get_pair_copula(vc, 1, 1:2))
  expect_error(get_pair_copula(vc, 0, 1))
  expect_error(get_pair_copula(vc, 1, 0))
  expect_error(get_pair_copula(vc, 12, 1))
  expect_error(get_pair_copula(vc, 1, 12))

  # test get_all_pair_copulas
  expect_equiv(pcs, get_all_pair_copulas(vc))
  expect_equiv(pcs[1:2], get_all_pair_copulas(vc, 1:2))
  expect_error(get_all_pair_copulas(12))
  expect_error(get_all_pair_copulas(vc, 0))
  expect_error(get_all_pair_copulas(vc, 12))

  # test other getters
  expect_equiv(get_parameters(vc, 1, 1), coef(pcs[[1]][[1]]))
  expect_equiv(
    get_all_parameters(vc),
    lapply(pcs, function(tree) lapply(tree, coef))
  )
  expect_equiv(get_ktau(vc, 1, 1), par_to_ktau(bicop))
  expect_equiv(
    get_all_ktaus(vc),
    lapply(pcs, function(tree)
      lapply(tree, function(pc) par_to_ktau(pc)))
  )
  expect_equiv(get_family(vc, 1, 1), "bb1")
  expect_equiv(
    get_all_families(vc),
    lapply(pcs, function(tree)
      lapply(tree, function(pc) pc$family))
  )

  # test printed output of getters
  expect_output(print(get_all_pair_copulas(vc)))
  expect_output(print(get_all_pair_copulas(vc, 1)))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

pseudo_obs_cpp <- function(x, ties_method) {
    .Call(`_rvinecopulib_pseudo_obs_cpp`, x, ties_method)
}

bicop_check_cpp <- function(bicop_r) {
    invisible(.Call(`_rvinecopulib_bicop_check_cpp`, bicop_r))
}

bicop_select_cpp <- function(data, family_set, par_method, nonpar_method, mult, selcrit, weights, psi0, presel, num_threads, var_types) {
    .Call(`_rvinecopulib_bicop_select_cpp`, data, family_set, par_method, nonpar_method, mult, selcrit, weights, psi0, presel, num_threads, var_types)
}

bicop_pdf_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_pdf_cpp`, u, bicop_r)
}

bicop_cdf_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_cdf_cpp`, u, bicop_r)
}

bicop_hfunc1_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_hfunc1_cpp`, u, bicop_r)
}

bicop_hfunc2_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_hfunc2_cpp`, u, bicop_r)
}

bicop_hinv1_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_hinv1_cpp`, u, bicop_r)
}

bicop_hinv2_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_hinv2_cpp`, u, bicop_r)
}

bicop_sim_cpp <- function(bicop_r, n, qrng, seeds) {
    .Call(`_rvinecopulib_bicop_sim_cpp`, bicop_r, n, qrng, seeds)
}

bicop_loglik_cpp <- function(u, bicop_r) {
    .Call(`_rvinecopulib_bicop_loglik_cpp`, u, bicop_r)
}

bicop_par_to_tau_cpp <- function(bicop_r) {
    .Call(`_rvinecopulib_bicop_par_to_tau_cpp`, bicop_r)
}

bicop_tau_to_par_cpp <- function(bicop_r, tau) {
    .Call(`_rvinecopulib_bicop_tau_to_par_cpp`, bicop_r, tau)
}

rvine_structure_cpp <- function(rvine_structure_r, check, is_natural_order) {
    .Call(`_rvinecopulib_rvine_structure_cpp`, rvine_structure_r, check, is_natural_order)
}

rvine_structure_check_cpp <- function(rvine_struct, is_natural_order) {
    invisible(.Call(`_rvinecopulib_rvine_structure_check_cpp`, rvine_struct, is_natural_order))
}

rvine_structure_sim_cpp <- function(d, natural_order, seeds) {
    .Call(`_rvinecopulib_rvine_structure_sim_cpp`, d, natural_order, seeds)
}

rvine_matrix_check_cpp <- function(matrix) {
    invisible(.Call(`_rvinecopulib_rvine_matrix_check_cpp`, matrix))
}

vinecop_check_cpp <- function(vinecop_r) {
    invisible(.Call(`_rvinecopulib_vinecop_check_cpp`, vinecop_r))
}

vinecop_inverse_rosenblatt_cpp <- function(U, vinecop_r, cores) {
    .Call(`_rvinecopulib_vinecop_inverse_rosenblatt_cpp`, U, vinecop_r, cores)
}

vinecop_rosenblatt_cpp <- function(U, vinecop_r, cores) {
    .Call(`_rvinecopulib_vinecop_rosenblatt_cpp`, U, vinecop_r, cores)
}

vinecop_sim_cpp <- function(vinecop_r, n, qrng, cores, seeds) {
    .Call(`_rvinecopulib_vinecop_sim_cpp`, vinecop_r, n, qrng, cores, seeds)
}

vinecop_pdf_cpp <- function(u, vinecop_r, cores) {
    .Call(`_rvinecopulib_vinecop_pdf_cpp`, u, vinecop_r, cores)
}

vinecop_cdf_cpp <- function(u, vinecop_r, N, cores, seeds) {
    .Call(`_rvinecopulib_vinecop_cdf_cpp`, u, vinecop_r, N, cores, seeds)
}

vinecop_loglik_cpp <- function(u, vinecop_r, cores) {
    .Call(`_rvinecopulib_vinecop_loglik_cpp`, u, vinecop_r, cores)
}

vinecop_mbicv_cpp <- function(u, vinecop_r, psi0, cores) {
    .Call(`_rvinecopulib_vinecop_mbicv_cpp`, u, vinecop_r, psi0, cores)
}

vinecop_select_cpp <- function(data, structure, family_set, par_method, nonpar_method, mult, truncation_level, tree_criterion, threshold, selection_criterion, weights, psi0, select_truncation_level, select_threshold, preselect_families, show_trace, num_threads, var_types) {
    .Call(`_rvinecopulib_vinecop_select_cpp`, data, structure, family_set, par_method, nonpar_method, mult, truncation_level, tree_criterion, threshold, selection_criterion, weights, psi0, select_truncation_level, select_threshold, preselect_families, show_trace, num_threads, var_types)
}

fit_margins_cpp <- function(data, nlevels, mult, xmin, xmax, bw, deg, weights, num_threads) {
    .Call(`_rvinecopulib_fit_margins_cpp`, data, nlevels, mult, xmin, xmax, bw, deg, weights, num_threads)
}

#' Vine based distributions
#'
#' Density, distribution function and random generation
#' for the vine based distribution.
#'
#' @name vine_distributions
#' @aliases dvine pvine rvine dvine_dist pvine_dist rvine_dist
#' @param x evaluation points, either a length d vector or a d-column matrix,
#'   where d is the number of variables in the vine.
#' @param vine an object of class `"vine_dist"`.
#' @param cores number of cores to use; if larger than one, computations are
#'   done in parallel on `cores` batches .
#' @details
#' See [vine] for the estimation and construction of vine models.
#' Here, the density, distribution function and random generation
#' for the vine distributions are standard.
#'
#' The functions are based on [dvinecop()], [pvinecop()] and [rvinecop()] for
#' [vinecop] objects, and either [kde1d::dkde1d()], [kde1d::pkde1d()] and
#' [kde1d::qkde1d()] for estimated vines (i.e., output of [vine()]), or the
#' standard *d/p/q-xxx* from [stats::Distributions] for custom vines
#' (i.e., output of [vine_dist()]).
#' @return
#' `dvine()` gives the density, `pvine()` gives the distribution function,
#' and `rvine()` generates random deviates.
#'
#' The length of the result is determined by `n` for `rvine()`, and
#' the number of rows in `u` for the other functions.
#'
#' The `vine` object is recycled to the length of the
#' result.
#' @examples
#' # specify pair-copulas
#' bicop <- bicop_dist("bb1", 90, c(3, 2))
#' pcs <- list(
#'   list(bicop, bicop), # pair-copulas in first tree
#'   list(bicop) # pair-copulas in second tree
#' )
#'
#' # set up vine copula model
#' mat <- rvine_matrix_sim(3)
#' vc <- vine_dist(list(distr = "norm"), pcs, mat)
#'
#' # simulate from the model
#' x <- rvine(200, vc)
#' pairs(x)
#'
#' # evaluate the density and cdf
#' dvine(x[1, ], vc)
#' pvine(x[1, ], vc)
#' @rdname vine_methods
#' @export
dvine <- function(x, vine, cores = 1) {
  stopifnot(inherits(vine, "vine_dist"))
  if (NCOL(x) == 1) {
    x <- t(x)
  }

  x <- expand_factors(x)
  if (!is.null(vine$names)) {
    x <- x[, vine$names, drop = FALSE]
  }

  # prepare marginals if only one is specified
  d <- ncol(x)
  if (!inherits(vine, "vine") & depth(vine$margins) == 1) {
    vine$margins <- replicate(d, vine$margins, simplify = FALSE)
  }

  ## evaluate marginal densities
  margvals <- dpq_marg(x, vine, "d")

  if (!is.null(vine$copula)) {
    u <- compute_pseudo_obs(x, vine)
    vinevals <- dvinecop(u, vine$copula, cores)
  } else {
    vinevals <- rep(1, nrow(x))
  }

  ## final density estimate is product of marginals and copula density
  apply(cbind(margvals, vinevals), 1, prod)
}

#' @rdname vine_methods
#' @param n_mc number of samples used for quasi Monte Carlo integration.
#' @export
pvine <- function(x, vine, n_mc = 10^4, cores = 1) {
  stopifnot(inherits(vine, "vine_dist"))

  if (NCOL(x) == 1) {
    x <- t(x)
  }
  x <- expand_factors(x)
  if (!is.null(vine$names)) {
    x <- x[, vine$names, drop = FALSE]
  }

  # prepare marginals if only one is specified
  if (!inherits(vine, "vine") & depth(vine$margins) == 1) {
    vine$margins <- replicate(ncol(x), vine$margins, simplify = FALSE)
  }

  # PIT to copula data
  u <- compute_pseudo_obs(x, vine)

  # Evaluate copula if needed
  if (!is.null(vine$copula)) {
    vals <- pvinecop(u, vine$copula, n_mc, cores)
  } else {
    vals <- apply(u, 1, prod)
  }

  vals
}

#' @rdname vine_methods
#' @param n number of observations.
#' @param qrng if `TRUE`, generates quasi-random numbers using the multivariate
#' Generalized Halton sequence up to dimension 300 and the Generalized Sobol
#' sequence in higher dimensions (default `qrng = FALSE`).
#' @export
rvine <- function(n, vine, qrng = FALSE, cores = 1) {
  assert_that(inherits(vine, "vine_dist"), is.flag(qrng))

  # simulate copula data
  U <- rvinecop(n, vine$copula, qrng, cores)

  # prepare marginals if only one is specified
  if (!inherits(vine, "vine") & depth(vine$margins) == 1) {
    vine$margins <- replicate(dim(vine)[1], vine$margins, simplify = FALSE)
  }

  # use quantile transformation for marginals
  X <- dpq_marg(U, vine, "q")
  colnames(X) <- vine$names
  X
}

#' @export
print.vine_dist <- function(x, ...) {
  cat(dim(x)[1], "-dimensional vine distribution model ('vine_dist')", sep = "")
  print_truncation_info(x$copula)
  invisible(x)
}

#' @export
summary.vine_dist <- function(object, ...) {
  list(
    margins = get_vine_dist_margin_summary(object),
    copula = summary(object$copula, ...)
  )
}

get_vine_dist_margin_summary <- function(vd) {
  margins <- vd$margins
  if (length(margins) == 1) {
    margins <- rep(list(margins), dim(vd$copula)[1])
  }
  df <- data.frame(
    margin = seq_along(margins),
    distr = sapply(margins, function(x) x$distr)
  )
  class(df) <- c("summary_df", class(df))
  df
}

#' Predictions and fitted values for a vine copula model
#'
#' Predictions of the density and distribution function
#' for a vine copula model.
#'
#' @name vine_predict_and_fitted
#' @aliases fitted.vine predict.vine
#' @param object a `vine` object.
#' @param newdata points where the fit shall be evaluated.
#' @param what what to predict, either `"pdf"` or `"cdf"`.
#' @param n_mc number of samples used for quasi Monte Carlo integration when
#'    `what = "cdf"`.
#' @param cores number of cores to use; if larger than one, computations are
#'   done in parallel on `cores` batches .
#' @param ... unused.
#'
#' @return
#' `fitted()` and `predict()` have return values similar to [dvine()]
#' and [pvine()].
#' @export
#' @rdname predict_vine
#' @examples
#' x <- sapply(1:5, function(i) rnorm(50))
#' fit <- vine(x, copula_controls = list(family_set = "par"), keep_data = TRUE)
#' all.equal(predict(fit, x), fitted(fit), check.environment = FALSE)
predict.vine <- function(object, newdata, what = "pdf", n_mc = 10^4,
                         cores = 1, ...) {
  stopifnot(what %in% c("pdf", "cdf"))
  switch(
    what,
    "pdf" = dvine(newdata, object, cores),
    "cdf" = pvine(newdata, object, n_mc, cores)
  )
}

#' @rdname predict_vine
#' @export
fitted.vine <- function(object, what = "pdf", n_mc = 10^4, cores = 1, ...) {
  if (all(is.na(object$data))) {
    stop("data have not been stored, use keep_data = TRUE when fitting.")
  }
  stopifnot(what %in% c("pdf", "cdf"))
  switch(
    what,
    "pdf" = dvine(object$data, object, cores),
    "cdf" = pvine(object$data, object, n_mc, cores)
  )
}

#' @export
logLik.vine <- function(object, ...) {
  structure(object$loglik, "df" = object$npars)
}

#' @export
print.vine <- function(x, ...) {
  cat(dim(x)[1], "-dimensional vine distribution fit ('vine')", sep = "")
  print_truncation_info(x$copula)
  print_fit_info(x)
  invisible(x)
}

#' @export
summary.vine <- function(object, ...) {
  list(
    margins = get_vine_margin_summary(object),
    copula = summary(object$copula)
  )
}

get_vine_margin_summary <- function(object) {
  capture.output(info <- sapply(object$margins, summary))
  info <- as.data.frame(t(info))
  info <- cbind(
    data.frame(margin = seq_len(nrow(info)), name = object$names),
    info
  )
  class(info) <- c("summary_df", "data.frame")
  info
}


dpq_marg <- function(x, vine, what = "p") {
  d <- ncol(x)
  res <- lapply(seq_len(d),
                function(i) eval_one_dpq(x[, i], vine$margins[[i]], what))
  do.call(cbind, res)
}

eval_one_dpq <- function(x, margin, what = "p") {
  if (inherits(margin, "kde1d")) {
    dpq <- switch(what,
                  p = pkde1d(x, margin),
                  d = dkde1d(x, margin),
                  q = qkde1d(x, margin))
  } else {
    par <- margin[names(margin) != "distr"]
    par[[length(par) + 1]] <- x
    names(par)[[length(par)]] <- switch(what,
                                        p = "q",
                                        d = "x",
                                        q = "p")
    dpq <- do.call(get(paste0(what, margin$distr)), par)
  }
  if (is.factor(dpq))
    dpq <- as.data.frame(dpq)
  dpq
}

collate_u <- function(x) {
  if (!all(is.na(x$data))) {
    u <- dpq_marg(x$data, x)
    x$copula$data <- u
  }
  x
}

#' @export
dim.vine_dist <- function(x) {
  dim(x$copula)
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Fitting 'vine' models")

set.seed(5)
u <- sapply(1:5, function(i) rnorm(30))
fit <- vine(u, copula_controls = list(family_set = "nonpar"), keep_data = TRUE)

test_that("returns proper 'vine' object", {
  expect_s3_class(fit, "vine")
  expect_s3_class(fit, "vine_dist")
  expect_identical(
    names(fit),
    c(
      "margins", "margins_controls", "copula", "copula_controls",
      "npars", "loglik", "data", "weights", "nobs", "names", "var_levels"
    )
  )
})

test_that("S3 generics work", {
  expect_eql(predict(fit, u), fitted(fit))
  expect_eql(
    predict(fit, u, what = "cdf"),
    fitted(fit, what = "cdf"),
    tolerance = 0.01
  )
  expect_error(predict(fit, u, what = "hfunc1"))
  expect_length(attr(logLik(fit), "df"), 1)
  expect_length(predict(fit, u[1, ], what = "pdf"), 1)
})

test_that("print/summary generics work", {
  expect_output(print(fit))
  s <- summary(fit)
  expect_is(s$margins, c("summary_df", "data.frame"))
  expect_is(s$copula, c("summary_df", "data.frame"))
})

test_that("truncation works", {
  fit_truncated <- truncate_model(fit, trunc_lvl = 1)
  expect_silent(dvine(u, fit_truncated))
  expect_silent(rvine(50, fit_truncated))

  fit_truncated <- vine(u, copula_controls = list(
    family_set = "nonpar",
    trunc_lvl = 1
  ))
  expect_silent(dvine(u, fit_truncated))
  expect_silent(rvine(50, fit_truncated))
})

test_that("margins_controls works", {
  fit_mult <- vine(u, margins_controls = list(mult = 2))
  expect_eql(
    sapply(fit_mult$margins, "[[", "bw"),
    2 / log(6) * sapply(fit$margins, "[[", "bw")
  )

  fit_xmin <- vine(abs(u), margins_controls = list(xmin = 0, deg = 1, mult = 1:5))
  expect_eql(sapply(fit_xmin$margins, "[[", "xmin"), rep(0, 5))
  expect_eql(sapply(fit_xmin$margins, "[[", "deg"), rep(1, 5))
})

test_that("weights work", {
  w <- rexp(nrow(u))
  fit_weights <- vine(u, copula_controls = list(family_set = "nonpar"),
                     weights = w, keep_data = TRUE)
  expect_eql(fit_weights$weights, w)
  expect_false(identical(fit$margins[[1]], fit_weights$margins[[1]]))
})

test_that("d = 1 works", {
  vc <- vine(runif(20))
  expect_eql(dim(summary(vc)$margins)[1], 1)
  expect_eql(dim(summary(vc)$copula)[1], 0)
})

#' Fit and select bivariate copula models
#'
#' Fit a bivariate copula model for continuous or discrete data. The family
#' can be selected automatically from a vector of options.
#'
#' @param data a matrix or data.frame with at least two columns, containing the
#'   (pseudo-)observations for the two variables (copula data should have
#'   approximately uniform margins). More columns are required for discrete
#'   models, see *Details*.
#' @param family_set a character vector of families; see *Details* for
#'   additional options.
#' @param par_method the estimation method for parametric models, either `"mle"`
#'   for maximum likelihood or `"itau"` for inversion of Kendall's tau (only
#'   available for one-parameter families and `"t"`.
#' @param nonpar_method the estimation method for nonparametric models, either
#'   `"constant"` for the standard transformation estimator, or
#'   `"linear"`/`"quadratic"` for the local-likelihood approximations of order
#'   one/two.
#' @param mult multiplier for the smoothing parameters of nonparametric
#'   families. Values larger than 1 make the estimate more smooth, values less
#'   than 1 less smooth.
#' @param selcrit criterion for family selection, either `"loglik"`, `"aic"`,
#'   `"bic"`, `"mbic"`. For `vinecop()` there is the additional option
#'   `"mbicv"`.
#' @param weights optional vector of weights for each observation.
#' @param psi0 see [mBICV()].
#' @param presel whether the family set should be thinned out according to
#'   symmetry characteristics of the data.
#' @param keep_data whether the data should be stored (necessary for using
#'   [fitted()]).
#' @param cores number of cores to use; if more than 1, estimation for multiple
#'   families is done in parallel.
#' @param var_types variable types, a length 2 vector; e.g., `c("c", "c")` for
#'   both continuous (default), or `c("c", "d")` for first variable continuous
#'   and second discrete.
#'
#' @details
#'
#' If there are missing data (i.e., `NA` entries), incomplete observations are
#' discarded before fitting the copula.
#'
#' ## Discrete variables
#'
#' When at least one variable is discrete, more than two columns are required
#' for `data`: the first \eqn{n \times 2} block contains realizations of
#' \eqn{F_{X_1}(x_1), F_{X_2}(x_2)}. The second \eqn{n \times 2} block contains
#' realizations of \eqn{F_{X_1}(x_1^-), F_{X_1}(x_1^-)}. The minus indicates a
#' left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
#' \eqn{F_{X_1}(x_1^-) = F_{X_1}(x_1 - 1)}. For continuous variables the left
#' limit and the cdf itself coincide. Respective columns can be omitted in the
#' second block.
#'
#' ## Family collections
#'
#' The `family_set` argument accepts all families in `bicop_dist()` plus the
#' following convenience definitions:
#'
#' * `"all"` contains all the families,
#'
#' * `"parametric"` contains the parametric families (all except `"tll"`),
#'
#' * `"nonparametric"` contains the nonparametric families (`"indep"` and
#' `"tll"`)
#'
#' * `"onepar"` contains the parametric families with a single parameter,
#'
#' (`"gaussian"`, `"clayton"`, `"gumbel"`, `"frank"`, and `"joe"`),
#'
#' * `"twopar"` contains the parametric families with two parameters
#' (`"t"`, `"bb1"`, `"bb6"`, `"bb7"`, and `"bb8"`),
#'
#' * `"elliptical"` contains the elliptical families,
#'
#' * `"archimedean"` contains the archimedean families,
#'
#' * `"BB"` contains the BB families,
#'
#' * `"itau"` families for which estimation by Kendall's tau inversion is
#' available (`"indep"`,`"gaussian"`, `"t"`,`"clayton"`, `"gumbel"`,
#' `"frank"`, `"joe"`).
#'
#' @return
#' An object inheriting from classes `bicop` and  `bicop_dist` . In addition to
#' the entries contained in `bicop_dist()`, objects from the `bicop` class
#' contain:
#'
#' * `data` (optionally, if `keep_data = TRUE` was used), the dataset that was
#' passed to [bicop()].
#'
#' * `controls`, a `list` with the set of fit controls that was passed to
#' [bicop()].
#'
#' * `loglik` the log-likelihood.
#'
#' * `nobs`, an `integer` with the number of observations that was used to fit
#' the model.
#'
#' @seealso [bicop_dist()], [plot.bicop()], [contour.bicop()], [dbicop()],
#'   [pbicop()], [hbicop()], [rbicop()]
#'
#' @examples
#' ## fitting a continuous model from simulated data
#' u <- rbicop(100, "clayton", 90, 3)
#' fit <- bicop(u, family_set = "par")
#' summary(fit)
#'
#' ## compare fit with true model
#' contour(fit)
#' contour(bicop_dist("clayton", 90, 3), col = 2, add = TRUE)
#'
#' ## fit a model from discrete data
#' x_disc <- qpois(u, 1)  # transform to Poisson margins
#' plot(x_disc)
#' udisc <- cbind(ppois(x_disc, 1), ppois(x_disc - 1, 1))
#' fit_disc <- bicop(udisc, var_types = c("d", "d"))
#' summary(fit_disc)
#' @export
bicop <- function(data, var_types = c("c", "c"), family_set = "all",
                  par_method = "mle", nonpar_method = "quadratic", mult = 1,
                  selcrit = "aic", weights = numeric(), psi0 = 0.9,
                  presel = TRUE, keep_data = FALSE, cores = 1) {
  assert_that(
    is.character(family_set),
    is.string(par_method),
    is.string(nonpar_method),
    is.number(mult), mult > 0,
    is.string(selcrit),
    is.numeric(weights),
    is.number(psi0), psi0 > 0, psi0 < 1,
    is.flag(presel),
    is.flag(keep_data),
    is.number(cores), cores > 0,
    correct_var_types(var_types)
  )

  # check if families known (w/ partial matching) and expand convenience defs
  family_set <- process_family_set(family_set, par_method)

  ## fit and select copula model
  data <- if_vec_to_matrix(data)
  bicop <- bicop_select_cpp(
    data = data,
    family_set = family_set,
    par_method = par_method,
    nonpar_method = nonpar_method,
    mult = mult,
    selcrit = selcrit,
    weights = weights,
    psi0 = psi0,
    presel = presel,
    num_threads = cores,
    var_types = var_types
  )

  ## add information about the fit
  bicop$names <- colnames(data)
  if (keep_data) {
    bicop$data <- data
  }
  bicop$controls <- list(
    family_set = family_set,
    par_method = par_method,
    nonpar_method = nonpar_method,
    mult = mult,
    selcrit = selcrit,
    weights = weights,
    psi0 = psi0,
    presel = presel
  )
  bicop$nobs <- nrow(data)

  as.bicop(bicop)
}

as.bicop <- function(object) {
  if (!all(c("family", "rotation", "parameters", "npars") %in% names(object))) {
    stop("object cannot be coerced to class 'bicop'")
  }
  structure(object, class = c("bicop", "bicop_dist"))
}

#' Bivariate copula models
#'
#' Create custom bivariate copula models by specifying the family, rotation,
#' parameters, and variable types.
#'
#' ## Implemented families
#'
#' | type          | name                  | name in R     |
#' |---------------|-----------------------|---------------|
#' | -             | Independence          | "indep"       |
#' | Elliptical    | Gaussian              | "gaussian"    |
#' | "             | Student t             | "student"     |
#' | Archimedean   | Clayton               | "clayton"     |
#' | "             | Gumbel                | "gumbel"      |
#' | "             | Frank                 | "frank"       |
#' | "             | Joe                   | "joe"         |
#' | "             | Clayton-Gumbel (BB1)  | "bb1"         |
#' | "             | Joe-Gumbel (BB6)      | "bb6"         |
#' | "             | Joe-Clayton (BB7)     | "bb7"         |
#' | "             | Joe-Frank (BB8)       | "bb8"         |
#' | Nonparametric | Transformation kernel | "tll"         |
#'
#' @return
#'
#' An object of class `bicop_dist`, i.e., a list containing:
#'
#' * `family`, a `character` indicating the copula family.
#'
#' * `rotation`, an `integer` indicating the rotation (i.e., either 0, 90, 180,
#' or 270).
#'
#' * `parameters`, a `numeric` vector or matrix of parameters.
#'
#' * `npars`, a `numeric` with the (effective) number of parameters.
#'
#' * `var_types`, the variable types.
#'
#' @param family the copula family, a string containing the family name (see
#' *Details* for all possible families).
#' @param rotation the rotation of the copula, one of `0`, `90`, `180`, `270`.
#' @param parameters a vector or matrix of copula parameters.
#' @param var_types variable types, a length 2 vector; e.g., `c("c", "c")` for
#'   both continuous (default), or `c("c", "d")` for first variable continuous
#'   and second discrete.
#'
#' @seealso [bicop_dist()], [plot.bicop()], [contour.bicop()], [dbicop()],
#'   [pbicop()], [hbicop()], [rbicop()]
#' @export
#' @examples
#' ## Clayton 90° copula with parameter 3
#' cop <- bicop_dist("clayton", 90, 3)
#' cop
#' str(cop)
#'
#' ## visualization
#' plot(cop)
#' contour(cop)
#' plot(rbicop(200, cop))
#'
#' ## BB8 copula model for discrete data
#' cop_disc <- bicop_dist("bb8", 0, c(2, 0.5), var_types = c("d", "d"))
#' cop_disc
#'
bicop_dist <- function(family = "indep", rotation = 0, parameters = numeric(0),
                       var_types = c("c", "c")) {
  assert_that(is.string(family), is.number(rotation), is.numeric(parameters))
  assert_that(correct_var_types(var_types))
  if (family %in% setdiff(family_set_nonparametric, "indep")) {
    stop("bicop_dist should not be used directly with nonparametric families.")
  }

  family <- family_set_all[pmatch(family, family_set_all)]
  dist <- list(
    family = family,
    rotation = rotation,
    parameters = as.matrix(parameters),
    var_types = var_types,
    npars = length(parameters)
  )
  bicop_check_cpp(dist)
  structure(dist, class = "bicop_dist")
}

# fixes problems with change in all.equal() behavior in R 4.1.x
expect_eql <- function(...) expect_equal(..., check.environment = FALSE)
expect_equiv <- function(...) expect_equivalent(..., check.environment = FALSE)

context("Fitting 'bicop' models")

set.seed(0)
dist <- bicop_dist("gumbel", 90, 3)
u <- rbicop(20, dist)
fit <- bicop(u, keep_data = TRUE)

test_that("returns proper 'bicop' object", {
  expect_s3_class(fit, "bicop")
  expect_s3_class(fit, "bicop_dist")
  expect_identical(
    names(fit),
    c("family", "rotation", "parameters", "var_types",
      "npars", "loglik", "data", "controls", "nobs")
  )

  fit <- bicop(u, family = "tll", keep_data = FALSE)
  expect_s3_class(fit, "bicop")
  expect_s3_class(fit, "bicop_dist")
  expect_identical(
    names(fit),
    c("family", "rotation", "parameters", "var_types",
      "npars", "loglik", "controls", "nobs")
  )

  colnames(u) <- paste(1:2)
  expect_identical(
    names(bicop(u, family = "indep")),
    c("family", "rotation", "parameters", "var_types",
      "npars", "loglik", "names", "controls", "nobs")
  )
})

test_that("family sets (w/ partial matching)", {
  bicop(u, family = "arch")
  bicop(u, family = "nonp")
  bicop(u, family = "elli")
  bicop(u, family = "onep")
  bicop(u, family = "two")
  bicop(u, family = "bbs")
  bicop(u, family = "itau")
  expect_warning(bicop(u, par_method = "itau"))
  expect_error(bicop(u, family = "asdf"))
})

test_that("S3 generics work", {
  expect_eql(predict(fit, u, what = "pdf"), fitted(fit, what = "pdf"))
  expect_eql(predict(fit, u, what = "cdf"), fitted(fit, what = "cdf"))
  expect_eql(predict(fit, u, what = "hfunc1"), fitted(fit, what = "hfunc1"))
  expect_eql(predict(fit, u, what = "hfunc2"), fitted(fit, what = "hfunc2"))
  expect_eql(predict(fit, u, what = "hinv1"), fitted(fit, what = "hinv1"))
  expect_eql(predict(fit, u, what = "hinv2"), fitted(fit, what = "hinv2"))
  u <- as.data.frame(u)
  expect_equiv(logLik(fit), sum(log(predict(fit, u, what = "pdf"))))
  expect_output(print(fit))
  expect_output(summary(fit))
  expect_output(print(bicop(u, family = "nonp")))
})

#' Truncate a vine copula model
#'
#' Extracts a truncated sub-vine based on a truncation level supplied by user.
#'
#' While a vine model for a `d` dimensional random vector contains at most `d-1`
#' nested trees, this function extracts a sub-model based on a given truncation
#' level.
#'
#' For instance, `truncate_model(object, 1)` results in a 1-truncated
#' vine (i.e., a vine with a single tree). Similarly `truncate_model(object, 2)`
#' results in a 2-truncated vine (i.e., a vine with two trees). Note that
#' `truncate_model(truncate_model(object, 1), 2)` returns a 1-truncated vine.
#'
#' @param object a model object.
#' @param ... further arguments passed to specific methods.
#'
#' @export
#' @examples
#' # specify pair-copulas
#' bicop <- bicop_dist("bb1", 90, c(3, 2))
#' pcs <- list(
#'   list(bicop, bicop), # pair-copulas in first tree
#'   list(bicop) # pair-copulas in second tree
#' )
#'
#' # specify R-vine matrix
#' mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
#'
#' # set up vine structure
#' structure <- as_rvine_structure(mat)
#'
#' # truncate the model
#' truncate_model(structure, 1)
#'
#' # set up vine copula model
#' vc <- vinecop_dist(pcs, mat)
#'
#' # truncate the model
#' truncate_model(vc, 1)
truncate_model <- function(object, ...) {
  UseMethod("truncate_model")
}

#' @export
#' @param trunc_lvl tree level after which the vine copula should be truncated.
#' @rdname truncate_model
truncate_model.rvine_structure <- function(object, trunc_lvl, ...) {
  check_trunc_lvl(object, trunc_lvl)
  if (trunc_lvl < dim(object)["trunc_lvl"]) {
    object$struct_array <- object$struct_array[seq_len(trunc_lvl)]
    object$trunc_lvl <- trunc_lvl
  }
  object
}

#' @export
#' @rdname truncate_model
truncate_model.rvine_matrix <- function(object, trunc_lvl, ...) {
  check_trunc_lvl(object, trunc_lvl)
  d <- dim(object)["dim"]
  trees_to_truncate <- setdiff(seq_len(d - 1), seq_len(trunc_lvl))
  for (tree in trees_to_truncate) {
    object[tree, seq_len(d - tree)] <- 0
  }
  attr(object, "trunc_lvl") <- trunc_lvl
  object
}

#' @export
#' @rdname truncate_model
truncate_model.vinecop_dist <- function(object, trunc_lvl, ...) {
  check_trunc_lvl(object, trunc_lvl)
  if (trunc_lvl < dim(object)["trunc_lvl"]) {
    object <- adjust_fit_stats(object, trunc_lvl)
    object$structure <- truncate_model(object$structure, trunc_lvl)
    object$pair_copulas <- object$pair_copulas[seq_len(trunc_lvl)]
  }
  object
}

#' @export
#' @rdname truncate_model
truncate_model.vine_dist <- function(object, trunc_lvl, ...) {
  check_trunc_lvl(object, trunc_lvl)
  if (trunc_lvl < dim(object)["trunc_lvl"]) {
    object <- adjust_fit_stats(object, trunc_lvl)
    object$copula <- truncate_model(object$copula, trunc_lvl)
  }
  object
}

#' internal function
#' @noRd
truncate_column <- function(column, trunc_lvl) {
  column[1:min(length(column), trunc_lvl)]
}

#' internal function
#' @noRd
check_trunc_lvl <- function(object, trunc_lvl) {
  msg <- paste0(
    "trunc_lvl should be a number between 0 and the number of trees (",
    dim(object)["dim"] - 1, ")."
  )
  assert_that(
    is.number(trunc_lvl),
    trunc_lvl == round(trunc_lvl),
    trunc_lvl >= 0,
    trunc_lvl < dim(object)["dim"],
    msg = msg
  )
  if (trunc_lvl > dim(object)["trunc_lvl"]) {
    warning("truncation has no effect; vine is already ",
      dim(object)["trunc_lvl"], "-truncated.",
      call. = FALSE
    )
  }
  invisible(TRUE)
}

#' internal function
#' @noRd
get_truncated_pcs <- function(object, trunc_lvl) {
  check_trunc_lvl(object, trunc_lvl)
  if (!is.null(object$copula)) {
    pcs <- object$copula$pair_copulas[-seq_len(trunc_lvl)]
  } else {
    pcs <- object$pair_copulas[-seq_len(trunc_lvl)]
  }
  unlist(pcs, recursive = FALSE)
}

#' internal function
#' @noRd
adjust_fit_stats <- function(object, trunc_lvl) {
  trunc_pcs <- get_truncated_pcs(object, trunc_lvl)
  if (length(trunc_pcs) == 0) { # model is unchanged
    return(object)
  }

  # adjust npars for truncation
  trunc_npars <- sum(sapply(trunc_pcs, function(x) x[["npars"]]))
  object$npars <- object$npars - trunc_npars

  # adjust loglik for truncation
  if (!is.na(object$loglik)) {
    trunc_loglik <- sum(sapply(trunc_pcs, function(x) x[["loglik"]]))
    object$loglik <- object$loglik - trunc_loglik
  }

  object
}

#!/bin/bash

git clone --depth 1 git@github.com:vinecopulib/vinecopulib.git -b prep-release --single-branch

rm -rf ../inst/include/vinecopulib/*
mv ./vinecopulib/include/* ../inst/include

sed -i '11i#define INTERFACED_FROM_R' ./../inst/include/vinecopulib/misc/tools_interface.hpp

rm -rf ./../inst/include/vinecopulib/mainpage.h
rm -rf vinecopulib

Update following an email by Brian Ripley. The library is now compiled without 
forcing c++11, but boost uses features that were deprecated in later versions. 
As suggested by Ripley, we now add the `-D_HAS_AUTO_PTR_ETC=0` include flag to
disable the problematic parts.

## Test environments

* CRAN win builder (release)
* Windows Server 2019 (release)
* macOS (release)
* ubuntu 20.04 (release, oldrel, devel)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicop.R
\name{bicop}
\alias{bicop}
\title{Fit and select bivariate copula models}
\usage{
bicop(
  data,
  var_types = c("c", "c"),
  family_set = "all",
  par_method = "mle",
  nonpar_method = "quadratic",
  mult = 1,
  selcrit = "aic",
  weights = numeric(),
  psi0 = 0.9,
  presel = TRUE,
  keep_data = FALSE,
  cores = 1
)
}
\arguments{
\item{data}{a matrix or data.frame with at least two columns, containing the
(pseudo-)observations for the two variables (copula data should have
approximately uniform margins). More columns are required for discrete
models, see \emph{Details}.}

\item{var_types}{variable types, a length 2 vector; e.g., \code{c("c", "c")} for
both continuous (default), or \code{c("c", "d")} for first variable continuous
and second discrete.}

\item{family_set}{a character vector of families; see \emph{Details} for
additional options.}

\item{par_method}{the estimation method for parametric models, either \code{"mle"}
for maximum likelihood or \code{"itau"} for inversion of Kendall's tau (only
available for one-parameter families and \code{"t"}.}

\item{nonpar_method}{the estimation method for nonparametric models, either
\code{"constant"} for the standard transformation estimator, or
\code{"linear"}/\code{"quadratic"} for the local-likelihood approximations of order
one/two.}

\item{mult}{multiplier for the smoothing parameters of nonparametric
families. Values larger than 1 make the estimate more smooth, values less
than 1 less smooth.}

\item{selcrit}{criterion for family selection, either \code{"loglik"}, \code{"aic"},
\code{"bic"}, \code{"mbic"}. For \code{vinecop()} there is the additional option
\code{"mbicv"}.}

\item{weights}{optional vector of weights for each observation.}

\item{psi0}{see \code{\link[=mBICV]{mBICV()}}.}

\item{presel}{whether the family set should be thinned out according to
symmetry characteristics of the data.}

\item{keep_data}{whether the data should be stored (necessary for using
\code{\link[=fitted]{fitted()}}).}

\item{cores}{number of cores to use; if more than 1, estimation for multiple
families is done in parallel.}
}
\value{
An object inheriting from classes \code{bicop} and  \code{bicop_dist} . In addition to
the entries contained in \code{bicop_dist()}, objects from the \code{bicop} class
contain:
\itemize{
\item \code{data} (optionally, if \code{keep_data = TRUE} was used), the dataset that was
passed to \code{\link[=bicop]{bicop()}}.
\item \code{controls}, a \code{list} with the set of fit controls that was passed to
\code{\link[=bicop]{bicop()}}.
\item \code{loglik} the log-likelihood.
\item \code{nobs}, an \code{integer} with the number of observations that was used to fit
the model.
}
}
\description{
Fit a bivariate copula model for continuous or discrete data. The family
can be selected automatically from a vector of options.
}
\details{
If there are missing data (i.e., \code{NA} entries), incomplete observations are
discarded before fitting the copula.
\subsection{Discrete variables}{

When at least one variable is discrete, more than two columns are required
for \code{data}: the first \eqn{n \times 2} block contains realizations of
\eqn{F_{X_1}(x_1), F_{X_2}(x_2)}. The second \eqn{n \times 2} block contains
realizations of \eqn{F_{X_1}(x_1^-), F_{X_1}(x_1^-)}. The minus indicates a
left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
\eqn{F_{X_1}(x_1^-) = F_{X_1}(x_1 - 1)}. For continuous variables the left
limit and the cdf itself coincide. Respective columns can be omitted in the
second block.
}

\subsection{Family collections}{

The \code{family_set} argument accepts all families in \code{bicop_dist()} plus the
following convenience definitions:
\itemize{
\item \code{"all"} contains all the families,
\item \code{"parametric"} contains the parametric families (all except \code{"tll"}),
\item \code{"nonparametric"} contains the nonparametric families (\code{"indep"} and
\code{"tll"})
\item \code{"onepar"} contains the parametric families with a single parameter,
}

(\code{"gaussian"}, \code{"clayton"}, \code{"gumbel"}, \code{"frank"}, and \code{"joe"}),
\itemize{
\item \code{"twopar"} contains the parametric families with two parameters
(\code{"t"}, \code{"bb1"}, \code{"bb6"}, \code{"bb7"}, and \code{"bb8"}),
\item \code{"elliptical"} contains the elliptical families,
\item \code{"archimedean"} contains the archimedean families,
\item \code{"BB"} contains the BB families,
\item \code{"itau"} families for which estimation by Kendall's tau inversion is
available (\code{"indep"},\code{"gaussian"}, \code{"t"},\code{"clayton"}, \code{"gumbel"},
\code{"frank"}, \code{"joe"}).
}
}
}
\examples{
## fitting a continuous model from simulated data
u <- rbicop(100, "clayton", 90, 3)
fit <- bicop(u, family_set = "par")
summary(fit)

## compare fit with true model
contour(fit)
contour(bicop_dist("clayton", 90, 3), col = 2, add = TRUE)

## fit a model from discrete data
x_disc <- qpois(u, 1)  # transform to Poisson margins
plot(x_disc)
udisc <- cbind(ppois(x_disc, 1), ppois(x_disc - 1, 1))
fit_disc <- bicop(udisc, var_types = c("d", "d"))
summary(fit_disc)
}
\seealso{
\code{\link[=bicop_dist]{bicop_dist()}}, \code{\link[=plot.bicop]{plot.bicop()}}, \code{\link[=contour.bicop]{contour.bicop()}}, \code{\link[=dbicop]{dbicop()}},
\code{\link[=pbicop]{pbicop()}}, \code{\link[=hbicop]{hbicop()}}, \code{\link[=rbicop]{rbicop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairs_copula_data.R
\name{pairs_copula_data}
\alias{pairs_copula_data}
\title{Exploratory pairs plot for copula data}
\usage{
pairs_copula_data(data, ...)
}
\arguments{
\item{data}{the data (must lie in the unit hypercube).}

\item{\dots}{other parameters passed to \code{pairs.default()},
\code{contour.bicop()}, \code{points.default()}, \code{hist.default()}, or \code{bicop()}.}
}
\description{
This function provides pair plots for copula data. It shows
bivariate contour plots on the lower panel, scatter plots and
correlations on the upper panel and histograms on the diagonal panel.
}
\examples{
u <- replicate(3, runif(100))
pairs_copula_data(u)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicop_methods.R
\name{par_to_ktau}
\alias{par_to_ktau}
\alias{ktau_to_par}
\title{Conversion between Kendall's tau and parameters}
\usage{
par_to_ktau(family, rotation, parameters)

ktau_to_par(family, tau)
}
\arguments{
\item{family}{a copula family (see \code{\link[=bicop_dist]{bicop_dist()}}) or a \link{bicop_dist} object.}

\item{rotation}{the rotation of the copula, one of \code{0}, \code{90}, \code{180}, \code{270}.}

\item{parameters}{vector or matrix of copula parameters, not used when
\code{family} is a \code{bicop_dist} object.}

\item{tau}{Kendall's \eqn{\tau}.}
}
\description{
Conversion between Kendall's tau and parameters
}
\examples{
# the following are equivalent
par_to_ktau(bicop_dist("clayton", 0, 3))
par_to_ktau("clayton", 0, 3)

ktau_to_par("clayton", 0.5)
ktau_to_par(bicop_dist("clayton", 0, 3), 0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvine_structure.R
\name{rvine_structure}
\alias{rvine_structure}
\alias{rvine_matrix}
\alias{is.rvine_structure}
\alias{is.rvine_matrix}
\alias{dvine_structure}
\alias{cvine_structure}
\title{R-vine structure}
\usage{
rvine_structure(order, struct_array = list(), is_natural_order = FALSE)

cvine_structure(order, trunc_lvl = Inf)

dvine_structure(order, trunc_lvl = Inf)

rvine_matrix(matrix)
}
\arguments{
\item{order}{a vector of positive integers.}

\item{struct_array}{a list of vectors of positive integers. The vectors
represent rows of the r-rvine structure and the number of elements have to
be compatible with the \code{order} vector. If empty, the model is 0-truncated.}

\item{is_natural_order}{whether \code{struct_array} is assumed to be provided in
natural order already (a structure is in natural order if the anti-
diagonal is 1, .., d from bottom left to top right).}

\item{trunc_lvl}{the truncation level}

\item{matrix}{an R-vine matrix, see \emph{Details}.}
}
\value{
Either an \code{rvine_structure} or an \code{rvine_matrix}.
}
\description{
R-vine structures are compressed representations encoding the tree structure
of the vine, i.e. the conditioned/conditioning variables of each edge. The
functions \verb{[cvine_structure()]} or \verb{[dvine_structure()]} give a simpler way
to construct C-vines (every tree is a star) and D-vines (every tree is a
path), respectively (see \emph{Examples}).
}
\details{
The R-vine structure is essentially a lower-triangular matrix/triangular
array, with a notation that differs from the one in the VineCopula package.
An example array is

\if{html}{\out{<div class="sourceCode">}}\preformatted{4 4 4 4
3 3 3
2 2
1
}\if{html}{\out{</div>}}

which encodes the following pair-copulas:

\tabular{lll}{
tree \tab  edge \tab pair-copulas   \cr
0    \tab  0    \tab \verb{(1, 4)}       \cr
\tab  1    \tab \verb{(2, 4)}       \cr
\tab  2    \tab \verb{(3, 4)}       \cr
1    \tab  0    \tab \verb{(1, 3; 4)}    \cr
\tab  1    \tab \verb{(2, 3; 4)}    \cr
2    \tab  0    \tab \verb{(1, 2; 3, 4)}
}

An R-vine structure can be converted to an R-vine matrix using
\code{\link[=as_rvine_matrix]{as_rvine_matrix()}}, which encodes the same model with a square matrix filled
with zeros. For instance, the matrix corresponding to the structure above is:

\if{html}{\out{<div class="sourceCode">}}\preformatted{4 4 4 4
3 3 3 0
2 2 0 0
1 0 0 0
}\if{html}{\out{</div>}}

Similarly, an R-vine matrix can be converted to an R-vine structure using
\code{\link[=as_rvine_structure]{as_rvine_structure()}}.

Denoting by \code{M[i, j]} the array entry in row \code{i} and column \code{j} (the
pair-copula index for edge \code{e} in tree \code{t} of a \code{d} dimensional vine is
\verb{(M[d + 1 - e, e], M[t, e]; M[t - 1, e], ..., M[1, e])}. Less formally,
\enumerate{
\item Start with the counter-diagonal element of column \code{e} (first conditioned
variable).
\item Jump up to the element in row \code{t} (second conditioned variable).
\item Gather all entries further up in column \code{e} (conditioning set).
}

Internally, the diagonal is stored separately from the off-diagonal elements,
which are stored as a triangular array. For instance, the off-diagonal
elements off the structure above are stored as

\if{html}{\out{<div class="sourceCode">}}\preformatted{4 4 4
3 3
2
}\if{html}{\out{</div>}}

for the structure above. The reason is that it allows for parsimonious
representations of truncated models. For instance, the 2-truncated model is
represented by the same diagonal and the following truncated triangular
array:

\if{html}{\out{<div class="sourceCode">}}\preformatted{4 4 4
3 3
}\if{html}{\out{</div>}}

A valid R-vine structure or matrix must satisfy several conditions which are
checked when \code{\link[=rvine_structure]{rvine_structure()}}, \code{\link[=rvine_matrix]{rvine_matrix()}}, or some coercion methods
(see \code{\link[=as_rvine_structure]{as_rvine_structure()}} and \verb{as_rvine_matrix(}) are called:
\enumerate{
\item It can only contain numbers between 1 and d (and additionally zeros for
R-vine matrices).
\item The anti-diagonal must contain the numbers 1, ..., d.
\item The anti-diagonal entry of a column must not be contained in any column
further to the right.
\item The entries of a column must be contained in all columns to the left.
\item The proximity condition must hold: For all t = 1, ..., d - 2 and e = 1,
..., d - t there must exist an index j > d, such that
\verb{(M[t, e], \{M[1, e], ..., M[t - 1, e]\})} equals either
\verb{(M[d + 1 - j, j], \{M[1, j], ..., M[t - 1, j]\})} or
\verb{(M[t - 1, j], \{M[d + 1 - j, j], M[1, j], ..., M[t - 2, j]\})}.
}

Condition 5 already implies conditions 2-4, but is more difficult to check by
hand.
}
\examples{

# R-vine structures can be constructed from the order vector and struct_array
rvine_structure(order = 1:4, struct_array = list(
  c(4, 4, 4),
  c(3, 3),
  2
))

# R-vine matrices can be constructed from standard matrices
mat <- matrix(c(4, 3, 2, 1, 4, 3, 2, 0, 4, 3, 0, 0, 4, 0, 0, 0), 4, 4)
rvine_matrix(mat)

# coerce to R-vine structure
str(as_rvine_structure(mat))

# truncate and construct the R-vine matrix
mat[3, 1] <- 0
rvine_matrix(mat)

# or use directly the R-vine structure constructor
rvine_structure(order = 1:4, struct_array = list(
  c(4, 4, 4),
  c(3, 3)
))

# throws an error
mat[3, 1] <- 5
try(rvine_matrix(mat))

# C-vine structure
cvine <- cvine_structure(1:5)
cvine
plot(cvine)

# D-vine structure
dvine <- dvine_structure(c(1, 4, 2, 3, 5))
dvine
plot(dvine)

}
\seealso{
\code{\link[=as_rvine_structure]{as_rvine_structure()}}, \code{\link[=as_rvine_matrix]{as_rvine_matrix()}},
\code{\link[=plot.rvine_structure]{plot.rvine_structure()}}, \code{\link[=plot.rvine_matrix]{plot.rvine_matrix()}},
\code{\link[=rvine_structure_sim]{rvine_structure_sim()}}, \code{\link[=rvine_matrix_sim]{rvine_matrix_sim()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{emp_cdf}
\alias{emp_cdf}
\title{Corrected Empirical CDF}
\usage{
emp_cdf(x)
}
\arguments{
\item{x}{numeric vector of observations}
}
\value{
A function with signature \verb{function(x)} that returns \eqn{F_n(x)}.
}
\description{
The empirical CDF with tail correction, ensuring that its output is never
0 or 1.
}
\details{
The corrected empirical CDF is defined as
\deqn{
F_n(x) = \frac{1}{n + 1} \min\biggl\{1, \sum_{i = 1}^n 1(X_i \le x)\biggr\}
}
}
\examples{
# fit ECDF on simulated data
x <- rnorm(100)
cdf <- emp_cdf(x)

# output is bounded away from 0 and 1
cdf(-50)
cdf(50)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rosenblatt.R
\name{rosenblatt}
\alias{rosenblatt}
\alias{inverse_rosenblatt}
\title{(Inverse) Rosenblatt transform}
\usage{
rosenblatt(x, model, cores = 1)

inverse_rosenblatt(u, model, cores = 1)
}
\arguments{
\item{x}{matrix of evaluation points; must be in \eqn{(0, 1)^d} for copula
models.}

\item{model}{a model object; classes currently supported are
\code{bicop_dist()}, \code{vinecop_dist()}, and \code{vine_dist()}.}

\item{cores}{if \verb{>1}, computation is parallelized over \code{cores} batches (rows
of \code{u}).}

\item{u}{matrix of evaluation points; must be in \eqn{(0, 1)^d}.}
}
\description{
The Rosenblatt transform takes data generated from a model and turns it into
independent uniform variates, The inverse Rosenblatt transform computes
conditional quantiles and can be used simulate from a stochastic model,
see \emph{Details}.
}
\details{
The Rosenblatt transform (Rosenblatt, 1952) \eqn{U = T(V)} of a random vector
\eqn{V = (V_1,\ldots,V_d) ~ F} is defined as
\deqn{
  U_1= F(V_1), U_{2} = F(V_{2}|V_1), \ldots, U_d =F(V_d|V_1,\ldots,V_{d-1}),
}
where \eqn{F(v_k|v_1,\ldots,v_{k-1})} is the conditional distribution of
\eqn{V_k} given \eqn{V_1 \ldots, V_{k-1}, k = 2,\ldots,d}. The vector
\eqn{U  = (U_1, \dots, U_d)} then contains independent standard uniform
variables. The inverse operation
\deqn{
  V_1 = F^{-1}(U_1), V_{2} = F^{-1}(U_2|U_1), \ldots,
  V_d =F^{-1}(U_d|U_1,\ldots,U_{d-1}),
}
can be used to simulate from a distribution. For any copula \eqn{F}, if
\eqn{U} is a vector of independent random variables, \eqn{V = T^{-1}(U)} has
distribution \eqn{F}.

The formulas above assume a vine copula model with order \eqn{d, \dots, 1}.
More generally, \code{rosenblatt()} returns the variables
\deqn{
  U_{M[d + 1- j, j]}= F(V_{M[d + 1- j, j]} | V_{M[d - j, j - 1]}, \dots, V_{M[1, 1]}),
}
where \eqn{M} is the structure matrix. Similarly, \code{inverse_rosenblatt()}
returns
\deqn{
  V_{M[d + 1- j, j]}= F^{-1}(U_{M[d + 1- j, j]} | U_{M[d - j, j - 1]}, \dots, U_{M[1, 1]}).
}
}
\examples{
# simulate data with some dependence
x <- replicate(3, rnorm(200))
x[, 2:3] <- x[, 2:3] + x[, 1]
pairs(x)

# estimate a vine distribution model
fit <- vine(x, copula_controls = list(family_set = "par"))

# transform into independent uniforms
u <- rosenblatt(x, fit)
pairs(u)

# inversion
pairs(inverse_rosenblatt(u, fit))

# works similarly for vinecop models
vc <- fit$copula
rosenblatt(pseudo_obs(x), vc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vinecop.R
\name{vinecop}
\alias{vinecop}
\title{Fitting vine copula models}
\usage{
vinecop(
  data,
  var_types = rep("c", NCOL(data)),
  family_set = "all",
  structure = NA,
  par_method = "mle",
  nonpar_method = "constant",
  mult = 1,
  selcrit = "aic",
  weights = numeric(),
  psi0 = 0.9,
  presel = TRUE,
  trunc_lvl = Inf,
  tree_crit = "tau",
  threshold = 0,
  keep_data = FALSE,
  show_trace = FALSE,
  cores = 1
)
}
\arguments{
\item{data}{a matrix or data.frame with at least two columns, containing the
(pseudo-)observations for the two variables (copula data should have
approximately uniform margins). More columns are required for discrete
models, see \emph{Details}.}

\item{var_types}{variable types, a length d vector; e.g., \code{c("c", "c")} for
two continuous variables, or \code{c("c", "d")} for first variable continuous
and second discrete.}

\item{family_set}{a character vector of families; see \code{\link[=bicop]{bicop()}} for
additional options.}

\item{structure}{an \code{rvine_structure} object, namely a compressed
representation of the vine structure, or an object that can be coerced into
one (see \code{\link[=rvine_structure]{rvine_structure()}} and \code{\link[=as_rvine_structure]{as_rvine_structure()}}). The dimension
must be \code{length(pair_copulas[[1]]) + 1}; \code{structure = NA} performs
automatic selection based on Dissman's algorithm. See \emph{Details} for partial
selection of the structure.}

\item{par_method}{the estimation method for parametric models, either \code{"mle"}
for maximum likelihood or \code{"itau"} for inversion of Kendall's tau (only
available for one-parameter families and \code{"t"}.}

\item{nonpar_method}{the estimation method for nonparametric models, either
\code{"constant"} for the standard transformation estimator, or
\code{"linear"}/\code{"quadratic"} for the local-likelihood approximations of order
one/two.}

\item{mult}{multiplier for the smoothing parameters of nonparametric
families. Values larger than 1 make the estimate more smooth, values less
than 1 less smooth.}

\item{selcrit}{criterion for family selection, either \code{"loglik"}, \code{"aic"},
\code{"bic"}, \code{"mbic"}. For \code{vinecop()} there is the additional option
\code{"mbicv"}.}

\item{weights}{optional vector of weights for each observation.}

\item{psi0}{prior probability of a non-independence copula (only used for
\code{selcrit = "mbic"} and \code{selcrit = "mbicv"}).}

\item{presel}{whether the family set should be thinned out according to
symmetry characteristics of the data.}

\item{trunc_lvl}{the truncation level of the vine copula; \code{Inf} means no
truncation, \code{NA} indicates that the truncation level should be selected
automatically by \code{\link[=mBICV]{mBICV()}}.}

\item{tree_crit}{the criterion for tree selection, one of \code{"tau"}, \code{"rho"},
\code{"hoeffd"}, \code{"mcor"}, or \code{"joe"} for Kendall's \eqn{\tau}, Spearman's
\eqn{\rho}, Hoeffding's \eqn{D}, maximum correlation, or logarithm of
the partial correlation, respectively.}

\item{threshold}{for thresholded vine copulas; \code{NA} indicates that the
threshold should be selected automatically by \code{\link[=mBICV]{mBICV()}}.}

\item{keep_data}{whether the data should be stored (necessary for using
\code{\link[=fitted]{fitted()}}).}

\item{show_trace}{logical; whether a trace of the fitting progress should be
printed.}

\item{cores}{number of cores to use; if more than 1, estimation of pair
copulas within a tree is done in parallel.}
}
\value{
Objects inheriting from \code{vinecop} and \code{vinecop_dist} for \code{\link[=vinecop]{vinecop()}}. In
addition to the entries provided by \code{\link[=vinecop_dist]{vinecop_dist()}}, there are:
\itemize{
\item \code{threshold}, the (set or estimated) threshold used for thresholding the
vine.
\item \code{data} (optionally, if \code{keep_data = TRUE} was used), the dataset that was
passed to \code{\link[=vinecop]{vinecop()}}.
\item \code{controls}, a \code{list} with fit controls that was passed to \code{\link[=vinecop]{vinecop()}}.
\item \code{nobs}, the number of observations that were used to fit the model.
}
}
\description{
Automated fitting and model selection for vine copula models with continuous
or discrete data. Selection of the structure is performed using the algorithm
of Dissmann et al. (2013).
}
\details{
\subsection{Missing data}{

If there are missing data (i.e., \code{NA} entries), incomplete observations are
discarded before fitting a pair-copula. This is done on a pair-by-pair basis
so that the maximal available information is used.
}

\subsection{Discrete variables}{

The dependence measures used to select trees (default: Kendall's tau) are
corrected for ties (see \link[wdm:wdm]{wdm::wdm}).

Let \code{n} be the number of observations and \code{d} the number of variables.
When at least one variable is discrete, two types of
"observations" are required in \code{data}: the first \verb{n  x  d} block
contains realizations of \eqn{F_{X_j}(X_j)}. The second \verb{n  x  d}
block contains realizations of \eqn{F_{X_j}(X_j^-)}. The minus indicates a
left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
\eqn{F_{X_j}(X_j^-) = F_{X_j}(X_j - 1)}. For continuous variables the left
limit and the cdf itself coincide. Respective columns can be omitted in the
second block.
}

\subsection{Partial structure selection}{

It is possible to fix the vine structure only in the first trees and select
the remaining ones automatically. To specify only the first \code{k} trees, supply
a \code{k}-truncated \code{rvine_structure()} or \code{rvine_matrix()}. All trees up to
\code{trunc_lvl} will then be selected automatically.
}
}
\examples{
## simulate dummy data
x <- rnorm(30) * matrix(1, 30, 5) + 0.5 * matrix(rnorm(30 * 5), 30, 5)
u <- pseudo_obs(x)

## fit and select the model structure, family and parameters
fit <- vinecop(u)
summary(fit)
plot(fit)
contour(fit)

## select by log-likelihood criterion from one-paramter families
fit <- vinecop(u, family_set = "onepar", selcrit = "bic")
summary(fit)

## Gaussian D-vine
fit <- vinecop(u, structure = dvine_structure(1:5), family = "gauss")
plot(fit)
contour(fit)

## Partial structure selection with only first tree specified
structure <- rvine_structure(order = 1:5, list(rep(5, 4)))
structure
fit <- vinecop(u, structure = structure, family = "gauss")
plot(fit)

## 1-truncated model with random structure
fit <- vinecop(u, structure = rvine_structure_sim(5), trunc_lvl = 1)
contour(fit)

## Model for discrete data
x <- qpois(u, 1)  # transform to Poisson margins
# we require two types of observations (see Details)
u_disc <- cbind(ppois(x, 1), ppois(x - 1, 1))
fit <- vinecop(u_disc, var_types = rep("d", 5))

## Model for mixed data
x <- qpois(u[, 1], 1)  # transform first variable to Poisson margin
# we require two types of observations (see Details)
u_disc <- cbind(ppois(x, 1), u[, 2:5], ppois(x - 1, 1))
fit <- vinecop(u_disc, var_types = c("d", rep("c", 4)))
}
\references{
Dissmann, J. F., E. C. Brechmann, C. Czado, and D. Kurowicka (2013).
\emph{Selecting and estimating regular vine copulae and application to
financial returns.} Computational Statistics & Data Analysis, 59 (1),
52-69.
}
\seealso{
\code{\link[=vinecop]{vinecop()}}, \code{\link[=dvinecop]{dvinecop()}}, \code{\link[=pvinecop]{pvinecop()}}, \code{\link[=rvinecop]{rvinecop()}},
\code{\link[=plot.vinecop]{plot.vinecop()}}, \code{\link[=contour.vinecop]{contour.vinecop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vinecop_plot.R
\name{plot.vinecop_dist}
\alias{plot.vinecop_dist}
\alias{contour.vinecop_dist}
\alias{plot.vinecop}
\alias{contour.vinecop}
\title{Plotting \code{vinecop_dist} and \code{vinecop} objects.}
\usage{
\method{plot}{vinecop_dist}(x, tree = 1, var_names = "ignore", edge_labels = NULL, ...)

\method{plot}{vinecop}(x, tree = 1, var_names = "ignore", edge_labels = NULL, ...)

\method{contour}{vinecop_dist}(x, tree = "ALL", cex.nums = 1, ...)

\method{contour}{vinecop}(x, tree = "ALL", cex.nums = 1, ...)
}
\arguments{
\item{x}{\code{vinecop_dist} object.}

\item{tree}{\code{"ALL"} or integer vector; specifies which trees are
plotted.}

\item{var_names}{integer; specifies how to make use of variable names:
\itemize{
\item `"ignore"`` = variable names are ignored,
\item `"use"`` = variable names are used to annotate vertices,
\item `"legend"`` = uses numbers in plot and adds a legend for variable names,
\item `"hide"`` = no numbers or names, just the node.
}}

\item{edge_labels}{character; options are:
\itemize{
\item \code{"family"} = pair-copula family (see \verb{[bicop_dist()]}),
\item `"tau"`` = pair-copula Kendall's tau
\item `"family_tau"`` = pair-copula family and Kendall's tau,
\item `"pair"`` = the name of the involved variables.
}}

\item{\dots}{Unused for \code{plot} and passed to
\code{\link[rvinecopulib:contour.bicop]{contour.bicop}} for \code{contour}.}

\item{cex.nums}{numeric; expansion factor for font of the numbers.}
}
\description{
There are two plotting generics for \code{vinecop_dist} objects.
\code{plot.vinecop_dist} plots one or all trees of a given R-vine copula
model. Edges can be labeled with information about the corresponding
pair-copula. \code{contour.vinecop_dist} produces a matrix of contour plots
(using \code{\link[rvinecopulib:plot.bicop]{plot.bicop}}).
}
\details{
If you want the contour boxes to be perfect squares, the plot height should
be \code{1.25/length(tree)*(d - min(tree))} times the plot width.

The \code{plot()} method returns an object that (among other things) contains the
\code{igraph} representation of the graph; see \emph{Examples}.
}
\examples{
# set up vine copula model
u <- matrix(runif(20 * 10), 20, 10)
vc <- vinecop(u, family = "indep")

# plot
plot(vc, tree = c(1, 2))
plot(vc, edge_labels = "pair")

# extract igraph representation
plt <- plot(vc, edge_labels = "family_tau")
igr_obj <- get("g", plt$plot_env)[[1]]
igr_obj  # print object
igraph::E(igr_obj)$name  # extract edge labels

# set up another vine copula model
pcs <- lapply(1:3, function(j) # pair-copulas in tree j
  lapply(runif(4 - j), function(cor) bicop_dist("gaussian", 0, cor)))
mat <- rvine_matrix_sim(4)
vc <- vinecop_dist(pcs, mat)

# contour plot
contour(vc)
}
\seealso{
\code{\link[rvinecopulib:vinecop_dist]{vinecop_dist}},
\code{\link[rvinecopulib:plot.bicop]{plot.bicop}}
}
\author{
Thomas Nagler, Thibault Vatter
}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_rvine_structure.R
\name{as_rvine_structure}
\alias{as_rvine_structure}
\alias{as_rvine_matrix}
\alias{as_rvine_structure.rvine_structure}
\alias{as_rvine_matrix.rvine_structure}
\alias{as_rvine_structure.list}
\alias{as_rvine_matrix.list}
\alias{as_rvine_structure.rvine_matrix}
\alias{as_rvine_matrix.rvine_matrix}
\alias{as_rvine_structure.matrix}
\alias{as_rvine_matrix.matrix}
\title{Coerce various kind of objects to R-vine structures and matrices}
\usage{
as_rvine_structure(x, ...)

as_rvine_matrix(x, ...)

\method{as_rvine_structure}{rvine_structure}(x, ..., validate = FALSE)

\method{as_rvine_matrix}{rvine_structure}(x, ..., validate = FALSE)

\method{as_rvine_structure}{list}(x, ..., is_natural_order = FALSE)

\method{as_rvine_matrix}{list}(x, ..., is_natural_order = FALSE)

\method{as_rvine_structure}{rvine_matrix}(x, ..., validate = FALSE)

\method{as_rvine_matrix}{rvine_matrix}(x, ..., validate = FALSE)

\method{as_rvine_structure}{matrix}(x, ..., validate = TRUE)

\method{as_rvine_matrix}{matrix}(x, ..., validate = TRUE)
}
\arguments{
\item{x}{An object of class \code{rvine_structure}, \code{rvine_matrix}, \code{matrix} or
\code{list} that can be coerced into an R-vine structure or R-vine matrix
(see \emph{Details}).}

\item{...}{Other arguments passed on to individual methods.}

\item{validate}{When `TRUE``, verifies that the input is a valid
rvine-structure (see \emph{Details}). You may want to suppress this when you
know that you already have a valid structure and you want to save some time,
or to explicitly enable it if you have a structure that you want to re-check.}

\item{is_natural_order}{A flag indicating whether the \code{struct_array} element
of \code{x} is assumed to be provided in natural order already (a structure is in
natural order if the anti-diagonal is 1, .., d from bottom left to top
right).}
}
\value{
Either an object of class \code{rvine_structure} or of class
\code{rvine_matrix} (see \code{\link[=rvine_structure]{rvine_structure()}} or \code{\link[=rvine_matrix]{rvine_matrix()}}).
}
\description{
\code{as_rvine_structure} and \code{as_rvine_matrix} are new S3 generics allowing to
coerce objects into R-vine structures and matrices (see \code{\link[=rvine_structure]{rvine_structure()}}
and \code{\link[=rvine_matrix]{rvine_matrix()}}).
}
\details{
The coercion to \code{rvine_structure} and \code{rvine_matrix} can be applied to
different kind of objects Currently, \code{rvine_structure}, \code{rvine_matrix},
\code{matrix} and \code{list} are supported.

For \code{as_rvine_structure}:
\itemize{
\item \code{rvine_structure} : the main use case is to re-check an object via
\code{validate = TRUE}.
\item \code{rvine_matrix} and \code{matrix} : allow to coerce matrices into R-vine
structures (see \code{\link[=rvine_structure]{rvine_structure()}} for more details). The main difference
between \code{rvine_matrix} and \code{matrix} is the nature of the validity
checks.
\item \code{list} : must contain named elements \code{order} and \code{struct_array} to be
coerced into an R-vine structure (see \code{\link[=rvine_structure]{rvine_structure()}} for more details).
}

For \code{as_rvine_matrix}:
\itemize{
\item \code{rvine_structure} : allow to coerce an \code{rvine_structure} into an
R-vine matrix (useful e.g. for printing).
\item \code{rvine_matrix}: similar to \code{as_rvine_structure} for \code{rvine_structure},
the main use case is to re-check an object via \code{validate = TRUE}.
\item \code{matrix} : allow to coerce matrices into R-vine
matrices (mainly by checking that the matrix defines a valid
R-vine, see \code{\link[=rvine_matrix]{rvine_matrix()}} for more details).
\item \code{list} : must contain named elements \code{order} and \code{struct_array} to be
coerced into an R-vine matrix (see \code{\link[=rvine_structure]{rvine_structure()}} for more details).
}
}
\examples{
# R-vine structures can be constructed from the order vector and struct_array
rvine_structure(order = 1:4, struct_array = list(
  c(4, 4, 4),
  c(3, 3),
  2
))

# ... or a similar list can be coerced into an R-vine structure
as_rvine_structure(list(order = 1:4, struct_array = list(
  c(4, 4, 4),
  c(3, 3),
  2
)))

# similarly, standard matrices can be coerced into R-vine structures
mat <- matrix(c(4, 3, 2, 1, 4, 3, 2, 0, 4, 3, 0, 0, 4, 0, 0, 0), 4, 4)
as_rvine_structure(mat)

# or truncate and construct the structure
mat[3, 1] <- 0
as_rvine_structure(mat)

# throws an error
mat[3, 1] <- 5
try(as_rvine_structure(mat))
}
\seealso{
rvine_structure rvine_matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vinecop.R
\name{vinecop_dist}
\alias{vinecop_dist}
\title{Vine copula models}
\usage{
vinecop_dist(pair_copulas, structure, var_types = rep("c", dim(structure)[1]))
}
\arguments{
\item{pair_copulas}{A nested list of '\code{\link[=bicop_dist]{bicop_dist()}}' objects, where
\code{pair_copulas[[t]][[e]]} corresponds to the pair-copula at edge \code{e} in
tree \code{t}.}

\item{structure}{an \code{rvine_structure} object, namely a compressed
representation of the vine structure, or an object that can be coerced into
one (see \code{\link[=rvine_structure]{rvine_structure()}} and \code{\link[=as_rvine_structure]{as_rvine_structure()}}). The dimension
must be \code{length(pair_copulas[[1]]) + 1}; \code{structure = NA} performs
automatic selection based on Dissman's algorithm. See \emph{Details} for partial
selection of the structure.}

\item{var_types}{variable types, a length d vector; e.g., \code{c("c", "c")} for
two continuous variables, or \code{c("c", "d")} for first variable continuous
and second discrete.}
}
\value{
Object of class \code{vinecop_dist}, i.e., a list containing:
\itemize{
\item \code{pair_copulas}, a list of lists. Each element of \code{pair_copulas} corresponds
to a tree, which is itself a list of \code{\link[=bicop_dist]{bicop_dist()}} objects.
\item \code{structure}, a compressed representation of the vine structure, or an
object that can be coerced into one (see \code{\link[=rvine_structure]{rvine_structure()}} and
\code{\link[=as_rvine_structure]{as_rvine_structure()}}).
\item \code{npars}, a \code{numeric} with the number of (effective) parameters.
\item \code{var_types} the variable types.
}
}
\description{
Create custom vine copula models by specifying the pair-copulas, structure,
and variable types.
}
\examples{
# specify pair-copulas
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(
  list(bicop, bicop), # pair-copulas in first tree
  list(bicop) # pair-copulas in second tree
)

# specify R-vine matrix
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)

# set up vine copula model
vc <- vinecop_dist(pcs, mat)

# visualization
plot(vc)
contour(vc)

# simulate from the model
pairs(rvinecop(200, vc))
}
\seealso{
\code{\link[=rvine_structure]{rvine_structure()}}, \code{\link[=rvine_matrix]{rvine_matrix()}}, \code{\link[=vinecop]{vinecop()}},
\code{\link[=plot.vinecop_dist]{plot.vinecop_dist()}}, \code{\link[=contour.vinecop_dist]{contour.vinecop_dist()}}, \code{\link[=dvinecop]{dvinecop()}},
\code{\link[=pvinecop]{pvinecop()}}, \code{\link[=rvinecop]{rvinecop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vine_methods.R
\name{vine_distributions}
\alias{vine_distributions}
\alias{dvine}
\alias{pvine}
\alias{rvine}
\alias{dvine_dist}
\alias{pvine_dist}
\alias{rvine_dist}
\title{Vine based distributions}
\usage{
dvine(x, vine, cores = 1)

pvine(x, vine, n_mc = 10^4, cores = 1)

rvine(n, vine, qrng = FALSE, cores = 1)
}
\arguments{
\item{x}{evaluation points, either a length d vector or a d-column matrix,
where d is the number of variables in the vine.}

\item{vine}{an object of class \code{"vine_dist"}.}

\item{cores}{number of cores to use; if larger than one, computations are
done in parallel on \code{cores} batches .}

\item{n_mc}{number of samples used for quasi Monte Carlo integration.}

\item{n}{number of observations.}

\item{qrng}{if \code{TRUE}, generates quasi-random numbers using the multivariate
Generalized Halton sequence up to dimension 300 and the Generalized Sobol
sequence in higher dimensions (default \code{qrng = FALSE}).}
}
\value{
\code{dvine()} gives the density, \code{pvine()} gives the distribution function,
and \code{rvine()} generates random deviates.

The length of the result is determined by \code{n} for \code{rvine()}, and
the number of rows in \code{u} for the other functions.

The \code{vine} object is recycled to the length of the
result.
}
\description{
Density, distribution function and random generation
for the vine based distribution.
}
\details{
See \link{vine} for the estimation and construction of vine models.
Here, the density, distribution function and random generation
for the vine distributions are standard.

The functions are based on \code{\link[=dvinecop]{dvinecop()}}, \code{\link[=pvinecop]{pvinecop()}} and \code{\link[=rvinecop]{rvinecop()}} for
\link{vinecop} objects, and either \code{\link[kde1d:dkde1d]{kde1d::dkde1d()}}, \code{\link[kde1d:dkde1d]{kde1d::pkde1d()}} and
\code{\link[kde1d:dkde1d]{kde1d::qkde1d()}} for estimated vines (i.e., output of \code{\link[=vine]{vine()}}), or the
standard \emph{d/p/q-xxx} from \link[stats:Distributions]{stats::Distributions} for custom vines
(i.e., output of \code{\link[=vine_dist]{vine_dist()}}).
}
\examples{
# specify pair-copulas
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(
  list(bicop, bicop), # pair-copulas in first tree
  list(bicop) # pair-copulas in second tree
)

# set up vine copula model
mat <- rvine_matrix_sim(3)
vc <- vine_dist(list(distr = "norm"), pcs, mat)

# simulate from the model
x <- rvine(200, vc)
pairs(x)

# evaluate the density and cdf
dvine(x[1, ], vc)
pvine(x[1, ], vc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvine_structure.R
\name{rvine_structure_sim}
\alias{rvine_structure_sim}
\alias{rvine_matrix_sim}
\title{Simulate R-vine structures}
\usage{
rvine_structure_sim(d, natural_order = FALSE)

rvine_matrix_sim(d, natural_order = FALSE)
}
\arguments{
\item{d}{the number of variables}

\item{natural_order}{boolean; whether the structures should be in natural
order (counter-diagonal is \code{1:d}).}
}
\description{
Simulates from a uniform distribution over all R-vine structures on d
variables. \code{rvine_structure_sim()} returns an \code{\link[=rvine_structure]{rvine_structure()}} object,
\code{rvine_matrix_sim()} an \code{\link[=rvine_matrix]{rvine_matrix()}}.
}
\examples{
rvine_structure_sim(10)

rvine_structure_sim(10, natural_order = TRUE)  # counter-diagonal is 1:d

rvine_matrix_sim(10)
}
\seealso{
\code{\link[=rvine_structure]{rvine_structure()}}, \code{\link[=rvine_matrix]{rvine_matrix()}},
\code{\link[=plot.rvine_structure]{plot.rvine_structure()}}, \code{\link[=plot.rvine_matrix]{plot.rvine_matrix()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicop_plot.R
\name{plot.bicop_dist}
\alias{plot.bicop_dist}
\alias{contour.bicop_dist}
\alias{plot.bicop}
\alias{contour.bicop}
\title{Plotting tools for \code{bicop_dist} and \code{bicop} objects}
\usage{
\method{plot}{bicop_dist}(x, type = "surface", margins, size, ...)

\method{plot}{bicop}(x, type = "surface", margins, size, ...)

\method{contour}{bicop_dist}(x, margins = "norm", size = 100L, ...)

\method{contour}{bicop}(x, margins = "norm", size = 100L, ...)
}
\arguments{
\item{x}{\code{bicop_dist object.}}

\item{type}{plot type; either \code{"surface"} or \code{"contour"}.}

\item{margins}{options are: \code{"unif"} for the original copula density,
\code{"norm"} for the transformed density with standard normal margins,
\code{"exp"} with standard exponential margins, and  \code{"flexp"} with
flipped exponential margins. Default is \code{"norm"} for \code{type =
  "contour"}, and \code{"unif"} for \code{type = "surface"}.}

\item{size}{integer; the plot is based on values on a \verb{size x size} grid,
default is 100.}

\item{\dots}{optional arguments passed to \code{\link{contour}} or
\code{\link{wireframe}}.}
}
\description{
There are several options for plotting bicop_dist objects. The density of a
bivariate copula density can be visualized as surface/perspective or contour
plot. Optionally, the density can be coupled with standard normal margins
(default for contour plots).
}
\examples{

## construct bicop_dist object for a student t copula
obj <- bicop_dist(family = "t", rotation = 0, parameters = c(0.7, 4))

## plots
plot(obj) # surface plot of copula density
contour(obj) # contour plot with standard normal margins
contour(obj, margins = "unif") # contour plot of copula density
}
\seealso{
\code{\link[=bicop_dist]{bicop_dist()}}, \code{\link[graphics:contour]{graphics::contour()}}, \code{\link[=wireframe]{wireframe()}}
}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicop.R
\name{bicop_dist}
\alias{bicop_dist}
\title{Bivariate copula models}
\usage{
bicop_dist(
  family = "indep",
  rotation = 0,
  parameters = numeric(0),
  var_types = c("c", "c")
)
}
\arguments{
\item{family}{the copula family, a string containing the family name (see
\emph{Details} for all possible families).}

\item{rotation}{the rotation of the copula, one of \code{0}, \code{90}, \code{180}, \code{270}.}

\item{parameters}{a vector or matrix of copula parameters.}

\item{var_types}{variable types, a length 2 vector; e.g., \code{c("c", "c")} for
both continuous (default), or \code{c("c", "d")} for first variable continuous
and second discrete.}
}
\value{
An object of class \code{bicop_dist}, i.e., a list containing:
\itemize{
\item \code{family}, a \code{character} indicating the copula family.
\item \code{rotation}, an \code{integer} indicating the rotation (i.e., either 0, 90, 180,
or 270).
\item \code{parameters}, a \code{numeric} vector or matrix of parameters.
\item \code{npars}, a \code{numeric} with the (effective) number of parameters.
\item \code{var_types}, the variable types.
}
}
\description{
Create custom bivariate copula models by specifying the family, rotation,
parameters, and variable types.
}
\details{
\subsection{Implemented families}{\tabular{lll}{
   type \tab name \tab name in R \cr
   - \tab Independence \tab "indep" \cr
   Elliptical \tab Gaussian \tab "gaussian" \cr
   " \tab Student t \tab "student" \cr
   Archimedean \tab Clayton \tab "clayton" \cr
   " \tab Gumbel \tab "gumbel" \cr
   " \tab Frank \tab "frank" \cr
   " \tab Joe \tab "joe" \cr
   " \tab Clayton-Gumbel (BB1) \tab "bb1" \cr
   " \tab Joe-Gumbel (BB6) \tab "bb6" \cr
   " \tab Joe-Clayton (BB7) \tab "bb7" \cr
   " \tab Joe-Frank (BB8) \tab "bb8" \cr
   Nonparametric \tab Transformation kernel \tab "tll" \cr
}

}
}
\examples{
## Clayton 90° copula with parameter 3
cop <- bicop_dist("clayton", 90, 3)
cop
str(cop)

## visualization
plot(cop)
contour(cop)
plot(rbicop(200, cop))

## BB8 copula model for discrete data
cop_disc <- bicop_dist("bb8", 0, c(2, 0.5), var_types = c("d", "d"))
cop_disc

}
\seealso{
\code{\link[=bicop_dist]{bicop_dist()}}, \code{\link[=plot.bicop]{plot.bicop()}}, \code{\link[=contour.bicop]{contour.bicop()}}, \code{\link[=dbicop]{dbicop()}},
\code{\link[=pbicop]{pbicop()}}, \code{\link[=hbicop]{hbicop()}}, \code{\link[=rbicop]{rbicop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vine.R
\name{vine}
\alias{vine}
\alias{vine_dist}
\title{Vine copula models}
\usage{
vine(
  data,
  margins_controls = list(mult = NULL, xmin = NaN, xmax = NaN, bw = NA, deg = 2),
  copula_controls = list(family_set = "all", structure = NA, par_method = "mle",
    nonpar_method = "constant", mult = 1, selcrit = "aic", psi0 = 0.9, presel = TRUE,
    trunc_lvl = Inf, tree_crit = "tau", threshold = 0, keep_data = FALSE, show_trace =
    FALSE, cores = 1),
  weights = numeric(),
  keep_data = FALSE,
  cores = 1
)

vine_dist(margins, pair_copulas, structure)
}
\arguments{
\item{data}{a matrix or data.frame. Discrete variables have to be declared as
\code{ordered()}.}

\item{margins_controls}{a list with arguments to be passed to
\code{\link[kde1d:kde1d]{kde1d::kde1d()}}. Currently, there can be
\itemize{
\item \code{mult} numeric vector of length one or d; all bandwidths for marginal
kernel density estimation are multiplied with \code{mult}. Defaults to
\code{log(1 + d)} where \code{d} is the number of variables after applying
\code{rvinecopulib:::expand_factors()}.
\item \code{xmin} numeric vector of length d; see \code{\link[kde1d:kde1d]{kde1d::kde1d()}}.
\item \code{xmax} numeric vector of length d; see \code{\link[kde1d:kde1d]{kde1d::kde1d()}}.
\item \code{bw} numeric vector of length d; see \code{\link[kde1d:kde1d]{kde1d::kde1d()}}.
\item \code{deg} numeric vector of length one or d; \code{\link[kde1d:kde1d]{kde1d::kde1d()}}.
}}

\item{copula_controls}{a list with arguments to be passed to \code{\link[=vinecop]{vinecop()}}.}

\item{weights}{optional vector of weights for each observation.}

\item{keep_data}{whether the original data should be stored; if you want to
store the pseudo-observations used for fitting the copula, use the
\code{copula_controls} argument.}

\item{cores}{the number of cores to use for parallel computations.}

\item{margins}{A list with with each element containing the specification of a
marginal \link[stats:Distributions]{stats::Distributions}. Each marginal specification
should be a list with containing at least the distribution family (\code{"distr"})
and optionally the parameters, e.g.
\code{list(list(distr = "norm"), list(distr = "norm", mu = 1), list(distr = "beta", shape1 = 1, shape2 = 1))}.
Note that parameters that have no default values have to be provided.
Furthermore, if \code{margins} has length one, it will be recycled for every component.}

\item{pair_copulas}{A nested list of 'bicop_dist' objects, where
\code{pair_copulas[[t]][[e]]} corresponds to the pair-copula at edge \code{e} in
tree \code{t}.}

\item{structure}{an \code{rvine_structure} object, namely a compressed
representation of the vine structure, or an object that can be coerced
into one (see \code{\link[=rvine_structure]{rvine_structure()}} and \code{\link[=as_rvine_structure]{as_rvine_structure()}}).
The dimension must be \code{length(pair_copulas[[1]]) + 1}.}
}
\value{
Objects inheriting from \code{vine_dist} for \code{\link[=vine_dist]{vine_dist()}}, and
\code{vine} and \code{vine_dist} for \code{\link[=vine]{vine()}}.

Objects from the \code{vine_dist} class are lists containing:
\itemize{
\item \code{margins}, a list of marginals (see below).
\item \code{copula}, an object of the class \code{vinecop_dist}, see \code{\link[=vinecop_dist]{vinecop_dist()}}.
}

For objects from the \code{vine} class, \code{copula} is also an object of the class
\code{vine}, see \code{\link[=vinecop]{vinecop()}}. Additionally, objects from the \code{vine} class contain:
\itemize{
\item \code{margins_controls}, a \code{list} with the set of fit controls that was passed
to \code{\link[kde1d:kde1d]{kde1d::kde1d()}} when estimating the margins.
\item \code{copula_controls}, a \code{list} with the set of fit controls that was passed
to \code{\link[=vinecop]{vinecop()}} when estimating the copula.
\item \code{data} (optionally, if \code{keep_data = TRUE} was used), the dataset that was
passed to \code{\link[=vine]{vine()}}.
\item \code{nobs}, an \code{integer} containing the number of observations that was used
to fit the model.
}

Concerning \code{margins}:
\itemize{
\item For objects created with \code{\link[=vine_dist]{vine_dist()}}, it simply corresponds to the \code{margins}
argument.
\item For objects created with \code{\link[=vine]{vine()}}, it is a list of objects of class \code{kde1d},
see \code{\link[kde1d:kde1d]{kde1d::kde1d()}}.
}
}
\description{
Automated fitting or creation of custom vine copula models
}
\details{
\code{vine_dist()} creates a vine copula by specifying the margins, a nested list
of \code{bicop_dist} objects and a quadratic structure matrix.

\code{vine()} provides automated fitting for vine copula models.
\code{margins_controls} is a list with the same parameters as
\code{\link[kde1d:kde1d]{kde1d::kde1d()}} (except for \code{x}). \code{copula_controls} is a list
with the same parameters as \code{\link[=vinecop]{vinecop()}} (except for \code{data}).
}
\examples{
# specify pair-copulas
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(
  list(bicop, bicop), # pair-copulas in first tree
  list(bicop) # pair-copulas in second tree
)

# specify R-vine matrix
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)

# set up vine copula model with Gaussian margins
vc <- vine_dist(list(distr = "norm"), pcs, mat)

# show model
summary(vc)

# simulate some data
x <- rvine(50, vc)

# estimate a vine copula model
fit <- vine(x, copula_controls = list(family_set = "par"))
summary(fit)

## model for discrete data
x <- as.data.frame(x)
x[, 1] <- ordered(round(x[, 1]), levels = seq.int(-5, 5))
fit_disc <- vine(x, copula_controls = list(family_set = "par"))
summary(fit_disc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicop_methods.R
\name{bicop_distributions}
\alias{bicop_distributions}
\alias{dbicop}
\alias{pbicop}
\alias{rbicop}
\alias{hbicop}
\alias{dbicop_dist}
\alias{pbicop_dist}
\alias{rbicop_dist}
\alias{hbicop_dist}
\title{Bivariate copula distributions}
\usage{
dbicop(u, family, rotation, parameters, var_types = c("c", "c"))

pbicop(u, family, rotation, parameters, var_types = c("c", "c"))

rbicop(n, family, rotation, parameters, qrng = FALSE)

hbicop(
  u,
  cond_var,
  family,
  rotation,
  parameters,
  inverse = FALSE,
  var_types = c("c", "c")
)
}
\arguments{
\item{u}{evaluation points, a matrix with at least two columns, see
\emph{Details}.}

\item{family}{the copula family, a string containing the family name (see
\code{\link{bicop}} for all possible families).}

\item{rotation}{the rotation of the copula, one of \code{0}, \code{90}, \code{180}, \code{270}.}

\item{parameters}{a vector or matrix of copula parameters.}

\item{var_types}{variable types, a length 2 vector; e.g., \code{c("c", "c")} for
both continuous (default), or \code{c("c", "d")} for first variable continuous
and second discrete.}

\item{n}{number of observations. If `length(n) > 1``, the length is taken to
be the number required.}

\item{qrng}{if \code{TRUE}, generates quasi-random numbers using the bivariate
Generalized Halton sequence (default \code{qrng = FALSE}).}

\item{cond_var}{either \code{1} or \code{2}; \code{cond_var = 1} conditions on the first
variable, \code{cond_var = 2} on the second.}

\item{inverse}{whether to compute the h-function or its inverse.}
}
\value{
\code{dbicop()} gives the density, \code{pbicop()} gives the distribution function,
\code{rbicop()} generates random deviates, and \code{hbicop()} gives the h-functions
(and their inverses).

The length of the result is determined by \code{n} for \code{rbicop()}, and
the number of rows in \code{u} for the other functions.

The numerical arguments other than \code{n} are recycled to the length of the
result.
}
\description{
Density, distribution function, random generation and h-functions (with their
inverses) for the bivariate copula distribution.
}
\details{
See \link{bicop} for the various implemented copula families.

The copula density is defined as joint density divided by marginal
densities, irrespective of variable types.

H-functions (\code{hbicop()}) are conditional distributions derived
from a copula. If \eqn{C(u, v) = P(U \le u, V \le v)} is a copula, then
\deqn{h_1(u, v) = P(V \le v | U = u) = \partial C(u, v) / \partial u,}
\deqn{h_2(u, v) = P(U \le u | V = v) = \partial C(u, v) / \partial v.}
In other words, the H-function number refers to the conditioning variable.
When inverting H-functions, the inverse is then taken with respect to the
other variable, that is \code{v} when \code{cond_var = 1} and \code{u} when \code{cond_var = 2}.
\subsection{Discrete variables}{

When at least one variable is discrete, more than two columns are required
for \code{u}: the first \eqn{n \times 2} block contains realizations of
\eqn{F_{X_1}(x_1), F_{X_2}(x_2)}. The second \eqn{n \times 2} block contains
realizations of \eqn{F_{X_1}(x_1^-), F_{X_1}(x_1^-)}. The minus indicates a
left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
\eqn{F_{X_1}(x_1^-) = F_{X_1}(x_1 - 1)}. For continuous variables the left
limit and the cdf itself coincide. Respective columns can be omitted in the
second block.
}
}
\note{
The functions can optionally be used with a \link{bicop_dist} object, e.g.,
\code{dbicop(c(0.1, 0.5), bicop_dist("indep"))}.
}
\examples{
## evaluate the copula density
dbicop(c(0.1, 0.2), "clay", 90, 3)
dbicop(c(0.1, 0.2), bicop_dist("clay", 90, 3))

## evaluate the copula cdf
pbicop(c(0.1, 0.2), "clay", 90, 3)

## simulate data
plot(rbicop(500, "clay", 90, 3))

## h-functions
joe_cop <- bicop_dist("joe", 0, 3)
# h_1(0.1, 0.2)
hbicop(c(0.1, 0.2), 1, "bb8", 0, c(2, 0.5))
# h_2^{-1}(0.1, 0.2)
hbicop(c(0.1, 0.2), 2, joe_cop, inverse = TRUE)

## mixed discrete and continuous data
x <- cbind(rpois(10, 1), rnorm(10, 1))
u <- cbind(ppois(x[, 1], 1), pnorm(x[, 2]), ppois(x[, 1] - 1, 1))
pbicop(u, "clay", 90, 3, var_types = c("d", "c"))

}
\seealso{
\code{\link[=bicop_dist]{bicop_dist()}}, \code{\link[=bicop]{bicop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vine_methods.R
\name{vine_predict_and_fitted}
\alias{vine_predict_and_fitted}
\alias{predict.vine}
\alias{fitted.vine}
\title{Predictions and fitted values for a vine copula model}
\usage{
\method{predict}{vine}(object, newdata, what = "pdf", n_mc = 10^4, cores = 1, ...)

\method{fitted}{vine}(object, what = "pdf", n_mc = 10^4, cores = 1, ...)
}
\arguments{
\item{object}{a \code{vine} object.}

\item{newdata}{points where the fit shall be evaluated.}

\item{what}{what to predict, either \code{"pdf"} or \code{"cdf"}.}

\item{n_mc}{number of samples used for quasi Monte Carlo integration when
\code{what = "cdf"}.}

\item{cores}{number of cores to use; if larger than one, computations are
done in parallel on \code{cores} batches .}

\item{...}{unused.}
}
\value{
\code{fitted()} and \code{predict()} have return values similar to \code{\link[=dvine]{dvine()}}
and \code{\link[=pvine]{pvine()}}.
}
\description{
Predictions of the density and distribution function
for a vine copula model.
}
\examples{
x <- sapply(1:5, function(i) rnorm(50))
fit <- vine(x, copula_controls = list(family_set = "par"), keep_data = TRUE)
all.equal(predict(fit, x), fitted(fit), check.environment = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vinecop_methods.R
\name{vinecop_distributions}
\alias{vinecop_distributions}
\alias{dvinecop}
\alias{pvinecop}
\alias{rvinecop}
\alias{dvinecop_dist}
\alias{pvinecop_dist}
\alias{rvinecop_dist}
\title{Vine copula distributions}
\usage{
dvinecop(u, vinecop, cores = 1)

pvinecop(u, vinecop, n_mc = 10^4, cores = 1)

rvinecop(n, vinecop, qrng = FALSE, cores = 1)
}
\arguments{
\item{u}{matrix of evaluation points; must contain at least d columns, where
d is the number of variables in the vine. More columns are required for
discrete models, see \emph{Details}.}

\item{vinecop}{an object of class \code{"vinecop_dist"}.}

\item{cores}{number of cores to use; if larger than one, computations are
done in parallel on \code{cores} batches .}

\item{n_mc}{number of samples used for quasi Monte Carlo integration.}

\item{n}{number of observations.}

\item{qrng}{if \code{TRUE}, generates quasi-random numbers using the multivariate
Generalized Halton sequence up to dimension 300 and the Generalized Sobol
sequence in higher dimensions (default \code{qrng = FALSE}).}
}
\value{
\code{dvinecop()} gives the density, \code{pvinecop()} gives the distribution function,
and \code{rvinecop()} generates random deviates.

The length of the result is determined by \code{n} for \code{rvinecop()}, and
the number of rows in \code{u} for the other functions.

The \code{vinecop} object is recycled to the length of the
result.
}
\description{
Density, distribution function and random generation for the vine copula
distribution.
}
\details{
See \code{\link[=vinecop]{vinecop()}} for the estimation and construction of vine copula
models.

The copula density is defined as joint density divided by marginal
densities, irrespective of variable types.
\subsection{Discrete variables}{

When at least one variable is discrete, two types of
"observations" are required in \code{u}: the first \eqn{n \; x \; d} block
contains realizations of \eqn{F_{X_j}(X_j)}. The second \eqn{n \; x \; d}
block contains realizations of \eqn{F_{X_j}(X_j^-)}. The minus indicates a
left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
\eqn{F_{X_j}(X_j^-) = F_{X_j}(X_j - 1)}. For continuous variables the left
limit and the cdf itself coincide. Respective columns can be omitted in the
second block.
}
}
\examples{
## simulate dummy data
x <- rnorm(30) * matrix(1, 30, 5) + 0.5 * matrix(rnorm(30 * 5), 30, 5)
u <- pseudo_obs(x)

## fit a model
vc <- vinecop(u, family = "clayton")

# simulate from the model
u <- rvinecop(100, vc)
pairs(u)

# evaluate the density and cdf
dvinecop(u[1, ], vc)
pvinecop(u[1, ], vc)

## Discrete models
vc$var_types <- rep("d", 5)  # convert model to discrete

# with discrete data we need two types of observations (see Details)
x <- qpois(u, 1)  # transform to Poisson margins
u_disc <- cbind(ppois(x, 1), ppois(x - 1, 1))

dvinecop(u_disc[1:5, ], vc)
pvinecop(u_disc[1:5, ], vc)

# simulated data always has uniform margins
pairs(rvinecop(200, vc))
}
\seealso{
\code{\link[=vinecop_dist]{vinecop_dist()}}, \code{\link[=vinecop]{vinecop()}}, \code{\link[=plot.vinecop]{plot.vinecop()}}, \code{\link[=contour.vinecop]{contour.vinecop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvine_structure.R
\name{plot.rvine_structure}
\alias{plot.rvine_structure}
\alias{plot.rvine_matrix}
\title{Plotting R-vine structures}
\usage{
\method{plot}{rvine_structure}(x, ...)

\method{plot}{rvine_matrix}(x, ...)
}
\arguments{
\item{x}{an \code{rvine_structure} or \code{rvine_matrix} object.}

\item{...}{passed to \code{plot.vinecop_dist()}.}
}
\description{
Plot one or all trees of an R-vine structure.
}
\examples{
plot(cvine_structure(1:5))
plot(rvine_structure_sim(5))
mat <- rbind(c(1, 1, 1), c(2, 2, 0), c(3, 0, 0))
plot(rvine_matrix(mat))
plot(rvine_matrix_sim(5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vinecop_methods.R
\name{vinecop_predict_and_fitted}
\alias{vinecop_predict_and_fitted}
\alias{predict.vinecop}
\alias{fitted.vinecop}
\title{Predictions and fitted values for a vine copula model}
\usage{
\method{predict}{vinecop}(object, newdata, what = "pdf", n_mc = 10^4, cores = 1, ...)

\method{fitted}{vinecop}(object, what = "pdf", n_mc = 10^4, cores = 1, ...)
}
\arguments{
\item{object}{a \code{vinecop} object.}

\item{newdata}{points where the fit shall be evaluated.}

\item{what}{what to predict, either \code{"pdf"} or \code{"cdf"}.}

\item{n_mc}{number of samples used for quasi Monte Carlo integration when
\code{what = "cdf"}.}

\item{cores}{number of cores to use; if larger than one, computations are
done in parallel on \code{cores} batches.}

\item{...}{unused.}
}
\value{
\code{fitted()} and \code{predict()} have return values similar to \code{\link[=dvinecop]{dvinecop()}}
and \code{\link[=pvinecop]{pvinecop()}}.
}
\description{
Predictions of the density and distribution function
for a vine copula model.
}
\details{
\code{fitted()} can only be called if the model was fit with the
\code{keep_data = TRUE} option.
\subsection{Discrete variables}{

When at least one variable is discrete, two types of
"observations" are required in \code{newdata}: the first \eqn{n \; x \; d} block
contains realizations of \eqn{F_{X_j}(X_j)}. The second \eqn{n \; x \; d}
block contains realizations of \eqn{F_{X_j}(X_j^-)}. The minus indicates a
left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
\eqn{F_{X_j}(X_j^-) = F_{X_j}(X_j - 1)}. For continuous variables the left
limit and the cdf itself coincide. Respective columns can be omitted in the
second block.
}
}
\examples{
u <- sapply(1:5, function(i) runif(50))
fit <- vinecop(u, family = "par", keep_data = TRUE)
all.equal(predict(fit, u), fitted(fit), check.environment = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{pseudo_obs}
\alias{pseudo_obs}
\title{Pseudo-Observations}
\usage{
pseudo_obs(x, ties_method = "average", lower_tail = TRUE)
}
\arguments{
\item{x}{vector or matrix random variates to be converted (column wise) to
pseudo-observations.}

\item{ties_method}{similar to \code{ties.method} of \code{\link[=rank]{rank()}} (only \code{"average"},
\code{"first"} and \code{"random"} currently available).}

\item{lower_tail}{\code{logical} which, if `FALSE``, returns the
pseudo-observations when applying the empirical marginal survival
functions.}
}
\value{
a vector of matrix of the same dimension as the input containing the
pseudo-observations.
}
\description{
Compute the pseudo-observations for the given data matrix.
}
\details{
Given \code{n} realizations \eqn{x_i=(x_{i1}, \ldots,x_{id})},
\eqn{i \in \left\lbrace 1, \ldots,n \right\rbrace }
of a random vector \code{X}, the pseudo-observations are defined via
\eqn{u_{ij}=r_{ij}/(n+1)} for
\eqn{i \in \left\lbrace 1, \ldots,n \right\rbrace}
and
\eqn{j \in \left\lbrace 1, \ldots,d \right\rbrace }, where
\eqn{r_{ij}} denotes the rank of \eqn{x_{ij}} among all \eqn{x_{kj}},
\eqn{k \in \left\lbrace 1, \ldots,n \right\rbrace }.

The pseudo-observations can thus also be computed by component-wise applying
the empirical distribution functions to the data and scaling the result by
\eqn{n/(n+1)}. This asymptotically negligible scaling factor is used to force
the variates to fall inside the open unit hypercube, for example, to avoid
problems with density evaluation at the boundaries.

When \code{lower_tail = FALSE}, then \code{pseudo_obs()} simply returns
\code{1 - pseudo_obs()}.
}
\examples{
# pseudo-observations for a vector
pseudo_obs(rnorm(10))

# pseudo-observations for a matrix
pseudo_obs(cbind(rnorm(10), rnorm(10)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getters.R
\name{getters}
\alias{getters}
\alias{get_structure}
\alias{get_pair_copula}
\alias{get_all_pair_copulas}
\alias{get_parameters}
\alias{get_all_parameters}
\alias{get_ktau}
\alias{get_all_ktaus}
\alias{get_matrix}
\alias{get_family}
\alias{get_all_families}
\title{Extracts components of \code{bicop_dist} and \code{vinecop_dist} objects}
\usage{
get_structure(object)

get_pair_copula(object, tree = NA, edge = NA)

get_parameters(object, tree = NA, edge = NA)

get_ktau(object, tree = NA, edge = NA)

get_family(object, tree = NA, edge = NA)

get_all_pair_copulas(object, trees = NA)

get_all_parameters(object, trees = NA)

get_all_ktaus(object, trees = NA)

get_all_families(object, trees = NA)
}
\arguments{
\item{object}{a \code{bicop_dist}, \code{vinecop_dist} or \code{vine_dist} object.}

\item{tree}{tree index (not required if \code{object} is of class \code{bicop_dist}).}

\item{edge}{edge index (not required if \code{object} is of class \code{bicop_dist}).}

\item{trees}{the trees to extract from \code{object} (\code{trees = NA} extracts all
trees).}
}
\value{
The structure matrix, or pair-copulas, their parameters, Kendall's taus,
or families.
}
\description{
Extracts either the structure matrix  (for \code{vinecop_dist} only), or
pair-copulas, their parameters, Kendall's taus, or families (for \code{bicop_dist}
and \code{vinecop_dist}).
}
\details{
#' The \link{get_structure} method (for \code{vinecop_dist} or \code{vine_dist}
objects only) extracts the structure (see \link{rvine_structure} for more
details).

The \link{get_matrix} method (for \code{vinecop_dist} or \code{vine_dist} objects only)
extracts the structure matrix (see \link{rvine_structure} for more details).

The other \code{get_xyz} methods for \code{vinecop_dist} or \code{vine_dist} objects return
the entries corresponding to the pair-copula indexed by its \code{tree} and
\code{edge}. When \code{object} is of class \code{bicop_dist}, \code{tree} and \code{edge} are not
required. \cr
\itemize{
\item \code{\link[=get_pair_copula]{get_pair_copula()}} = the pair-copula itself (see \link{bicop}).
\item \code{\link[=get_parameters]{get_parameters()}} = the parameters of the pair-copula (i.e., a \code{numeric}
scalar, vector, or matrix).
\item \code{\link[=get_family]{get_family()}} = a character for the family (see \link{bicop} for
implemented families).
\item \code{\link[=get_ktau]{get_ktau()}} = a \code{numeric} scalar with the pair-copula Kendall's tau.
}

The \code{get_all_xyz} methods (for \code{vinecop_dist} or \code{vine_dist} objects only)
return lists of lists, with each element corresponding to a tree in \code{trees},
and then elements of the sublists correspond to edges. The returned lists
have two additional attributes:
\itemize{
\item \code{"d"} = the dimension of the model.
\item \code{"trees"} = the extracted trees.
}
}
\examples{
# specify pair-copulas
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(
  list(bicop, bicop), # pair-copulas in first tree
  list(bicop) # pair-copulas in second tree
)

# specify R-vine matrix
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)

# set up vine copula model
vc <- vinecop_dist(pcs, mat)

# get the structure
get_structure(vc)
all(get_matrix(vc) == mat)

# get pair-copulas
get_pair_copula(vc, 1, 1)
get_all_pair_copulas(vc)
all.equal(get_all_pair_copulas(vc), pcs,
          check.attributes = FALSE, check.environment = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vinecop_methods.R
\name{mBICV}
\alias{mBICV}
\title{Modified vine copula Bayesian information criterion (mBICv)}
\usage{
mBICV(object, psi0 = 0.9, newdata = NULL)
}
\arguments{
\item{object}{a fitted \code{vinecop} object.}

\item{psi0}{baseline prior probability of a non-independence copula.}

\item{newdata}{optional; a new data set.}
}
\description{
Calculates the modified vine copula Bayesian information criterion.
}
\details{
The modified vine copula Bayesian information criterion (mBICv) is defined as

\deqn{BIC = -2 loglik +  \nu log(n) - 2 \sum_{t=1}^{d - 1} (q_t log(\psi_0^t)
- (d - t - q_t) log(1 - \psi_0^t)) }

where \eqn{\mathrm{loglik}} is the log-likelihood and \eqn{\nu} is the
(effective) number of parameters of the model, \eqn{t} is the tree level
\eqn{\psi_0} is the prior probability of having a non-independence copula and
\eqn{q_t} is the number of non-independence copulas in tree \eqn{t}. The
mBICv is a consistent model selection criterion for parametric sparse vine
copula models.
}
\examples{
u <- sapply(1:5, function(i) runif(50))
fit <- vinecop(u, family = "par", keep_data = TRUE)
mBICV(fit, 0.9) # with a 0.9 prior probability of a non-independence copula
mBICV(fit, 0.1) # with a 0.1 prior probability of a non-independence copula
}
\references{
Nagler, T., Bumann, C., Czado, C. (2019). Model selection for
sparse high-dimensional vine copulas with application to portfolio risk.
\emph{Journal of Multivariate Analysis, in press}
(\url{https://arxiv.org/pdf/1801.09739.pdf})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicop_methods.R
\name{bicop_predict_and_fitted}
\alias{bicop_predict_and_fitted}
\alias{predict.bicop_dist}
\alias{predict.bicop}
\alias{fitted.bicop}
\title{Predictions and fitted values for a bivariate copula model}
\usage{
\method{predict}{bicop_dist}(object, newdata, what = "pdf", ...)

\method{fitted}{bicop}(object, what = "pdf", ...)
}
\arguments{
\item{object}{a \code{bicop} object.}

\item{newdata}{points where the fit shall be evaluated.}

\item{what}{what to predict, one of \code{"pdf"}, \code{"cdf"}, \code{"hfunc1"}, \code{"hfunc2"},
\code{"hinv1"}, \code{"hinv2"}.}

\item{...}{unused.}
}
\value{
\code{fitted()} and \code{logLik()} have return values similar to \code{\link[=dbicop]{dbicop()}},
\code{\link[=pbicop]{pbicop()}}, and \code{\link[=hbicop]{hbicop()}}.
}
\description{
Predictions of the density, distribution function, h-functions (with their
inverses) for a bivariate copula model.
}
\details{
\code{fitted()} can only be called if the model was fit with the
\code{keep_data = TRUE} option.
\subsection{Discrete variables}{

When at least one variable is discrete, more than two columns are required
for \code{newdata}: the first \eqn{n \times 2} block contains realizations of
\eqn{F_{X_1}(x_1), F_{X_2}(x_2)}. The second \eqn{n \times 2} block contains
realizations of \eqn{F_{X_1}(x_1^-), F_{X_1}(x_1^-)}. The minus indicates a
left-sided limit of the cdf. For, e.g., an integer-valued variable, it holds
\eqn{F_{X_1}(x_1^-) = F_{X_1}(x_1 - 1)}. For continuous variables the left
limit and the cdf itself coincide. Respective columns can be omitted in the
second block.
}
}
\examples{
# Simulate and fit a bivariate copula model
u <- rbicop(500, "gauss", 0, 0.5)
fit <- bicop(u, family = "par", keep_data = TRUE)

# Predictions
all.equal(predict(fit, u, "hfunc1"), fitted(fit, "hfunc1"),
          check.environment = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/truncate_model.R
\name{truncate_model}
\alias{truncate_model}
\alias{truncate_model.rvine_structure}
\alias{truncate_model.rvine_matrix}
\alias{truncate_model.vinecop_dist}
\alias{truncate_model.vine_dist}
\title{Truncate a vine copula model}
\usage{
truncate_model(object, ...)

\method{truncate_model}{rvine_structure}(object, trunc_lvl, ...)

\method{truncate_model}{rvine_matrix}(object, trunc_lvl, ...)

\method{truncate_model}{vinecop_dist}(object, trunc_lvl, ...)

\method{truncate_model}{vine_dist}(object, trunc_lvl, ...)
}
\arguments{
\item{object}{a model object.}

\item{...}{further arguments passed to specific methods.}

\item{trunc_lvl}{tree level after which the vine copula should be truncated.}
}
\description{
Extracts a truncated sub-vine based on a truncation level supplied by user.
}
\details{
While a vine model for a \code{d} dimensional random vector contains at most \code{d-1}
nested trees, this function extracts a sub-model based on a given truncation
level.

For instance, \code{truncate_model(object, 1)} results in a 1-truncated
vine (i.e., a vine with a single tree). Similarly \code{truncate_model(object, 2)}
results in a 2-truncated vine (i.e., a vine with two trees). Note that
\code{truncate_model(truncate_model(object, 1), 2)} returns a 1-truncated vine.
}
\examples{
# specify pair-copulas
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(
  list(bicop, bicop), # pair-copulas in first tree
  list(bicop) # pair-copulas in second tree
)

# specify R-vine matrix
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)

# set up vine structure
structure <- as_rvine_structure(mat)

# truncate the model
truncate_model(structure, 1)

# set up vine copula model
vc <- vinecop_dist(pcs, mat)

# truncate the model
truncate_model(vc, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvinecopulib.R
\docType{package}
\name{rvinecopulib}
\alias{rvinecopulib}
\title{High Performance Algorithms for Vine Copula Modeling}
\description{
Provides an interface to 'vinecopulib', a C++ library for vine copula
modeling based on 'Boost' and 'Eigen'. The 'rvinecopulib' package implements
the core features of the popular 'VineCopula' package, in particular
inference algorithms for both vine copula and bivariate copula models.
Advantages over 'VineCopula' are a sleeker and more modern API, improved
performances, especially in high dimensions, nonparametric and
multi-parameter families. The 'rvinecopulib' package includes 'vinecopulib'
as header-only C++ library (currently version 0.6.2). Thus users do not need
to install 'vinecopulib' itself in order to use 'rvinecopulib'. Since their
initial releases, 'vinecopulib' is licensed under the MIT License, and
'rvinecopulib' is licensed under the GNU GPL version 3.
}
\examples{
## bicop_dist objects
bicop_dist("gaussian", 0, 0.5)
str(bicop_dist("gauss", 0, 0.5))
bicop <- bicop_dist("clayton", 90, 3)

## bicop objects
u <- rbicop(500, "gauss", 0, 0.5)
fit1 <- bicop(u, family = "par")
fit1

## vinecop_dist objects
## specify pair-copulas
bicop <- bicop_dist("bb1", 90, c(3, 2))
pcs <- list(
  list(bicop, bicop), # pair-copulas in first tree
  list(bicop) # pair-copulas in second tree
)
## specify R-vine matrix
mat <- matrix(c(1, 2, 3, 1, 2, 0, 1, 0, 0), 3, 3)
## build the vinecop_dist object
vc <- vinecop_dist(pcs, mat)
summary(vc)

## vinecop objects
u <- sapply(1:3, function(i) runif(50))
vc <- vinecop(u, family = "par")
summary(vc)

## vine_dist objects
vc <- vine_dist(list(distr = "norm"), pcs, mat)
summary(vc)

## vine objects
x <- sapply(1:3, function(i) rnorm(50))
vc <- vine(x, copula_controls = list(family_set = "par"))
summary(vc)
}
\author{
Thomas Nagler, Thibault Vatter
}
\keyword{package}
